#include <dbsim.h>
#include <object.h>
#include <node.h>
#include <op.h>
#include <query.h>
#include <stdio.h>
#include <cost.h>
#include <language.h>
#include <db.h>

#ifndef BUFMGR_H
#include <bufmgr.h>
#endif

#include "bfbucket.h"

int BFBucket::numBFBuckets = 0;
BFBucket* BFBucket::freeBucket = NULL;       // no free buckets initially
extern bool PlotMode;
extern int Inst_StartIO;

//////////////////////////////////////////////////////////////////////////////

// Constructor.

BufMgr::BufMgr(int numBuffers, Node *owningNode)
{
  int i;

  myNode = owningNode;
  if (myNode == 0) {
    printf("BufMgr: constructor fed bad myNode.\n");
    Panic();  // BUG! terminate simulation!
  }

  numBuf = numBuffers;
  InvalidFramesHead = InvalidFramesTail = NIL;
  ValidFramesHead = ValidFramesTail = NIL;
  LovedFramesHead = LovedFramesTail = NIL;
  LikedFramesHead = LikedFramesTail = NIL;
  HatedFramesHead = HatedFramesTail = NIL;
  
  invalidQTable = new qtable("bufmgr-invalid");
  validQTable = new qtable("bufmgr-valid");
  loveQTable = new qtable("bufmgr-love");
  likeQTable = new qtable("bufmgr-like");
  hateQTable = new qtable("bufmgr-hate");
  
  ClearStats(); // set buffer stats to initial conditions

  // Allocates array of frame descriptors for each frame in buffer pool
    
  bufTable = new Frame[numBuf];
  if (!bufTable ) {
    printf("BufMgr() node %d: Couldn't create buffer table.\n", myNode->id);
    Panic();  // terminate simulation!
  }

  // Places all frames in the invalid queue initially
    
  for (i = 0; i < numBuf; i++) 
    MoveFrame(&bufTable[i], INVALID);
  
  // Allocates and initiallizes the PID->frame mapping hash table

  hashSize = (numBuf) + 1;
  hashTable = (BFBucket**) new char[sizeof(BFBucket*) * hashSize];
  if ( ! hashTable )  {
    printf("BufMgr() node %d: Couldn't create hashTable.\n", myNode->id);
    Panic();  // terminate simulation!
  }
  for (i = 0; i < hashSize; i++) 
    hashTable[i] = NIL;
}

//////////////////////////////////////////////////////////////////////////////

BufMgr::~BufMgr() {

  delete (char*) hashTable;
  delete [] bufTable;
//  delete [numBuf] bufTable; kbaix
}

//////////////////////////////////////////////////////////////////////////////

// Resets buffer stats to initial conditions.
void BufMgr::ClearStats() 
{
  numFree = numBuf;
  bufferHits = bufferRefs = noIOReads = noIOWrites = 0;
  invalidQTable->reset(); 
  validQTable->reset();
  hateQTable->reset(); 
  likeQTable->reset(); 
  loveQTable->reset();
}
  
//////////////////////////////////////////////////////////////////////////////

// Moves a frame from wherever it currently lives to the end of the specified
// queue.  Validity checks and modifies the frame state if necessary.

void BufMgr::MoveFrame(Frame* f, int toQ) {

  Frame **head;  // the ADDRESS of a circular frame queue head pointer
  Frame **tail;  // the ADDRESS of a circular frame queue tail pointer

  // Misc err checks...

  if (f == NIL) {
    printf("Frame::move received a NIL frame to move\n");
    Panic();
  }
  
  // If the frame is currently queued somewhere, finds the addresses of the
  // queue head and tail, then dequeues it.

  if (f->qid != NONE) {
    if (f->qid == HATE) {
      head = &HatedFramesHead;
      tail = &HatedFramesTail;
      hateQTable->note_exit();
    } else if (f->qid == LIKE) {
      head = &LikedFramesHead;
      tail = &LikedFramesTail;
      likeQTable->note_exit();
    } else if (f->qid == LOVE) {
      head = &LovedFramesHead;
      tail = &LovedFramesTail;
      loveQTable->note_exit();
    } else if (f->qid == INVALID) {
      head = &InvalidFramesHead;
      tail = &InvalidFramesTail;
      invalidQTable->note_exit();
    } else if (f->qid == VALID) {
      head = &ValidFramesHead;
      tail = &ValidFramesTail;
      validQTable->note_exit();
    } else {
      printf("BufMgr node %d::MoveFrame:  bad qid %d found in frame %d\n", 
	     myNode->id, f->qid, FrameIndex(f));
      Panic();  // BUG! terminate simulation!
    }
      
    if (*head == *tail) {  // only one element
      *head = *tail = NIL;
    } else {
      f->prev->next = f->next;
      f->next->prev = f->prev;
      if (f == *tail) {
	*tail = f->prev;
      } else if (f == *head) {
	*head = f->next;
      }
    }
  }
  
  // Sets the new qid in the frame, queues it to a new queue if necessary,
  // and adjusts the frame state if need be.
    
  f->qid = toQ;

  if (f->qid != NONE) {
    bool queueAtTail = TRUE;       // default is LRU
    if (f->qid == HATE) {
      queueAtTail = FALSE;         // hated pages are MRU
      head = &HatedFramesHead;
      tail = &HatedFramesTail;
      hateQTable->note_entry();
    } else if (f->qid == LIKE) {
      head = &LikedFramesHead;
      tail = &LikedFramesTail;
      likeQTable->note_entry();
    } else if (f->qid == LOVE) {
      head = &LovedFramesHead;
      tail = &LovedFramesTail;
      loveQTable->note_entry();
    } else if (f->qid == VALID) {
      head = &ValidFramesHead;
      tail = &ValidFramesTail;
      validQTable->note_entry();
    } else if (f->qid == INVALID) {
      head = &InvalidFramesHead;
      tail = &InvalidFramesTail;
      invalidQTable->note_entry();
      f->valid = FALSE;
      if (f->reserved) {
	printf("BufMgr node %d::MoveFrame:  attempt to move reserved frame %d to invalid frame queue.\n", 
	       myNode->id, FrameIndex(f));
	Panic();  // BUG! terminate simulation!
      }
    } else {
      printf("BufMgr node %d::MoveFrame:  bad qid %d found in frame %d\n", 
	     myNode->id, f->qid, FrameIndex(f));
      Panic();  // BUG! terminate simulation!
    }
      
    if (*head == NIL) {        // q is empty
      *head = f;
      *tail = f;
      f->next = f->prev = f;
    } else {
      (*head)->prev = f;    // circular link from head
      f->next = *head;      // circular link to the head
      (*tail)->next = f;
      f->prev = *tail;
      if (queueAtTail)      // LRU chain
	*tail = f;
      else
	*head = f;          // MRU chain   
    }
  }
}

//////////////////////////////////////////////////////////////////////////////

// Page cannot be fixed or in transit to/from disk -- so must search list
// until this condition is true.

Frame* BufMgr::FindUnusedFrame(int qID) {

  register Frame *f, *qHead;
  BFBucket *htEntry;

  if (qID == HATE)
     qHead = HatedFramesHead;
  else if (qID == LIKE)
     qHead = LikedFramesHead;
  else if (qID == LOVE)
     qHead = LovedFramesHead;
  else if (qID == INVALID)        
     qHead = InvalidFramesHead;
  else if (qID == VALID)        
     qHead = ValidFramesHead;
  else {
    printf("BufMgr::FindUnusedFrame got bad qID %d\n", qID);
    Panic();  // BUG! terminate simulation!
  }
      
  if (qHead == NIL) {
    return NIL;
  }

  // do first element outside of while loop, rest within.

  f = qHead;
  htEntry = GetHashTableEntry(f->pid);
  if ((htEntry->inTransit == FALSE) &&
      (f->fixCnt == 0)) {
    MoveFrame(f, NONE); // dequeue it
    return f;
  }

  f = f->next;
  while (f != qHead) {
    htEntry = GetHashTableEntry(f->pid);
    if ((htEntry->inTransit == FALSE) &&
	(f->fixCnt == 0)) {
      MoveFrame(f, NONE); // dequeue it
      return f;
    }
    f = f->next;
  } 
  return NIL;
}

//////////////////////////////////////////////////////////////////////////////

// same as lookup except return pointer to
// hash table entry and return even if page pid is in transit.

BFBucket *BufMgr::GetHashTableEntry(PID pid) {

  register BFBucket* p;

  p = hashTable[Hash(pid)];
  while (p != NIL) {
    if (p->pid.diskId < 0) {
      printf("GetHashTableEntry: got a negative disk %d in pid.\n",
	     p->pid.diskId);
      Panic();  // BUG! terminate simulation!
    }
    if (p->pid.pageId < 0) {
      printf("GetHashTableEntry: got a negative page id %d in pid.\n");
      Panic();  // BUG! terminate simulation!
    }

    if ((p->pid.diskId != pid.diskId) ||
	(p->pid.pageId != pid.pageId))  {
      // not in this slot, skip it 
      p = p->next;
    } else {
      return p;
    }
  }
    
  // there is no mapping for the page 

  printf("Node %d:BufMgr::GetHashTableEntry couldn't find pid %d.%d.\n", 
	 myNode->id, pid.diskId, pid.pageId);
  Panic();    // BUG! terminate simulation!
  return NIL; // make compiler happy about return value

}

//////////////////////////////////////////////////////////////////////////////

// Write frame "index" to disk.  Sets "inTransit" to
// TRUE before IO, FALSE afterwards.  Also, wakes up
// any waiters on ioComplete when the io is finished.

void BufMgr::WriteFrame(int index) {

  register BFBucket *htEntry;  // the entry corresponding to the frame

  // get corresponding hash table entry.
  htEntry = GetHashTableEntry(bufTable[index].pid);
  
  // mark frame in-transit.
  htEntry->inTransit = TRUE;

  // do the write
  myNode->cpu.use(((float) Inst_StartIO) / (Mips * 1000)); //milliseconds
  // Where to account for this cpu and disk usage in stats?  Nowhere for now!
  myNode->disks[bufTable[index].pid.diskId]->use(bufTable[index].pid.pageId,
					        DISK_WRITE, NO_PREFETCH);
  // signal that I/O is complete
  htEntry->inTransit = FALSE;

  // wake up any waiters.
  if (htEntry->ioComplete != NIL) {
     // wake up waiters
     htEntry->ioComplete->set(); 
  }
  noIOWrites++;
}

//////////////////////////////////////////////////////////////////////////////
//
// Given (diskId, pageNumber) pair, finds index of page in bufTable.
//
// This is complicated by the following situations:
//
// 1.  The page could be mapped into the buffer pool but
//     still "in transit" from disk to the buffer pool.
// 2.  The page could be in the buffer pool but in the process
//     of being written out to disk to "flush" the page.  
//
// It is possible that a process could wait() for the page to be read in, 
// but have the page be deleted again before the waiting process had a chance
// to start running again after the wait.
// To resolve these points there is a while loop within LookUp() that
// continually checks the page until it is certain that either the
// page is resident or that it is not.

int BufMgr::LookUp(PID pid) {

  register BFBucket* p;

  p = hashTable[Hash(pid)];
  while (p != NIL) {
    if ((p->pid.diskId != pid.diskId) ||
        (p->pid.pageId != pid.pageId))  {
      // not in this slot, skip it  
      p = p->next;
    } else {
      // there is a mapping for the page   
      if ( !(p->inTransit) ) {
        // no IO in progress, mapping is valid 
        return p->bufIndex;
      } else {
        // page is mapped but undergoing I/O, wait for it   
        if (p->ioComplete == NIL) {
            // allocate an event to wait on
            p->ioComplete = new global_event("ioComplete");
        }

        p->ioComplete->wait();
        // here I/O is complete, wake up and try loop again   
        // return to beginning of bucket 
        p = hashTable[Hash(pid)];
      }
    }
  }
  // there is no mapping for the page  
  return eNOTFOUND;
}

//////////////////////////////////////////////////////////////////////////////

// reserve a mapping for an incoming page

void BufMgr::HashTableReserve(PID pid) {
  register BFBucket* p;

  p = hashTable[Hash(pid)];

  // Check if hashtable already contains what is to be inserted 
  for ( ; p; p = p->next) {
    if ((p->pid.diskId == pid.diskId) &&
	(p->pid.pageId == pid.pageId)) {
      printf("Node %d::HashTableReserve: tried to insert repeat pid %d.%d.\n",
	     myNode->id, pid.diskId, pid.pageId);
      Panic();  // BUG! terminate simulation!
    } 
  }

  // fallen through to end of list, insert new entry at head of list 
  p = new BFBucket;

  // give it invalid index, since real index is unknown
  p->bufIndex = ILLEGAL_BUFINDEX;  

  p->pid = pid;

  // mark undergoing I/O
  p->inTransit = TRUE;

  // an event will be created by waiter if needed
  p->ioComplete = NIL;

  p->next = hashTable[Hash(pid)];
  hashTable[Hash(pid)] = p;
}

//////////////////////////////////////////////////////////////////////////////

// Installs mapping from pid to buffer pool index.  This routine should only 
// be called on a page known not to be mapped or intransit.  Hash table 
// collisions are resolved by chaining.

void BufMgr::InsertInHashTable(int index) {

  char* routine = "BufMgr::InsertInHashTable()";
  register BFBucket* 	p;
  PID pid;

  pid = bufTable[index].pid;
  p = hashTable[Hash(pid)];

  for ( ; p; p = p->next) {
    // Check if hashtable already contains what is to be inserted 
    if (p->bufIndex == index) {
      printf("Node %d::InsertInHashTable: pid %d.%d already allocated %d.\n",
	     myNode->id, pid.diskId, pid.pageId, index);
      Panic();  // BUG! terminate simulation!
    } 
  }

  // fallen through to end of list, insert new entry at head of list 
  p = new BFBucket;

  p->bufIndex = index;
  p->pid = pid;
  p->ioComplete = NIL;
  p->inTransit = FALSE;
  p->next = hashTable[Hash(pid)];
  hashTable[Hash(pid)] = p;
}

//////////////////////////////////////////////////////////////////////////////

// Invalidate mapping from pid to buffer pool index.

void BufMgr::RemoveFromHashTable(int index) {
  register BFBucket*	p;
  register BFBucket**	pp;

  if (bufTable[index].fixCnt > 0) {
    printf("Node %d: BufMgr::RemoveFromHashTable removed fixed page %d.%d.\n",
	   myNode->id, bufTable[index].pid.diskId, bufTable[index].pid.pageId);
  }

  pp = &hashTable[Hash(bufTable[index].pid)];
  p = *pp;

  for ( ; p; pp = &(p->next), p = p->next)  {
    if (index == p->bufIndex)  {
      if (p->inTransit) {
	printf("Node %d:BufMgr::RemoveFromHashTable: removed in-transit page!\n",
	       myNode->id);
      } else {
	*pp = p->next; // adjust previous bfbucket element in chain
	if (p->ioComplete != NIL) {
	  delete p->ioComplete;
	  p->ioComplete = NIL;
	}
	delete p;
	return;
      }
    } 
  }
  printf("Node %d:BufMgr::RemoveFromHashTable: couldn't find index %d in frame %d.\n",
	 myNode->id,
	 index, 
	 Hash(bufTable[index].pid));
  Panic();  // BUG! terminate simulation!
}

//////////////////////////////////////////////////////////////////////////////
//
// Finds a buffer for a new page to be read into.
//
// Currently priorities are as follows:
//
//    1. An invalid page if one exists.
//    2. An unfixed and hated page that is not intransit if one exists 
//       Priority within hated pages is LRU.
//    3. An unfixed "liked" page that is not intransit if one exists.
//       priority within "liked" pages is LRU.
//    4. An unfixed "loved" page that is not intransit if one exists.
//       priority within "loved" pages is LRU.
//    5. An unfixed "valid" page that is not intransit if one exists.
//       priority within "valid" pages is by "value" (queue is non-empty only
//       for straight LRU or a value-based buffer manager)

Frame* BufMgr::NewBuffer() {
  Frame* victim = NIL;

  // Checks invalid frames first, and returns immediately if one is found.

  if (InvalidFramesHead != NIL) {  
    victim = InvalidFramesHead;
    MoveFrame(victim, NONE);    // dequeue frame 
    victim->valid = TRUE;
    return victim;
  } 

  // No invalid frames are available, so checks other lists of valid frames.

  if (victim == NIL) 
    victim = FindUnusedFrame(HATE);
  if (victim == NIL) 
    victim = FindUnusedFrame(LIKE);
  if (victim == NIL) 
    victim = FindUnusedFrame(LOVE);
  if (victim == NIL) 
    victim = FindUnusedFrame(VALID);

  // Panics if no more available buffers -- BUG IN SCHED!!

  if (victim == NIL) {    
    printf("BufMgr node %d::NewBuffer(): no free buffers!\n", myNode->id);
    Panic();  // BUG! terminate simulation!
  } 

  // We have a valid but not fixed, refed, or in transit.
       
  if (victim->dirty)  // write it to disk before reusing frame
    WriteFrame(FrameIndex(victim));
  
  RemoveFromHashTable(FrameIndex(victim)); // invalidate mapping for the frame.

  return victim;
}

//////////////////////////////////////////////////////////////////////////////

// Installs new page (diskId, pageId) in buffer pool of node,
// return index, also fix the page.  Returns an error if page is already
// in the buffer pool.  Places the frame on the requested queue.

int BufMgr::AllocPage(int diskId, int pageId, int fileID, int qID) {

  register int i;
  PID pid;
  Frame *f;
  
  pid.diskId = diskId;
  pid.pageId = pageId;
  
  // Rejects request if page is already mapped.

  i = LookUp(pid);
  if (i >= 0) {
    printf("BufMgr node %d::AllocPage: Tried to allocate existing page.\n", 
	   myNode->id);
    printf("page was disk = %d, pageId = %d\n", 
	   bufTable[i].pid.diskId, bufTable[i].pid.pageId);
    fflush(stdout);
    return(NOTOK);
  }
     
  // Get a new frame
  f = NewBuffer();   // terminates simulation if no free buffers (sched bug)

  f->Clear();
  f->fixCnt = 1;
  f->pid = pid;
  f->fileID = fileID;
  f->op = (Op *) ILLEGAL_OP;
  numFree--;
  InsertInHashTable(FrameIndex(f)); // maps page to frame
  MoveFrame(f, qID);  // moves frame to appropriate LRU chain

  return (OK);
}

//////////////////////////////////////////////////////////////////////////////

// Fake AllocPage --- does absolutely nothing.
// Included here for possible future use and
// to make all stores similar (AllocPage, WritePage).
// Used to model store writing out a bucket
// of a join using a pre-reserved join buffer frame.

int BufMgr::BucketAllocPage(int diskId, int pageId, int fileID, int qID) 
{
  // do a bunch of nothing to keep CC from complaining about params not used.
  if (diskId);
  if (pageId);
  if (fileID);
  if (qID);
  return (OK);
}

//////////////////////////////////////////////////////////////////////////////

// Fake read --- just does a disk read, doesn't affect the
// buffer pool at all.  Used to model select reading bucket
// of a join into pre-reserved join buffer frame.

int BufMgr::BucketReadPage(int diskId, int pageId, int prefetch, Op *op,
                           int pagesLeft)
{
    // read the page 
    myNode->cpu.use(((float) Inst_StartIO) / (Mips * 1000)); //milliseconds
    op->query->cnt_CPUInstr += (float) Inst_StartIO;
    myNode->disks[diskId]->use(pageId, DISK_READ, prefetch, op, pagesLeft,
			       TRUE);
    noIOReads++;

    return (OK);
}

//////////////////////////////////////////////////////////////////////////////

// returns index of page with PID pid.  Causes I/O if
// page is not already in buffer pool; also fixes page.
// notes file if given.  Moves page to appropriate queue.

int BufMgr::ReadPage(int diskId, int pageId, int fileID, int qID,
		     int prefetch, int pageType, int idxLvl, 
		     Op *op, int pagesLeft)
{
  register int i;
  BFBucket *htEntry;
  PID pid;
  Frame *f;

  if (pageType); // make compiler happy.  only used for probablistic bufmgr
  pid.diskId = diskId;
  pid.pageId = pageId;
  bufferRefs++;
  if (op != NULL) op->query->cnt_bufRef++;  // counts refs in op, if present

  // First checks if the page is mapped and resident -- i.e. a hit.
  if ((i = LookUp(pid)) >= 0)  {
    bufferHits++;
    if (op != NULL) op->query->cnt_bufHit++; // counts hits in op, if present
    if (bufTable[i].fixCnt++ == 0)  numFree--;

    // Moves the frame to higher priority LRU chain, if required.
    if (qID > bufTable[i].qid)
      MoveFrame(&bufTable[i], qID);
      
    return (OK);
  }
    
  // Must read the page in.  First reserves mapping to block concurrent users,
  // then gets a free buffer to read the page into.  Next, prepares the empty
  // frame descriptors to receive the page and starts the IO.  The hash table
  // entry for the page must be reserved before asking for a new buffer frame,
  // because this process may get suspended during the call to NewBuffer 
  // (in the case where the only available frame is dirty and must be cleaned).
  // If two processes request the same page which requires IO, some mark must
  // be left on the wall by the winner so that the loser does not start 
  // another IO for the same page.  This mark is the act of reserving a hash
  // table entry for the page.  This is also why the "intransit" flag and 
  // the IO completion events are stored in the hash table, and not in the
  // buffer frame table, since they both must be accessible on a PAGE basis,
  // and not on a FRAME basis because in this rare case.

  HashTableReserve(pid);  
  f = NewBuffer();   // terminates simulation if no free buffers (sched bug)
  numFree--;

  f->pid = pid;
  f->op = (Op *) ILLEGAL_OP;   // avoid coincidental release
  f->fileID = fileID;
  f->idxLvl = idxLvl;
  f->Clear();                  // clear descriptor 
  f->fixCnt = 1;               // fix the page 

  htEntry = GetHashTableEntry(pid); // install mapping from pid to index i. 
  if (htEntry == NIL) {
    printf("BufMgr node %d::ReadPage: couldn't find entry for pid %d.%d.\n",
              myNode->id, pid.diskId, pid.pageId);
    Panic();  // BUG! terminate simulation!
  } 
  htEntry->bufIndex = FrameIndex(f); // fix up hash tbl entry to point to frame
  htEntry->inTransit = TRUE;

  // Moves the frame to the appropriate LRU chain. 
  MoveFrame(f, qID);

  // read the page 
  myNode->cpu.use(((float) Inst_StartIO) / (Mips * 1000)); //milliseconds
    op->query->cnt_CPUInstr += (float) Inst_StartIO;
  myNode->disks[pid.diskId]->use(pid.pageId, DISK_READ, prefetch, op, pagesLeft);
  noIOReads++;
  
  // wake up anyone waiting on this page (in lookup.) 
  htEntry->inTransit = FALSE;
  if (htEntry->ioComplete != NIL) htEntry->ioComplete->set();

  return (OK);
}

//////////////////////////////////////////////////////////////////////////////

// makes sure page is in buffer pool and fixed, marks  page
// as dirty, increments buffer refs and hits.

int  BufMgr::WritePage(int diskId, int pageId, Op *op)
{
  register int i;
  register PID  pid;

  pid.diskId = diskId;
  pid.pageId = pageId;
  
  if ((i = LookUp(pid)) < 0)  {
    printf("BufMgr node%d::WritePage:Non-resident page (disk=%d,pageId=%d)\n",
	   myNode->id, pid.diskId, pid.pageId);
    return(NOTOK);
  } 
  
  if (bufTable[i].fixCnt == 0)  {
    printf("BufMgr node %d::WritePage: Tried to write unfixed page.\n", 
	   myNode->id);  
    printf("page was disk = %d, pageId = %d\n", bufTable[i].pid.diskId, 
	   bufTable[i].pid.pageId);
    return(NOTOK);
  } 
  
  bufTable[i].dirty = TRUE;
  if (op != NULL) op->query->cnt_bufDirty++;  // counts dirties if needed
      
  return (OK);
}


//////////////////////////////////////////////////////////////////////////////

// Fake write --- just does a disk write, doesn't affect the
// buffer pool at all.  Used to model store writing out bucket
// of a join from a pre-reserved join buffer frame.

int BufMgr::BucketWritePage(int diskId, int pageId, Op *op) 
{
    // write the page 
    myNode->cpu.use(((float) Inst_StartIO) / (Mips * 1000)); //milliseconds
    op->query->cnt_CPUInstr += (float) Inst_StartIO;
    myNode->disks[diskId]->use(pageId, DISK_WRITE, NO_PREFETCH, op);
    noIOWrites++;
    if (op != NULL) op->query->cnt_bufDirty++;  // counts dirties, if needed
    return (OK);

}

//////////////////////////////////////////////////////////////////////////////

// Deletes page from the buffer pool.  Assumes page was not fixed or dirty.

int BufMgr::DiscardPage(int bufIndex)
{

  if ((bufIndex < 0) || (bufIndex >= numBuf)) {
    printf("BufMgr node %d::DiscardPage: out of range buffer number.\n", 
	   myNode->id);
    Panic();  // BUG! terminate simulation!
  }

  if (bufTable[bufIndex].fixCnt != 0) {
    printf("BufMgr node %d::DiscardPage: page was fixed!.\n", myNode->id);
    Panic();  // BUG! terminate simulation!
  } 
  
  if (bufTable[bufIndex].dirty) {
    printf("BufMgr node %d::DiscardPage: page was dirty!.\n", myNode->id);
    Panic();  // BUG! terminate simulation!
  } 
  
  if (!bufTable[bufIndex].valid) {
    printf("BufMgr node %d::DiscardPage: page was invalid!.\n", myNode->id);
    Panic();  // BUG! terminate simulation!
  }
  numFree++;

  RemoveFromHashTable(bufIndex);      // unmap page from frame
  MoveFrame(&bufTable[bufIndex], INVALID);   // invalidate frame

  return (OK);
}

//////////////////////////////////////////////////////////////////////////////

//Reset the buffer pool to initial condition
//  i.e. put everything in invalid pages list
int BufMgr::ResetBuffer()
{
  int i;

  // Places all frames in the invalid queue
  for (i = 0; i < numBuf; i++) MoveFrame(&bufTable[i], INVALID);
  
  // Clear the hash table
  for (i = 0; i < hashSize; i++) {
    delete hashTable[i];
    hashTable[i] = NIL;
  }

  ClearStats();
  return(OK);
}


//////////////////////////////////////////////////////////////////////////////

// Decrement fix count for page in slot "bufIndex".

int BufMgr::UnfixPage(int diskId, int pageId)
{
  register int bufIndex;
  PID pid;

  pid.diskId = diskId;
  pid.pageId = pageId;
  bufIndex = LookUp(pid);

  if ((bufIndex < 0) || (bufIndex >= numBuf)) {
    printf("BufMgr node %d::UnfixPage: couldn't find page with index %d.\n",
	   myNode->id, bufIndex);
    printf("Was looking for PID (disk = %d, pageId = %d)\n",
	   pid.diskId, pid.pageId);
    Panic();  // BUG! terminate simulation!
  } 

  if (bufTable[bufIndex].fixCnt == 0) {
    printf("BufMgr node %d::UnfixPage: page %d was not fixed!\n", 
	   myNode->id,
	   bufIndex);
    printf("\t pid was %d.%d\n", 
	   bufTable[bufIndex].pid.diskId,
	   bufTable[bufIndex].pid.pageId);
    Panic();  // BUG! terminate simulation!
  } 

  if (bufTable[bufIndex].valid == FALSE) {
    printf("UnfixPage: Asked to unfix an invalid page!\n");
    Panic();  // BUG! terminate simulation!
  } 
  
  if (--bufTable[bufIndex].fixCnt == 0) {
    numFree++;

  // Makes sure the frame is at the end of its LRU chain (least likely to be
  // replaced).  An unfix request is the officially designated time of last 
  // access to a page.

    MoveFrame(&bufTable[bufIndex], bufTable[bufIndex].qid);
  }
    
  return (OK);
}

//////////////////////////////////////////////////////////////////////////////

// Release all pages reserved by op.

int BufMgr::ReleaseBufferPages(Op *op)
{
  // do sequential scan of buffer pool.  If this routine becomes a bottleneck,
  // should be replaced by routine that uses auxiliary data structure mapping 
  // op's to pages.

  register int i;

  for (i = 0; i < numBuf; i++) {
    if ((bufTable[i].op == op) && (bufTable[i].reserved == TRUE)) { // got one!
       bufTable[i].reserved = FALSE;
       bufTable[i].op = (Op *) ILLEGAL_OP;
       MoveFrame(&bufTable[i], INVALID); // invalidate frame
       numFree++;
    }
  }
  return (OK);
}


//////////////////////////////////////////////////////////////////////////////

// Release numReleasePages of the pages reserved by op.
int BufMgr::ReleaseReservedPages(Op *op, int numReleasePages)
{
  // do sequential scan of buffer pool.  If this routine becomes a bottleneck,
  // should be replaced by routine that uses auxiliary data structure mapping 
  // ops to pages.

  register int i;
  int relCnt = 0;

  for (i = 0; (i < numBuf) && (relCnt < numReleasePages); i++) {
    if ((bufTable[i].op == op) && (bufTable[i].reserved == TRUE)) { // got one!
       relCnt++;
       bufTable[i].reserved = FALSE;
       bufTable[i].op = (Op *) ILLEGAL_OP;
       MoveFrame(&bufTable[i], INVALID); // invalidate frame
       numFree++;
    }
  }

  // if less reserved pages found flag an error
  if (relCnt != numReleasePages)
  {
    printf("node %d::ReleaseReservedPages: wanted %d found %d",
	   myNode->id, numReleasePages, relCnt);
    Panic();  // BUG! terminate simulation!
  }
  return (OK);
}


//////////////////////////////////////////////////////////////////////////////

// Reserve numPages for op.  Returns number of pages actually reserved.  
// Actual reservation is made by "fix"ing the pages in the name of the op.
// File is set to ILLEGAL_FILE to avoid flushing, diskId set to ILLEGAL_FILE 
// to avoid false "hits."  Returns number of pages actually reserved.
//
// Note: if the option-taking is working properly, it should never fail to 
// reserve the requested number of pages.

int BufMgr::ReserveBufferPages(int numPages, Op *op)
{
  register int i;
  register Frame* f;
  int actualNumberReserved;

  if (numPages > numFree) {
    actualNumberReserved = numFree;
    numFree = 0;
    printf("BARF! Carey said this would never happen.\n");
  } else { 
    actualNumberReserved = numPages;
    numFree -= numPages;
  }

  // now actually reserve the pages.
  for (i = 0; i < actualNumberReserved; i++) {
    f = NewBuffer(); // terminates sim if no free buffers (sched bug)
    f->reserved = TRUE;
    f->op = op;
    f->fileID = ILLEGAL_FILE;     // avoid coinc flushings.
    f->idxLvl = 0;
    f->pid.diskId = ILLEGAL_DISK; // prev othr ops from finding
  }
  return actualNumberReserved;
}


//////////////////////////////////////////////////////////////////////////////

// Destructively flush a frame.  That is, after the flush, the frame is free 
// and invalid.  Does a disk write if necessary.

int BufMgr::FlushFrame(int frame) {

  BFBucket *htEntry;

  if(bufTable[frame].fixCnt != 0) {
    printf("node %d::FlushFRAME: tried to flush fixed frame %d!\n", 
	   myNode->id, frame);
    Panic();  // BUG! terminate simulation!
  }
  if (!bufTable[frame].valid) {
    printf("node %d::FlushFRAME: tried to flush invalid frame %d!\n", 
	   myNode->id, frame);
    Panic();  // BUG! terminate simulation!
  } 
  if (bufTable[frame].reserved) {
    printf("node %d::FlushFRAME: tried to flush reserved frame %d!\n", 
	   myNode->id, frame);
    Panic();  // BUG! terminate simulation!
  } 

  if (!bufTable[frame].dirty) {
    RemoveFromHashTable(frame);     // clean, just unmap page from frame
    MoveFrame(&bufTable[frame], INVALID);  // invalidate frame

  } else {   // page is dirty
    htEntry = GetHashTableEntry(bufTable[frame].pid);
    if (htEntry->inTransit) {
      // page is being written.  Only legal way this could happen is if 
      // NewBuffer() is trying to steal the frame.  In this case, do not do 
      // anything, since NewBuffer() is doing the flush.
      ;
    } else {  // frame is dirty but not in transit, write it out. 
      WriteFrame(frame);
      RemoveFromHashTable(frame);     // unmap page from frame
      MoveFrame(&bufTable[frame], INVALID);  // invalidate frame
    }
  }
  return (OK);
}

//////////////////////////////////////////////////////////////////////////////

// flushes all pages with homes in "file".  ASSUMPTION:  FlushFile() should 
// only be called when caller is certain that there is no other activity on 
// the file, since this is a "destructive flush" that results in 
// the buffer pool being void of any pages from "file."

int BufMgr::FlushFile(int fileID) {
  register int i;
  int status;
  
  // sequential scan of buffer pool.  Might need to be replaced
  // if this becomes a bottleneck.

  for (i = 0; i < numBuf; i++)  {
    if ((bufTable[i].valid) && (bufTable[i].fileID == fileID)) {
      status = FlushFrame(i);
      if (status != OK) {
	return (NOTOK);
      }
    }
  }
  return (OK);
}

//////////////////////////////////////////////////////////////////////////////

int BufMgr::FlushDisk(int diskId)
{
  register int i;
  int flushCount = 0;
  int status;
  
  for (i = 0; i < numBuf; i++)  {
    if ((bufTable[i].valid) && (bufTable[i].pid.diskId == diskId)) {
      status = FlushFrame(i);
      if (status != OK) {
	return (NOTOK);
      }
    }
  }
  return (OK);
}

//////////////////////////////////////////////////////////////////////////////

int BufMgr::FlushPage(int diskId, int pageId)
{
  PID pid;
  
  pid.diskId = diskId; 
  pid.pageId = pageId;
  
  int bufIndex = LookUp(pid);
  
  if ((bufIndex < 0) || (bufIndex >= numBuf)) {
    printf("BufMgr node %d::FlushPage: couldn't find page.\n", myNode->id);
    return (NOTOK);
  } else if (!bufTable[bufIndex].valid) {
    printf("BufMgr node %d::FlushPage: page was not valid!\n", myNode->id);
    return (NOTOK);
  } else return FlushFrame(bufIndex);
}

//////////////////////////////////////////////////////////////////////////////

// Print stats from buffer pool.
void BufMgr::PrintBufferPoolStats() 
{
  char plotID[256];
  
  if (PlotMode) { 
      sprintf(plotID, "BufN%d", myNode->id); // label line for graphgen tool
  } else {
      sprintf(plotID, "  ");
  }
      
  printf("%s  %4d  %5.3f  %8d  %8d  %8d  %8d  %8.1f  %8.1f  %8.1f  %8.1f  %8.1f\n", 
	 plotID,
	 myNode->id, 
	 (bufferRefs == 0) ? 0 : (float) bufferHits/(float) bufferRefs, 
	 bufferHits, 
	 bufferRefs, 
	 noIOWrites, 
	 noIOReads,
	 invalidQTable->qlen(),
	 validQTable->qlen(),
	 hateQTable->qlen(),
	 likeQTable->qlen(),
	 loveQTable->qlen()
	 );
  if (Flags[TRACE_BUF])
    PrintEverything();
}

//////////////////////////////////////////////////////////////////////////////

void BufMgr::PrintBufferPool() 
{
  printf("Buffer pool for node %d:\n", myNode->id);
  for (int i = 0; i < numBuf; i++) 
    PrintFrame(&bufTable[i]);
}

//////////////////////////////////////////////////////////////////////////////

void BufMgr::PrintHashTable() {

  BFBucket* p;
  
  printf("Printing Hash Table for node %d (non-empty entries).\n", myNode->id);
  for (int i = 0; i < hashSize; i++) {
    p = hashTable[ i ];
    if (p != NIL) printf("\n\t ht[ %d ]: ", i);
    while (p != NIL) {
      if (p->bufIndex >= 0) {
	printf("(index %d, pid %d.%d, intransit %d, address %x) ", 
	       p->bufIndex, 
	       p->pid.diskId,
	       p->pid.pageId,
	       p->inTransit,
	       (int) p);
      }
      p = p->next;
    }
  }
  printf("\n");
}

//////////////////////////////////////////////////////////////////////////////

void BufMgr::PrintFrameQ(int qID) {PrintFrameQ(qID, FALSE);}

void BufMgr::PrintFrameQ(int qID, bool countOnly) {

  Frame *f, *qHead;
  char qNameBuf[10], *qName = &qNameBuf[0];
  int count = 0;  // for sanity bound!!

  if (qID == HATE) {
     qHead = HatedFramesHead;
     qName = "hated";
  } else if (qID == LIKE) {
     qHead = LikedFramesHead;
     qName = "liked";
  } else if (qID == LOVE) {
     qHead = LovedFramesHead;
     qName = "loved";
  } else if (qID == INVALID) {
     qHead = InvalidFramesHead;
     qName = "invalid";
  } else if (qID == VALID) {
     qHead = ValidFramesHead;
     qName = "valid";
  } else {
    printf("BufMgr::PrinFrameQ got bad qID %d\n", qID);
    Panic();  // BUG! terminate simulation!
  }

  f = qHead;
  printf("Node %d: printing %s frame queue.\n", myNode->id, qName);

  if (f == NIL) {
    printf("\tempty list.\n");
  } else {
    PrintFrame(f);
    f = f->next;
    while ((f != qHead) && (count < numBuf)) {
      if (!countOnly) PrintFrame(f);
      f = f->next;
      count++;
    }
    printf("\n%d %s frames.\n", count, qName);
  }
}

//////////////////////////////////////////////////////////////////////////////

void BufMgr::PrintFrame(Frame* f) {

  printf("\t idx %6d, val %1d, dir %1d, rsv %1d, fix %3d, qid %2d, PID: %d.%d, ilvl %d\n", 
	 FrameIndex(f), f->valid, f->dirty, f->reserved, f->fixCnt, f->qid, 
	 f->pid.diskId, f->pid.pageId, f->idxLvl);
}

//////////////////////////////////////////////////////////////////////////////

void BufMgr::PrintEverything() {
  PrintFrameQ(INVALID);
  PrintFrameQ(VALID);
  PrintFrameQ(LOVE);
  PrintFrameQ(LIKE);
  PrintFrameQ(HATE);
  PrintBufferPool();
  PrintHashTable();
  fflush(stdout);
}


