#define BUFMGR_H

///////////////////////////////////////////////////////////////////////////////
//
// Definition of the base BufMgr class.  The base buffer manager is a 3 level
// LRU scheme (loved, liked, and hated pages -- LRU within each set).  
// Other types of buffer managers can be declared as subclasses of this class,
// and either add additional state data and methods, override methods, or
// possibly leave state data declared in the base class unused.  
//
// WARNING:  Some public methods that are not overridden by any subclass are
// not declared as virtual functions (late binding) for performance reasons.
// These will have to be converted to virtual functions if any new subclasses
// override them.
//
///////////////////////////////////////////////////////////////////////////////

#define eNOERROR 0
#define eNOTFOUND (-1)
#define eNOFREEBUFFER (-2)
#define ILLEGAL_DISK (-1)
#define ILLEGAL_FILE (-1)
#define ILLEGAL_PAGE (-1)
#define ILLEGAL_OP (-1)
#define ILLEGAL_BUFINDEX (-1)
#define MAXPAGESLEFT 1000

// Values for bufTable.qid, indicates which queue a frame lives on currently.
// \Max value is 15, since qid field is only 4 bits wide to save space.
// A low value for these constants means the page is more likely to be 
// replaced (an undesireable page), a high value means it is less likely to
// be replaced (a valuable page).  Thus the constants are ordered in increasing
// "value" of pages.  NONE is a special case, since unqueued frames are 
// currently ineligable for replacement.  Note that some of these queues
// may not be used by particular subclasses of the BufMgr class.  For 
// example, the hate, like, and love queues are only used by the base class
// 3 level LRU buffer manager, and none of the queues is used by the 
// probablistic buffer manager.

#define INVALID 1 /* free frames not mapped to any page */
#define HATE 2   /* e.g. data pages that are seq scans */
#define LIKE 3   /* e.g. data pages that are randomly accessed */
#define LOVE 4   /* e.g. index pages */
#define VALID 14 /* e.g. pages with valid disk mappings */
#define NONE  15 /* an unqueued frame */

extern "C" {
  void exit(int);
}

class Node;
class Op;
class BFBucket;

////////////////////////////////////////////////////////////////////////////

// a physical page id

typedef struct {
    int pageId;
    int diskId;
} PID;

////////////////////////////////////////////////////////////////////////////

// descriptor for a buffer pool page frame

struct  Frame {
  Frame*        next;           // ptr to next frame on q (or NIL if not on q)
  Frame*        prev;           // ptr to prev frame on q (or NIL if not on q)
  PID           pid;            // logical id of page kbaix
  Op            *op;            // op "owning" the page
  int           fileID;         // unique file/index identifier for the page
  short         idxLvl;         // needed for frag fencing only
  unsigned      qid : 4;        // = HATE, LIKE, LOVE, VALID, or INVALID
  unsigned	dirty : 1;	// TRUE if dirty (written)
  unsigned	valid : 1;	// TRUE if frame is mapped to a page
  unsigned      reserved : 1;   // TRUE if page has been reserved (by operator)
  short		fixCnt;		// fix charges

  Frame() {
    fixCnt = 0;
    qid = NONE;
    op = NIL;
    prev = next = NIL;
    dirty = valid = reserved = FALSE;
    fileID = ILLEGAL_FILE;
    op = (Op*) ILLEGAL_OP;
    idxLvl = 0;
    pid.pageId = ILLEGAL_PAGE;
    pid.diskId = ILLEGAL_DISK;
  };

  void Clear()	{               // initialize a page descriptor for a new user
    fixCnt = 0;
    dirty = FALSE;
    valid = TRUE;
    reserved = FALSE;
  };

};

////////////////////////////////////////////////////////////////////////////

class BufMgr {
 protected:
  int		numBuf;           // total # buffer frames 
  int		numFree;          // # remaining free buffers
  Frame*        HatedFramesHead;  // head of hated frame q
  Frame*        HatedFramesTail;  // tail of hated frame q
  Frame*        ValidFramesHead;  // head of valid frame q
  Frame*        ValidFramesTail;  // tail of valid frame q
  Frame*        InvalidFramesHead;// head of invalid frame q
  Frame*        InvalidFramesTail;// tail of invalid frame q
  Frame*        LovedFramesHead;  // head of loved frame q
  Frame*        LovedFramesTail;  // tail of loved frame q
  Frame*        LikedFramesHead;  // head of liked frame q
  Frame*        LikedFramesTail;  // tail of liked frame q
  Node*         myNode;           // node where this bufmgr instance lives
  Frame*	bufTable;         // array of descriptors for ea frame in buff
  BFBucket** 	hashTable;        // maps PIDS (log page ID) to buff pool pages
  int		hashSize;         // Size of PID->frame mapping hash table
  int 		bufferHits;       // statistics
  int		bufferRefs;       // statistics
  int           noIOReads;        // statistics
  int           noIOWrites;       // statistics
  qtable*       invalidQTable;    // # invalid pages in buffer pool
  qtable*       validQTable;      // # valid pages in buffer pool
  qtable*       hateQTable;       // # hated pages in buffer pool
  qtable*       likeQTable;       // # liked pages in buffer pool
  qtable*       loveQTable;       // # loved pages in buffer pool

  // Hash func to convert PID into an index into PID->frame mapping hash table
  int Hash(PID pid)  { return (pid.pageId * (pid.diskId+1)) % hashSize; }

  // Converts frame address to buffer pool index (to save space storing it).
  inline int FrameIndex(Frame* f) 
    { return f - bufTable; } 

  // Moves frame to specified q.  Prints err msg and panics if err chks fail.
  virtual void MoveFrame(Frame* f, int toQ);

  // Returns buffer index of page PID if found -- waits for any active IO 
  int LookUp(PID pid);  

  // Searches the specified qID and returns the address of an unfixed, 
  // non-IO-in-progress frame.  Frame is dequeued upon return.  
  // Returns NIL if no such frame exists on specified queue.
  virtual Frame* FindUnusedFrame(int qID);

  // Write frame "index" to disk.  Sets "inTransit" to TRUE before IO, FALSE 
  // afterwards.  Also, wakes up waiters on ioComplete when IO is finished.
  void WriteFrame(int index);

  // Destructively flush a frame.  That is, after the flush, the frame is free
  // and invalid.  Does a disk write if necessary.
  int FlushFrame(int frame);

  // Initializes descriptor for page in hashTable.  Collisions are resolved
  // by chaining additional BFBucket objects together on the constested entry.
  void InsertInHashTable(int index); 

  // Deletes descriptor for page in hashTable.
  void RemoveFromHashTable(int index); 

  // Reserves mapping for pid.  This has nothing to do with reserving buffer 
  // pool frames (i.e. the ReserveBufferPages method below)!  It is used to
  // the handle the case of 2 or more people requesting the same page when IO
  // is required to bring the page into the buffer.  In this case, multiple
  // requestors will race (via this method) to reserve a slot for this page
  // in the hash table.  The winner starts the IO, the loser(s) wait for it.
  void HashTableReserve(PID pid);  
    
  // Returns pointer to node in hash table representing specified page.
  BFBucket* GetHashTableEntry(PID pid);  

  // Returns address of a free buffer frame. Terminates simulation if none are
  // available, since this means there is a bug in the schedulers load 
  // control.
  virtual Frame* NewBuffer(); 

  // Resets buffer stats to initial conditions.
  void ClearStats();

 public:
  BufMgr(int numBuffers = DEFAULT_BUFSIZE, Node *owningNode = 0);
  ~BufMgr();

  // Installs page lid in buffer pool of node.  Records file  if given; 
  // this allows the pages of a file to be flushed as a group.  
  // Places the page frame on the requested queue, and defaults the queue
  // to LIKE if unspecified.
  virtual int 
    AllocPage(int diskId, int pageId, int fileID, int qID = LIKE);

  // returns index of page (diskId, pageId) .  (causes I/O if
  // page is not already in buffer pool; also fixes page.)
  // file arg allows pages of a file to be flushed as a group.  
  // Places the page frame on the requested queue.
  // op should be specified if buffer and disk stats are to be recorded.
  // Page type is used only by the probabilistic buffer manager
  // to decide the hit probability (= INDEX_PAGE or DATA_PAGE)
  virtual int   
    ReadPage(int diskId, int pageId, int fileID, int qID, int prefetch,
	     int pageType, int indexLevel, Op* op = NULL, 
	     int pagesLeft = MAXPAGESLEFT);

  // Fake read --- just does a disk read, doesnt affect the
  // buffer pool at all.  Used to model select reading bucket
  // of a join into pre-reserved join buffer frame.
  // op should be specified if buffer and disk stats are to be recorded.
  int 
    BucketReadPage(int diskId, int pageId, int prefetch, Op* op = NULL,
		   int pagesLeft = MAXPAGESLEFT);

  // makes sure page is in buffer pool and fixed, marks  page
  // as dirty, increments buffer refs and hits.  File is
  // not needed here, since it is set by preceding AllocPage();
  // return error.  
  // op should be specified if buffer and disk stats are to be recorded.
  virtual int   
    WritePage(int diskId, int pageId, Op* op = NULL);

  // Fake AllocPage --- does absolutely nothing.  Used to model store writing 
  // out a bucket of a join using a pre-reserved join buffer frame.
  virtual int 
    BucketAllocPage(int diskId, int pageId, int fileID, int qID = LIKE);

  // Fake write --- just does a disk write, doesnt affect the
  // buffer pool at all.  Used to model store writing out bucket
  // of a join from a pre-reserved join buffer frame.
  // op should be specified if buffer and disk stats are to be recorded.
  int 
    BucketWritePage(int diskId, int pageId, Op* op = NULL);

  // delete page lid from the buffer pool.
  // assumes page was not fixed or dity.
  virtual int 
    DiscardPage(int bufIndex);

  // reset the buffer pool to the initial condition
  virtual int 
    ResetBuffer();

  // Decrement fix count for page (diskId, pageId).  
  // returns error.
  virtual int 
    UnfixPage(int diskId, int pageId);

  // Returns nonzero if page is dirty.
  virtual int 
    IsDirty(int bufIndex) { return bufTable[bufIndex].dirty; }

  // Try to reserve numPages for op.  Returns number
  // of pages actually reserved.  Actual reservation is
  // made by "fix"ing the pages in the name of the op.  
  virtual int 
    ReserveBufferPages(int numPages, Op *op);

  // Release all pages reserved by op.
  virtual int 
    ReleaseBufferPages(Op *op);

  // Release numReleasePages reserved by op.
  virtual int 
    ReleaseReservedPages(Op *op, int numReleasePages);

  // Flush all pages for a given file to disk.  This
  // is a destructive flush --- after flushing the
  // page frames are available for another page.
  virtual int 
    FlushFile(int fileID);

  // flush all dirty pages from disk diskId
  virtual int 
    FlushDisk(int diskId);

  // flush page (diskId, pageId);
  virtual int 
    FlushPage(int diskId, int pageId);

  // Print stats from buffer pool.
  virtual void 
    PrintBufferPoolStats(); 

  // The following Print routines are useful in debugging.
  // Warning: in large configurations, each routine will
  // print a lot of information...
  virtual void PrintBufferPool();

  virtual void PrintHashTable();

  virtual void PrintFrame(Frame* f);

  virtual void PrintFrameQ(int qID);
  virtual void PrintFrameQ(int qID, bool countOnly);

  virtual void PrintEverything();

  // Called when an internal error is detected
  void Panic() {
    fflush(stdout);
    PrintEverything();
    exit(4);
  }
    
};

