
// ---------------------------------------------------------------------------
// db support code for managing files and fragments (i.e., physical files)
// in order to model a horizontally partitioned database, including indexed
// and unindexed files
//
// some notes:
//   - files are partitioned across a subset of the disks of a set of sites
//   - a file fragment occupies a set of extents on some particular disk
//   - all but the last extent is always fully allocated (all pages in use)
//   - extents are managed on a per-node (i.e., local filesystem) basis
//   - indices are now supported as well, and they\'re fragmented similarly
//   - indices have been implemented for use only with statically-sized files
// ---------------------------------------------------------------------------


// ---------------------------------------------------------------------------
// external stuff needed for database file management code
// ---------------------------------------------------------------------------

class Disk;


// ---------------------------------------------------------------------------
// constant definitions related to database file management
// ---------------------------------------------------------------------------

// extent overallocation factor for dynamically growing files
#define EXTENT_FACTOR 2

// return value for indicating completion of file and index scans
#define END_OF_SCAN -1

// constant for requesting scan of all fragments (i.e., disks) for file scans
#define ALL_FRAGMENTS -1

// error return used when physical or logical page id expected
#define BAD_PAGE -2

// size of page/record pointers (in bytes) for index calculations
#define PTR_SIZE 4

// maximum expected (and/or allowed) index height
#define MAX_HEIGHT 6

// possible index types for indexing a file
#define CLUSTERED    0
#define UNCLUSTERED  1

// possible page types for index-based scanning of a file
#define INDEX_PAGE  0
#define DATA_PAGE   1

// possible file types
//   perm - normal database files/indices
//   temps - temporary files created by stores for results
//   tempj - temporary files created by joins for buckets/overflows
//       special cased to save storage as these files are on only 1 node & disk
//   NOTE: only temps and tempj files can be deleted
#define PERM 0
#define TEMPS 1
#define TEMPJ 2

// printable versions of the two page types above
static char *pgType[2] = { "INDEX_PAGE", "DATA_PAGE" };
inline void dummy2() {if (pgType);} // avoid "::pgType unused" compiler msg

// ---------------------------------------------------------------------------
// file management classes - for file fragments, files, and file scans
// ---------------------------------------------------------------------------

class File;	       		// definition will appear later

////////////////////////////////////////////////////////////////////////////
//
//  FileFrag class - used (privately) in implementation of File class

class FileFrag {
    friend class File;
    friend class FileScan;
    friend class Index;
    friend class IndexFrag;
    friend class IndexScan;
private:
    static int numFileFrags;
    static class FileFrag* freeFileFrag;
    FileFrag*  nextReq;        // linked list of free filefrags
    int fragNo;	       	       // fragment id for debugging, etc
    Disk* disk;	       	       // physical disk holding this fragment
    int primaryExtent;         // 1st page of primary extent (only perm files)
    int primaryExtentSize;     // size of  primary extent
    int* pageMap;      	       // array of first phys page per secondary extent
    int tuplesPerPage; 	       // number of tuples per physical page
    int lastPgTuples;  	       // number of tuples on very last page
    int numTuples;     	       // current number of tuples in fragment
    int numPages;      	       // current number of pages in fragment
    int numExtents;    	       // current number of extents in fragment
    int maxExtents;    	       // maximum number of extents for pageMap

    FileFrag(int myFragNo, Disk* myDisk, int tupsPerPage, int tupCount = 0,
	     int typeId = PERM);
    ~FileFrag();

    void            addExtents(int newExtents);
    void            addSpace(int newTups, int& numPgs, int& firstPg,
                             int& firstTups);
    int        	    addPage(int newTups);
    int        	    lookupPage(int logiPage);
    inline void     getTupleInfo(int tupleNum, int &logicalPage,
    		                 int &tupsOnPage, int& tupsBefore);
    void       	    print();
public:
    inline int      size();     // size of frag, in pages
    inline void* operator new(size_t size);     // new is overloaded
    inline void  operator delete(void* oldReq); // delete is overloaded
};

////////////////////////////////////////////////////////////////////////////
//
//  File class - provides notion of a horizontally partitioned file

class File {
    friend class FileScan;
private:
    int* fragDisk;              // ids of storage disks at each node
    FileFrag*** fragment;       // fragment pointer array
    static int numFiles;
    static class File* freeFile;
    File*  nextReq;         // linked list of free files

public:
    int numFragments;  		// number of fragments in file
    int fileNo;		       	// must be unique accross both rels & indices
    int numNodes;	      	// number of nodes holding fragments
    Node** fragNode;   	        // array of storage site pointers 
    int numDisks;	       	// number of fragment disks per node
    int tupleSize;	       	// size of each tuple (in bytes)
    int numTuples;	       	// number of tuples in file
    int tuplesPerPage;	       	// occupancy of each page (in tuples)
    int fileType;               // temp/perm file
    float hitProb;              // buffer hit probability


    // form a file on specified 'nodes' and 'disks'. 
    //    defaults:  file type = permanent, probability hit ratio = 0
    File(int fileId, int nodeCount, Node** nodes, int diskCount, 
         int* disks, int tupCount, int tupSize, int typeId = PERM,
         float probHit = 0);
    ~File();

    inline FileFrag*  getFrag(Node* node, int disk);
    inline int        addPage(Node* node, int disk, int newTups);
    inline void       addSpace(Node* node, int disk, int newTups,
    		               int& numPgs, int& firstPg, int& firstTups);
    inline Node*      storageNode(int nodeNum);
    inline int        storageDisk(int diskNum);
    void              deleteFrag(Node* node, int disk);
    void              print();
    int        size();     // size of file, in pages
    inline int        sizePerNode(); // size of file on each node, in pages
    inline void* operator new(size_t size);      // new is overloaded
    inline void  operator delete(void* oldReq);  // delete is overloaded
};

////////////////////////////////////////////////////////////////////////////
//
//  FileScan class - iterator for scanning a node\'s portion of a file

class FileScan {
private:
    File* file;	       		// file being scanned
    Node* node;	       		// node where scan is occurring
    int curDisk;       		// disk currently being scanned
    FileFrag* curFrag; 		// fragment currently being scanned
    int curPage;       		// current logical page in fragment
    bool allFrags;     		// true if scanning all fragments
    static int numFileScans;
    static class FileScan* freeFileScan;
    FileScan*  nextReq;         // linked list of free filescans

public:
    FileScan(File* filePtr, Node* nodePtr, int diskNo = ALL_FRAGMENTS);

    int		   nextPage(int& physPage, int& disk, int& numTuples, 
			    int& pagesLeft);
    inline void* operator new(size_t size);      // new is overloaded
    inline void  operator delete(void* oldReq);  // delete is overloaded
};

// ---------------------------------------------------------------------------
// index management classes - for index fragments, indices, and index scans
// ---------------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////////
//
//  IndexFrag class - used (privately) in implementation of Index class

class Index;

class IndexFrag {
    friend class Index;
    friend class IndexScan;
private:
    FileFrag* data;    		// data file fragment being indexed
    FileFrag* nodes;   		// place to hold index pages themselves
    int indexType;     		// clustered or unclustered
    int numEntries;    		// number of <key, rid> pairs in index
    int fanout;	       		// number of entries per non-last page
    int numLevels;     		// height of this index fragment
    int firstPage[MAX_HEIGHT];	// first (logical) page of each level
    int lastFanout[MAX_HEIGHT];	// fanout of last page on each level

public:
    IndexFrag(FileFrag* myData, int keySize, int myIndexType);
    ~IndexFrag();

    void  print();
    inline int      size();     // size of frag, in pages
    inline int      sizePerLevel(int level);  // size of ifrag level, in pages
};

////////////////////////////////////////////////////////////////////////////
//
//  Index class - for indexing a horizontally partitioned file

class Index {
    friend class IndexScan;
private:
    IndexFrag* fragment[MAX_NODES] // fragment pointer array
		       [MAX_DISKSPERNODE];
public:
    int indexNo;	       	// must be unique accross both rels & indices
    File* data;		        // data file being indexed
    int keySize;      	        // size of each tuple (in bytes)
    int indexType;	       	// clustered or unclustered
    float hitProb;              // buffer hit probability

    Index(int myIndexNo, File* myDataFile, int myKeySize, int myIndexType,
	  float probHit = 0);
    ~Index();

    inline IndexFrag* getFrag(Node* node, int disk);
    void         print();
    int          size();     // size of index, in pages
    inline int   sizePerNode(); // size of index on each node, in pages
    inline bool  rootPage(int nodeID, int diskID, int pageID); 
    int          tupleSize() { return data->tupleSize; };
    int          numEntries() { return data->numTuples; };
    int          numNodes() { return data->numNodes; };
};

////////////////////////////////////////////////////////////////////////////
//
//  IndexScan class - iterator for scanning a node\'s portion of a file

class IndexScan {
private:
    // general info about current index scan state
    Index* index;      		// index being scanned
    Node* node;	       		// node where scan is occurring
    float selectivity; 		// desired scan selectivity
    int curDisk;       		// disk currently being scanned
    IndexFrag* curFrag;	       	// fragment currently being scanned
    int returnPoint;	       	// return point for scan iteration
    // info about current index fragment being scanned
    int scanEntries;	       	// size of index range being scanned
    int firstEntry, lastEntry; 	// actual range of scanned entries
    int pathPages[MAX_HEIGHT]; 	// path of pages for index search
    int pathEntries[MAX_HEIGHT];// per-page occupancy information
    int lastLeafPage;	       	// last page of sequence set scan
    int firstLeafEntries;      	// first leaf\'s relevant entry count
    int lastLeafEntries;       	// first leaf\'s relevant entry count
    // info about the exact state of the current scan, i.e., things that
    // would simply be local variables if iterators were available in C++
    int upperBound, relPage, firstLeafNo, lastLeafNo;
    int i, cnt, page, tupCount, tupsBefore, tupsLeft, tupleNo;
    bool point;                 // 1 tuple, randomly chosen disk

public:
    IndexScan(Index* indexPtr, Node* nodePtr, float selectFrac);

    int   nextPage(int& physPage, int& pageType, int& disk, 
                   int& numSteps, int& indexLevel);
    void  resetScan();
};

// ---------------------------------------------------------------------------
// inline routine bodies for file- and index-related classes
// ---------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////

inline int File::sizePerNode() { return size() / numNodes; };

//////////////////////////////////////////////////////////////////////////////

inline int Index::sizePerNode() { return size() / data->numNodes; };

//////////////////////////////////////////////////////////////////////////////

inline bool  Index::rootPage(int nodeID, int diskID, int pageID) {
  register IndexFrag* f = fragment[nodeID][diskID];
  return f->nodes->lookupPage(f->firstPage[f->numLevels - 1]) == pageID;
}

//////////////////////////////////////////////////////////////////////////////

inline int FileFrag::size() { return numPages; }; // in pages

//////////////////////////////////////////////////////////////////////////////

inline int IndexFrag::size() { return nodes->size(); }; // in pages

//////////////////////////////////////////////////////////////////////////////

inline int IndexFrag::sizePerLevel(int level) { 
  if (level == 0) 
    return 1; // root page
  else {
    int lvlIndex = (numLevels - 1) - level; 
    return firstPage[lvlIndex + 1] - firstPage[lvlIndex];
  }
};

//////////////////////////////////////////////////////////////////////////////

inline void FileFrag::getTupleInfo(int tupleNum, int &logicalPage,
			        int &tupsOnPage, int& tupsBefore) {
    logicalPage = tupleNum / tuplesPerPage;
    tupsOnPage  = (logicalPage == numPages-1) ? lastPgTuples : tuplesPerPage;
    tupsBefore  = tupleNum % tuplesPerPage;
};

//////////////////////////////////////////////////////////////////////////////

inline void File::addSpace(Node* node, int disk, int newTups, int& numPgs,
                    int& firstPg, int& firstTups) {
    numTuples += newTups;
    if (fileType == TEMPJ) {
        fragment[0][0]->addSpace(newTups,numPgs,firstPg,firstTups);
    } else {
        fragment[node->id][disk]->addSpace(newTups,numPgs,firstPg,firstTups);
    }
};

//////////////////////////////////////////////////////////////////////////////

inline int File::addPage(Node* node, int disk, int newTups) {
    int rsltPage;

    if (fileType == TEMPJ) {
        if ((rsltPage = fragment[0][0]->addPage(newTups)) >= 0) {
	    numTuples += newTups;
	}
    } else {
        if ((rsltPage = fragment[node->id][disk]->addPage(newTups)) >= 0) {
	    numTuples += newTups;
	}
    }
    return(rsltPage);
};

//////////////////////////////////////////////////////////////////////////////

inline Node* File::storageNode(int nodeNum) {
    return(fragNode[nodeNum]);
};

//////////////////////////////////////////////////////////////////////////////

inline int File::storageDisk(int diskNum) {
    return(fragDisk[diskNum]);
};

//////////////////////////////////////////////////////////////////////////////

inline FileFrag* File::getFrag(Node* node, int disk) {
    if (fileType == TEMPJ) 
        return(fragment[0][0]);
    else
        return(fragment[node->id][disk]);
};

//////////////////////////////////////////////////////////////////////////////

inline IndexFrag* Index::getFrag(Node* node, int disk) {
    return(fragment[node->id][disk]);
};

//////////////////////////////////////////////////////////////////////////////
//
// Service rtn to instantiate a relation (data file).  Three forms:
//
// form 1:  disk randomly chosen, starting node for consecutive string of 
//          nodeCnt nodes is also randomly chosen.

int  createRelation(int tupCount, int tupSize, int nodeCnt);

//////////////////////////////////////////////////////////////////////////////
//
// form 2:  user specifies disk, starting node for consecutive string of 
//          nodeCnt nodes is randomly chosen.

int  createRelation(int tupCount, int tupSize, int nodeCnt, int diskId);

//////////////////////////////////////////////////////////////////////////////
//
// form 3:  user specifies disk, and specific nodes (same disk on each node)

int createRelation(int tupCount, int tupSize, int nodeCnt, 
		   int *nodeIds, int diskId);

//////////////////////////////////////////////////////////////////////////////

inline void* File::operator new(size_t size) {
    File *newReq;

    if (freeFile == NULL) {
	// allocate a new File
	newReq = (File *) new char[size];
	if (Flags[TRACE_LEAK])
	  printf("new file %d\n", ++numFiles);
    } else {
	// reuse an existing File
        newReq = freeFile;
	freeFile = freeFile->nextReq;
    }
    return newReq;
};

//////////////////////////////////////////////////////////////////////////////

inline void File::operator delete(void* oldReq) {
    // put the File on the free list
    ((File *) oldReq)->nextReq = freeFile;
    freeFile = (File *) oldReq;
};




//////////////////////////////////////////////////////////////////////////////

inline void* FileFrag::operator new(size_t size) {
    FileFrag *newReq;

    if (freeFileFrag == NULL) {
	// allocate a new FileFrag
	newReq = (FileFrag *) new char[size];
	if (Flags[TRACE_LEAK])
	  printf("new filefrag %d\n", ++numFileFrags);
    } else {
	// reuse an existing FileFrag
        newReq = freeFileFrag;
	freeFileFrag = freeFileFrag->nextReq;
    }
    return newReq;
};

//////////////////////////////////////////////////////////////////////////////

inline void FileFrag::operator delete(void* oldReq) {
    // put the FileFrag on the free list
    ((FileFrag *) oldReq)->nextReq = freeFileFrag;
    freeFileFrag = (FileFrag *) oldReq;
};





//////////////////////////////////////////////////////////////////////////////

inline void* FileScan::operator new(size_t size) {
    FileScan *newReq;

    if (freeFileScan == NULL) {
	// allocate a new FileScan
	newReq = (FileScan *) new char[size];
	if (Flags[TRACE_LEAK])
	  printf("new filescan %d\n", ++numFileScans);
    } else {
	// reuse an existing FileScan
        newReq = freeFileScan;
	freeFileScan = freeFileScan->nextReq;
    }
    return newReq;
};

//////////////////////////////////////////////////////////////////////////////

inline void FileScan::operator delete(void* oldReq) {
    // put the FileScan on the free list
    ((FileScan *) oldReq)->nextReq = freeFileScan;
    freeFileScan = (FileScan *) oldReq;
};


