#include <dbsim.h>
#include <object.h>
#include <node.h>
#include <language.h>
#include <db.h>
#include <strings.h>
#include <math.h>

extern int NumNodes;
extern int DisksPerNode;

int File::numFiles = 0;
File* File::freeFile = NULL;

//---------------------------------------------------------------------------
// class implementation for Files
//---------------------------------------------------------------------------

File::File(int fileId, int nodeCount, Node** nodes, int diskCount, int* disks,
           int tupCount, int tupSize, int typeId, float probHit)
{
    // variables used to compute sizes and ids of file fragments, etc.
    int tupsPerFragment, extraTups, fragTups, fragId, n, d;
    Disk *curDisk;
    char buf[5], *ftypeName = &buf[0];
  
    if (Flags[TRACE_LOCK]) {
	switch (typeId) {
	  case PERM:
	    ftypeName = "PERM";
	    break;
	  case TEMPJ:
	    ftypeName = "TEMPJ";
	    break;
	  case TEMPS:
	    ftypeName = "TEMPS";
	    break;
	}
	printf("File contructor:  fid %d, #nodes %d, #disks %d, tups %d, tupsize %d, type %s, hitprob %f\n", fileId, nodeCount, diskCount, tupCount, tupSize, 
	       ftypeName, probHit);
    }

    // record the file id for display/debugging purposes
    fileNo = fileId;
    fileType = typeId;

    // record the set of storage sites for the file
    numNodes = nodeCount;

    // for temporary join files, file on single disk of one node
    //   special cased to save storage
    if (fileType == TEMPJ) {
        fragNode = new Node* [1];
        fragNode[0] = nodes[0];
	if (tupCount > 0) {
	    printf("ERROR: non empty temporary join file %d\n", fileId);
	    exit(NOTOK);
	}
    } else { // default: file for all disks on all nodes
        fragNode = new Node* [NumNodes];
        for (n = 0; n < numNodes; n++) {
	    fragNode[n] = nodes[n];
	}
    }       

    // record the particular disks where the file is to reside
    if (fileType == TEMPJ) {  // temporary join file on only one disk
	numDisks = 1; // force disks = 1 regardless of what caller says
        fragDisk = new int [1];
	fragDisk[0] = disks[0]; // put on first disk caller specified
    } else { // default: file may reside on all disks
	numDisks = diskCount; // copy number of disks from caller
        fragDisk = new int [DisksPerNode];
	for (d = 0; d < numDisks; d++) {
	    fragDisk[d] = disks[d];
	}
    }

    // record the file-specific database characteristics
    tupleSize = tupSize;
    numTuples = tupCount;
    tuplesPerPage = PageSize / tupleSize;
    hitProb = probHit;  // only for probabilistic buffer manager

    // allocate a set of file fragments for representing the file, with
    // one fragment residing on each storage disk of each file storage node,
    // using a sparse data structure that will mesh well with the related
    // data structures that David is using for identifing active sites
    numFragments = numNodes * numDisks;
    tupsPerFragment = numTuples / numFragments;
    extraTups = numTuples % numFragments;

    // temporary join file: only one fragment; special cased to save storage
    if (fileType == TEMPJ) { 
        fragment = new FileFrag** [1];
        fragment[0] = new FileFrag* [1];
	fragment[0][0] = NULL;
    } else { //temp-store/permanent file: may have fragment on each disk/node
        fragment = new FileFrag** [NumNodes];
	for (n = 0; n < NumNodes; n++) {
	    fragment[n] = new FileFrag* [DisksPerNode];
	}
	// begin by clearing the fragment pointer array
	for (n = 0; n < NumNodes; n++) {
	    for (d = 0; d < DisksPerNode; d++) {
	        fragment[n][d] = NULL;
	    }
	}
    }
    // now initialize the actual fragment pointers

    if (fileType == TEMPJ) { // temporary join file
       fragId = 0;
       curDisk = storageNode(0)->disks[storageDisk(0)];
       fragTups = tupsPerFragment;		// all sites get this many
       if (extraTups-- > 0) fragTups++;	// some sites get one extra
       fragment[0][0] = new FileFrag(fragId, curDisk, tuplesPerPage, fragTups,
				     fileType);
    }
    else { // temporary-store / permanent file
        fragId = 0;
	for (n = 0; n < numNodes; n++) {
            for (d = 0; d < numDisks; d++) {
	        curDisk = storageNode(n)->disks[storageDisk(d)];
		fragTups = tupsPerFragment;	   // all sites get this many
		if (extraTups-- > 0) fragTups++; // some sites get 1 extra
		fragment[storageNode(n)->id][storageDisk(d)] =
		        new FileFrag(fragId, curDisk, tuplesPerPage, fragTups,
				     fileType);
		fragId++;
	    }
	}
    }
}


int File::size()
{
    int pgcnt = 0;
    if (numFragments > 0) {
        if (fileType == TEMPJ) { // temporary join file: only one fragment
	    pgcnt += fragment[0][0]->numPages;
	} else { // temporary-store/permanent file: may have multiple fragments
	    for (int n = 0; n < numNodes; n++) {
	        for (int d = 0; d < numDisks; d++) {
		    pgcnt += 
		      fragment[storageNode(n)->id][storageDisk(d)]->numPages;
		}
	    }
	}
    }
    return pgcnt;
}


File::~File()
{
    if (numFragments > 0) {
        if (fileType == TEMPJ) { // temporary join file: only one fragment
	    delete fragment[0][0];
	    delete fragment[0];
	    delete fragment;
	    delete fragDisk;
	    delete fragNode;
	} else if (fileType == TEMPS) { // temporary store file
	    for (int n = 0; n < numNodes; n++) {
	        for (int d = 0; d < numDisks; d++) {
		    delete fragment[storageNode(n)->id][storageDisk(d)];
		}
	    }
	    for (n = 0; n < numNodes; n++) 
	      delete [] fragment[n];
//	      delete [DisksPerNode] fragment[n]; kbaix
	    delete [] fragment;
	    delete [] fragDisk;
	    delete [] fragNode;
//	    delete [numNodes] fragment;         kbaix
//	    delete [DisksPerNode] fragDisk;     kbaix
//	    delete [numNodes] fragNode;         kbaix
	} else { // permanent file: cannot be deleted
	    printf("ERROR: attempt to delete permanent file %d\n", fileNo);
	    exit(NOTOK);
	}
    }
}


void File::deleteFrag(Node* node, int disk)
{
//  printf("File::deleteFrag(node %d, disk %d)", node->id, disk);

    if (fileType == TEMPJ) {
        delete fragment[0][0];
    } else {
        delete fragment[node->id][disk];
    }
    numFragments--;
    if (numFragments == 0) {
	delete this;
//      printf(" -- deleted file itself");
    }
//  printf(" -- %d fragments left\n", numFragments);
};


void File::print()
{
    int n, d;

    printf("File summary for file %d...\n", fileNo);
    printf("\tnumNodes = %d, numDisks = %d\n", numNodes, numDisks);
    printf("\tnodes = { ");
    for (n = 0; n < numNodes; n++) {
	printf("%d ", storageNode(n)->id);
    }
    printf("}\n");
    printf("\tdisks = { ");
    for (d = 0; d < numDisks; d++) {
	printf("%d ", storageDisk(d));
    }
    printf("}\n");
    printf("\ttuplesPerPage = %d, numTuples = %d, tupleSize = %d\n",
            tuplesPerPage, numTuples, tupleSize);
    printf("\tFile contains %d fragments...\n", numFragments);
    if (fileType == TEMPJ) {
        printf("\t  Displaying first and last file fragment\n");
        fragment[0][0]->print();
    } else {
        for (n = 0; n < numNodes; n++) {
	    for (d = 0; d < numDisks; d++) {
	        printf("\t  Displaying file fragment[%d, %d]\n",
		       storageNode(n)->id, storageDisk(d));
		fragment[storageNode(n)->id][storageDisk(d)]->print();
	    }
	}
    }
}




