#include <dbsim.h>
#include <object.h>
#include <node.h>
#include <language.h>
#include <db.h>
#include <strings.h>
#include <math.h>

int FileFrag::numFileFrags = 0;
FileFrag* FileFrag::freeFileFrag = NULL;

//---------------------------------------------------------------------------
// class implementation for FileFrag
//---------------------------------------------------------------------------

FileFrag::FileFrag(int myFragNo, Disk* myDisk, int tupsPerPage, int tupCount,
		   int typeId)
{
    if (Flags[TRACE_CONTROL])
      printf("FileFrag constructor:  frag# %d, tups/page %d, tupCnt %d\n", 
	     myFragNo, tupsPerPage, tupCount);

    // record context for fragment
    fragNo = myFragNo;
    disk = myDisk;
    numTuples = tupCount;
    tuplesPerPage = tupsPerPage;
    int leftOverTuples = numTuples % tuplesPerPage;
    lastPgTuples = (leftOverTuples == 0) ? tuplesPerPage : leftOverTuples;

    // allocate required amount of disk space - initially one large primary
    //   extent containing all the pages of the fragment. used to save storage
    //   on the pageMap.
    if (typeId == PERM) {
	numPages = numTuples / tuplesPerPage + (leftOverTuples != 0);
	numExtents = maxExtents = 1;
	primaryExtentSize = numPages;
	primaryExtent = disk->AllocExtent(primaryExtentSize);
    } else {
	// no initial extents for temporary files.
	numPages = numExtents = maxExtents = 0;
	primaryExtent = primaryExtentSize = 0;
	pageMap = NULL;
    }
}


FileFrag::~FileFrag()
{
    // give back the space currently occupied by the fragment
    for (int i = 0; i < numExtents; i++) {
	disk->FreeExtent(pageMap[i]);
    }
    delete [] pageMap;
//    delete [maxExtents] pageMap; kbaix
}


void FileFrag::addExtents(int newExtents)
{
    // first allocate a bigger page map if necessary
    int newNumExtents = numExtents + newExtents;
    if (newNumExtents > maxExtents) {
	int* oldPageMap = pageMap;
	int oldMaxExtents = maxExtents;
	maxExtents = EXTENT_FACTOR * newNumExtents;
	pageMap = new int[maxExtents];
	for (int e = 0; e < numExtents; e++) {
	    pageMap[e] = oldPageMap[e];
	}
	delete [] oldPageMap;
//	delete [oldMaxExtents] oldPageMap; kbaix
    }

    // allocate desired number of additional disk extents
    for (int i = numExtents; i < newNumExtents; i++) {
	pageMap[i] = disk->AllocExtent();
    }
    numExtents = newNumExtents;
}


void FileFrag::addSpace(int newTups, int& numPgs, int& firstPg, int& firstTups)
{
    int origPages, newPages, freePages, freeTups, leftOverTuples;

    // first record the change up at the file level and remember current size
    origPages = numPages;

    // start by using up any tuple space that exists on the last in-use
    // page of the fragment, and then return if that's already enough;
    // also, compute some of the appropriate return value information
    freeTups = tuplesPerPage - lastPgTuples;
    if (freeTups > 0) {
	// one or more free tuple slots on last page
	firstPg = numPages - 1;
	firstTups = (newTups < freeTups) ? newTups : freeTups;
	printf("using %d free slots, so first page is %d\n", freeTups, firstPg);
    } else {
	// no free tuple slots available on last page
	firstPg = numPages;
	firstTups = (newTups < tuplesPerPage) ? newTups : tuplesPerPage;
	printf("no free tuple slots on last page\n");
    }
    if (newTups <= freeTups) {
        numTuples += newTups;
        leftOverTuples = numTuples % tuplesPerPage;
        lastPgTuples = (leftOverTuples == 0) ? tuplesPerPage : leftOverTuples;
	numPgs = 0;
	return;
    }
    numTuples += freeTups;	// this much space is now taken care of,
    newTups -= freeTups;        // so go ahead and decrement the request

    // next, use up the tuple space that exists on any as-yet-unused pages
    // of the fragment's last extent, and return as above if this is enough
    freePages = numExtents * EXTENT_SIZE - numPages;
    freeTups = freePages * tuplesPerPage;
    if (newTups <= freeTups) {
        numTuples += newTups;
	numPages += newTups/tuplesPerPage + (newTups % tuplesPerPage != 0);
        leftOverTuples = numTuples % tuplesPerPage;
        lastPgTuples = (leftOverTuples == 0) ? tuplesPerPage : leftOverTuples;
	numPgs = numPages - origPages;
	return;
    }
    numTuples += freeTups;	// this much more space is taken care of,
    numPages += freePages;	// both tuple-space-wise and page-space-wise,
    newTups -= freeTups;        // so decrement the space request further

    // finally, any remaining space needs must be met by allocating one or
    // more additional extents, so go ahead and handle that case that way
    newPages = newTups / tuplesPerPage + ((newTups % tuplesPerPage) != 0);
    addExtents(newPages / EXTENT_SIZE + ((newPages % EXTENT_SIZE) != 0));
    numTuples += newTups;
    leftOverTuples = numTuples % tuplesPerPage;
    lastPgTuples = (leftOverTuples == 0) ? tuplesPerPage : leftOverTuples;
    numPages += newPages;
    numPgs = numPages - origPages;
}


int FileFrag::addPage(int newTups)
{
    // start by checking for tuple space on the last in-use page of the
    // fragment, and scream if there is any - because there should not be
    // free tuple slots there if the fragment is being properly extended via
    // the addPage facility - and also make sure the current request is
    // for (at most) one page\'s worth of tuples
    if (((numPages > 0) && (lastPgTuples != tuplesPerPage)) ||
        (newTups > tuplesPerPage)) {
        printf("numpages %d lastpgtuples %d newTups %d tuplesPerPage %d\n",
	       numPages, lastPgTuples, newTups, tuplesPerPage);
	printf("BARF in addPage:  DeWitt said this wouldn't happen!\n");
	return(BAD_PAGE);
    }

    // next, see if there are any as-yet-unused pages in the fragment\'s
    // last extent, and if not, allocate one new extent to accomodate the
    // new page to be added to the fragment;  either way, then allocate a new
    // page at the end of the fragment
    int freePages = numExtents * EXTENT_SIZE - numPages;
    if (freePages == 0) {
	addExtents(1);
    }
    lastPgTuples = newTups;
    numTuples += newTups;
    numPages += 1;

    // assign values to be returned

    return(lookupPage(numPages - 1));
}


int FileFrag::lookupPage(int logiPage)
{
    if (logiPage >= primaryExtentSize) {
	// page in secondary extent so find proper extent number
	int extentNum = ((logiPage - primaryExtentSize) / EXTENT_SIZE);
	return(pageMap[extentNum] + (logiPage % EXTENT_SIZE));
    } else 
      // page in primary extent
      return(primaryExtent + logiPage);
}


void FileFrag::print()
{
    printf("\t    FileFrag %d:\n", fragNo);
    printf("\t\ttuplesPerPage = %d, numTuples = %d, lastPgTuples = %d\n",
            tuplesPerPage, numTuples, lastPgTuples);
    printf("\t\tnumPages = %d, numExtents = %d\n", numPages, numExtents);
    printf("\t\tpageMap = { ");
    for (int e = 0; e < numExtents; e++) {
	printf("%d ", pageMap[e]);
    }
    printf("}\n");
}

