#include <dbsim.h>
#include <object.h>
#include <node.h>
#include <language.h>
#include <db.h>
#include <strings.h>
#include <math.h>

//---------------------------------------------------------------------------
// class implementation for IndexScans
//---------------------------------------------------------------------------

IndexScan::IndexScan(Index* indexPtr, Node* nodePtr, float selectFrac)
{
    // prepare to scan all fragments of index on node
    index = indexPtr;
    node = nodePtr;
    selectivity = selectFrac;
    returnPoint = 0;
}

void IndexScan::resetScan() {
    // reset a point query scan
    returnPoint = 0;
}

int IndexScan::nextPage(int& physPage, int& pageType, int& disk, 
			int& numSteps, int& indexLevel)
{
    int maxCnt;
    indexLevel = 0;

    // figure out where to go based on current iteration state
    switch (returnPoint) {
        case 0:  break;
        case 1:  goto POINT1;
        case 2:  goto POINT2;
        case 3:  goto POINT3;
        case 4:  goto POINT4;
        case 5:  goto POINT5;
    }

    // loop through all disks that store fragments of the index
    for (curDisk = 0; curDisk < index->data->numDisks; curDisk++) {

	// for a given index fragment, select the range of its index
        // entries to be scanned
        if (selectivity <= 0) {
	    // single-tuple selection
	    curDisk = random(0, index->data->numDisks-1); // random disk
            curFrag = 
	      index->fragment[node->id][index->data->storageDisk(curDisk)];
	    scanEntries = 1;
	    point = TRUE;
	} else {
	    // multiple-tuple selection
            curFrag = 
	      index->fragment[node->id][index->data->storageDisk(curDisk)];
	    scanEntries = int(selectivity * curFrag->numEntries);
	    point = FALSE;
	}
        upperBound = curFrag->numEntries - scanEntries;
        firstEntry = random(0, upperBound);
        lastEntry = firstEntry + scanEntries - 1;

        // compute path through the index bottom-up
        relPage = firstEntry;
        for (i = 0; i < curFrag->numLevels; i++) {
	    // identify page to access at level i
	    relPage = relPage / curFrag->fanout;
	    pathPages[i] = curFrag->firstPage[i] + relPage;
	    if ( (i == curFrag->numLevels - 1) ||
	         (pathPages[i] == (curFrag->firstPage[i+1]-1)) ) {
	        // on right fringe of tree at root or below
	        pathEntries[i] = curFrag->lastFanout[i];
	    } else {
	        // not on right fringe of tree at all
	        pathEntries[i] = curFrag->fanout;
	    }
        }

        // compute stopping point of the leaf-level scan
        firstLeafNo  = firstEntry / curFrag->fanout;
        lastLeafNo   = lastEntry  / curFrag->fanout;
        lastLeafPage = curFrag->firstPage[0] + lastLeafNo;
        if (lastLeafNo <= firstLeafNo) {
	    // single-page scan at leaf level
	    firstLeafEntries = scanEntries;
	    lastLeafEntries  = scanEntries;
        } else {
	    // multi-page scan at leaf level
	    firstLeafEntries = scanEntries;
            firstLeafEntries = curFrag->fanout - (firstEntry % curFrag->fanout);
            lastLeafEntries  = lastEntry % curFrag->fanout + 1;
        }

        // now enumerate the index scan a page at a time, first descending
        // along the chosen search path to a leaf page of the index
        for (i = curFrag->numLevels-1; i >= 0; i--) {
	    cnt = int(log(pathEntries[i]) / log(2)) + 1;
            physPage = curFrag->nodes->lookupPage(pathPages[i]);
            pageType = INDEX_PAGE;
            disk     = index->data->storageDisk(curDisk);
            numSteps = (i > 0) ? cnt : (cnt + firstLeafEntries);
	    indexLevel = (curFrag->numLevels - 1) - i;
	    returnPoint = 1;
	    return(OK);
            POINT1: ;
        }
    
	// now check for special case of no matching index entries
	if (scanEntries == 0) {
            returnPoint = 5;
	    numSteps = 0;
	    return(END_OF_SCAN);
	}

        // now scan the rest of the chosen sequence set of leaves
        for (i = pathPages[0] + 1; i <= lastLeafPage; i++) {
            pageType = INDEX_PAGE;
            disk     = index->data->storageDisk(curDisk);
            physPage = curFrag->nodes->lookupPage(i);
	    numSteps = (i < lastLeafPage) ? curFrag->fanout : lastLeafEntries;
	    returnPoint = 2;
	    return(OK);
            POINT2: ;
        }
    
        // last but not least, scan the relevant file fragment pages
        if (curFrag->indexType == UNCLUSTERED) {
	    // access the file fragment in an unclustered manner
	    for (i = firstEntry; i <= lastEntry; i++) {
	        // setup return values for this page of sequence set
                tupleNo = random(0, upperBound);
    	        curFrag->data->getTupleInfo(tupleNo,page,tupCount,tupsBefore);
                physPage = curFrag->data->lookupPage(page);
                pageType = DATA_PAGE;
                disk     = index->data->storageDisk(curDisk);
                numSteps = 1;
		if (!point)
		  returnPoint = 3;
		else
		  returnPoint = 5;  // no more than one disk for point query
	        return(OK);
                POINT3: ;
	    }
        } else {
	    // access the file fragment in a clustered manner
	    for (i = firstEntry, tupsLeft = scanEntries;
	         i <= lastEntry;  i += cnt, tupsLeft -= cnt) {
    	        curFrag->data->getTupleInfo(i, page, tupCount, tupsBefore);
	        maxCnt = tupCount - tupsBefore;
	        cnt = (tupsLeft < maxCnt) ? tupsLeft : maxCnt;
                physPage = curFrag->data->lookupPage(page);
                pageType = DATA_PAGE;
                disk     = index->data->storageDisk(curDisk);
                numSteps = cnt;
		if (!point)
		  returnPoint = 4;
		else
		  returnPoint = 5;  // no more than one disk for point query
	        return(OK);
                POINT4: ;
	    }
        }
    }
    POINT5: ;
    returnPoint = 5;
    return(END_OF_SCAN);
}

