#ifndef diskh
#define diskh

// ---------------------------------------------------------------------------
// constant definitions and global variables for use in disk facility
// ---------------------------------------------------------------------------

// Physical disk parameters
extern float SeekFactor,           // seek time = SeekFactor*sqrt(seek dist)
             MinLatency,           // time in msecs
             DiskRotation,         // time in msecs for one rotation
             SettleTime,           // time for arm to settle after seek (msecs)
             XferRate,             // transfer rate in megabytes/second
             XferTime;             // page transfer time in milliseconds
extern int   PagesPerCyl,	   // disk cylinder capacity
             DiskCachePages,       // cache context size (zero -> no cache)
	     DiskCacheContexts;    // number of cache contexts
 
// Defaults for physical disk constants.  Meant to model the Fujitsu M2266 
// drive as closely as possible.  NOTE DEFAULTS ASSUME 8K PAGES!!!!!!!!!!!
#define DEFAULT_SEEKFACTOR 0.617
#define DEFAULT_MINLATENCY 0.0
#define DEFAULT_SETTLETIME 2.0
#define DEFAULT_DISKROTATION 16.667
#define DEFAULT_XFERRATE 3.09
#define DEFAULT_PAGESPERCYL 83
#define DEFAULT_DISKCACHEPAGES 4
#define DEFAULT_DISKCACHECONTEXTS 8
#define MAXPAGESLEFT 1000

const int DISKQLEN_INTERVAL = 60000;   // 1 minute
const float DISKQLEN_ALPHA = .25;      // exponential weighting factor

// head directions for disk
#define IN              0
#define OUT             1
static char *headDir[2] = { "IN", "OUT" };
inline void dummy() {if (headDir);} // avoid "::headDir unused" compiler msg

// possible disk access types and prefetch requests
#define DISK_READ       0
#define DISK_WRITE      1
#define NO_PREFETCH     0
#define PREFETCH        1

// empty disk cache indicator and maximum cache size (in #contexts)
#define NO_PAGE			-1

// ---------------------------------------------------------------------------
// class definition for DiskReq (descriptor for a disk request) 
// ---------------------------------------------------------------------------

class DiskReq {
private:
    static int numDReqs;
    static class DiskReq* freeReq;	// list of free requests
public:
    int     cylinder;			// cylinder number of request
    int     page;                       // page number of request
    int     prefetch;			// true if prefetching desired
    DiskReq *nextReq;			// linked list of I/O requests
    global_event   *ioReady;		// event to set when disk is ready
    double  startTime;                  // time req arrived at disk 
    DiskReq(int pageId = 0, int prefReq = NO_PREFETCH) {
      page     = pageId;
      cylinder = pageId / PagesPerCyl;
      prefetch = prefReq;
      nextReq = NULL;  // indicate currently unqueued
      startTime = clock; // note req arrival time
    };
    inline void* operator new(size_t size);  // new is overloaded
    inline void operator delete(void* oldReq); // delete  overloaded
};

inline void* DiskReq::operator new(size_t size) {
    DiskReq *newReq;

    if (freeReq == NULL) {
	// allocate a new disk request
	newReq = (DiskReq *) new char[size];
        newReq->ioReady = new global_event("ioReady");
	if (Flags[TRACE_LEAK])
	  printf("new diskreq %d\n", ++numDReqs);
    } else {
	// reuse an existing disk request
        newReq = freeReq;
	freeReq = freeReq->nextReq;
    }
    return newReq;
};

inline void DiskReq::operator delete(void* oldReq) {
    // put the disk request on the free list
    ((DiskReq *) oldReq)->nextReq = freeReq;
    freeReq = (DiskReq *) oldReq;
};

// ---------------------------------------------------------------------------
// class definition for DiskQueue (queue of disk requests)
// ---------------------------------------------------------------------------

// definition of disk queue class
class DiskQueue {
private:
    DiskReq qhead;		// sentinal for disk queue
public:
    DiskQueue()			// disk queue constructor
        {
            qhead.nextReq = NULL;
        }
    void print();			// print queue contents
    void enqueueIn(DiskReq *newReq);	// insert into "in" queue
    void enqueueOut(DiskReq *newReq);	// insert into "out" queue
    int nonempty()			// see if queue is empty
        {
    	    return(qhead.nextReq != NULL);
        };
    DiskReq *dequeue() 			// extract next request
        {
    	    DiskReq *next = qhead.nextReq;
    	    qhead.nextReq = next->nextReq;
    	    return(next);
        };
};


// ---------------------------------------------------------------------------
// definition of ExtentList class for tracking available disk extents
// ---------------------------------------------------------------------------

class Extent {
    static int numExtents;
    static class Extent* freeExtDesc;	// list of unused extent descriptors
public:
    int firstPage;			// first page of this extent
    int size;                           // size of the extent
    class Extent *nextFree;		// pointer to next extent descriptor
    inline void* operator new(size_t size);      // new is overloaded
    inline void operator delete(void* oldReq);   // delete is overloaded
};

inline void* Extent::operator new(size_t size) {
    Extent *newExt;

    if (freeExtDesc == NULL) {
	// allocate a new extent descriptor
	if (Flags[TRACE_LEAK])
	  printf("new Extent %d\n", ++numExtents);
	newExt = (Extent *) new char[size];
    } else {
	// reuse an existing extent descriptor
        newExt = freeExtDesc;
	freeExtDesc = freeExtDesc->nextFree;
    }
    return newExt;
};

inline void Extent::operator delete(void* oldExt) {
    // put the disk request on the free list
    ((Extent *) oldExt)->nextFree = freeExtDesc;
    freeExtDesc = (Extent *) oldExt;
};

// ---------------------------------------------------------------------------
// definition of disk cache class for holding multiple prefetch contexts
// ---------------------------------------------------------------------------

class CacheContext {
public:
    int firstCachedPage;    	// first page in cache context
    int lastCachedPage;     	// last page in cache context
};

class DiskCache {
    CacheContext *context;			// FIFO array of contexts
    int curContext;				// current cache context
public:
    DiskCache();
    CacheContext* whichContext(int pageId);	// lookup page in cache
    inline CacheContext* nextCacheContext();	// get new cache context
    void print();				// print cache contents
};


// ---------------------------------------------------------------------------
// class definition for Disk (disk i/o facility) with space management
// ---------------------------------------------------------------------------

// specification for disk subclass of facility class

class Op;

class Disk: public facility {
protected:
	char dname[24];		// name of this disk facility
	int curCylinder;	// current head position
	int curDirection;	// head is moving either "in" or "out"
	DiskQueue inQueue; 	// requests for when head is moving "in"
	DiskQueue outQueue; 	// requests for when head is moving "out"
	table  *servStat;	// stats on disk ARM svc time (cache ignored!)
	qtable *waitStat;	// statistics on waiting requests
	qtable *intQlenStat;	// interval disk q len stats
	float avgQlength;       // exponentially weighted avg q length
	double qlenIntStartTime;// time when dql interval period started
	int nextCleanPage;	// first page of next never-used extent
	Extent *freeExtents;	// list of used but now-available extents
        DiskCache cache;	// cache for holding prefetched pages
	bool pendingHandoff;	// true if disk handoff is in progress
	table* RRSvc;           // random read svc time (cache included)
	table* SRSvc;           // sequential read svc time (cache included)
	table* WSvc;            // write svc time (cache included)
	double totDemand;       // tot svc demand for entire simulation
	DiskQueue highPrQueue; 	// high priority requests
    
	void queueReq(DiskReq* req);  // queue req according to elevator alg
	DiskReq* getNextReq();        // get next req according to elevator alg
public:
	Disk();
	Disk(char*);
	~Disk();
        virtual void use(int pageId);
        virtual int use(int pageId, int reqType, bool prefetch, Op* op = NULL, int pagesLeft = MAXPAGESLEFT, bool bucketRead = FALSE);
	int AllocExtent(int extentSize = EXTENT_SIZE);
        void FreeExtent(int firstPage);
	void print();
	virtual void printStats();
	void clearIntStats();
	void observeQlen();

	inline float diskUtil() { return util();};
	inline float avgQlen() { return avgQlength;}; // expon wted avg qlen
	inline float diskQlen() { return waitStat->qlen();};
	inline float randReadSvc() { return RRSvc->mean(); };
	inline float seqReadSvc() { return SRSvc->mean(); };
	inline float writeSvc() { return WSvc->mean(); };
};


// ---------------------------------------------------------------------------
// inline functions for DiskCache class
// ---------------------------------------------------------------------------

inline CacheContext* DiskCache::nextCacheContext()
{
    curContext = (curContext + 1) % DiskCacheContexts;
    return(&(context[curContext]));
};

#endif


