#include <dbsim.h>
#include <object.h>
#include <disk.h>
#include <node.h>
#include <op.h>
#include <query.h>
#include <language.h>
#include <strings.h>
#include <math.h>

extern JobClass *JobClasses[];
extern Node *Nodes[];

int Extent::numExtents = 0;
int DiskReq::numDReqs = 0;
DiskReq* DiskReq::freeReq = NULL;	// no free disk requests initially
Extent* Extent::freeExtDesc = NULL;        // no free descriptors initially

// ---------------------------------------------------------------------------
// class implementation for Disk (disk i/o facility)
// ---------------------------------------------------------------------------

Disk::Disk(): facility("disk")
{
    strcpy(dname, "disk");
    curCylinder = 0;
    curDirection = IN;
    servStat = new table("disk-serv");
    RRSvc = new table("disk-RRSvc");
    RRSvc->record(20.0); // prime it
    SRSvc = new table("disk-SRSvc");
    SRSvc->record(7.0); // prime it
    WSvc = new table("disk-WSvc");
    WSvc->record(20.0); // prime it
    waitStat = new qtable("disk-wait");
    intQlenStat = new qtable("disk-qlen");
    qlenIntStartTime = 0.0;
    totDemand = 0.0;
    nextCleanPage = 0;
    freeExtents = NULL;
    pendingHandoff = FALSE;
    avgQlength = 0.0;
    if (Flags[TRACE_CONTROL])
      printf("disk instantiation complete\n");
}

///////////////////////////////////////////////////////////////////////////////

Disk::Disk(char *s): facility(s)
{
    strcpy(dname, s);
    curCylinder = 0;
    curDirection = IN;
    servStat = new table(s);
    RRSvc = new table(s);
    RRSvc->record(20.0); // prime it
    SRSvc = new table(s);
    SRSvc->record(7.0); // prime it
    WSvc = new table(s);
    WSvc->record(20.0); // prime it
    totDemand = 0.0;
    waitStat = new qtable(s);
    intQlenStat = new qtable(s);
    qlenIntStartTime = 0.0;
    nextCleanPage = 0;
    freeExtents = NULL;
    pendingHandoff = FALSE;
    avgQlength = 0.0;
    if (Flags[TRACE_CONTROL])
      printf("disk %s instantiation complete\n", dname);
}

///////////////////////////////////////////////////////////////////////////////

Disk::~Disk()
{
}

///////////////////////////////////////////////////////////////////////////////

void Disk::use(int pageId)	// satisfy C++ with DISK_READ default
{
    use(pageId, DISK_READ, NO_PREFETCH);
}

///////////////////////////////////////////////////////////////////////////////

void Disk::observeQlen() {
    qlenIntStartTime = clock;
    if (avgQlength == 0.0) {
	avgQlength = intQlenStat->qlen();
    } else {
	avgQlength = (1.0 - DISKQLEN_ALPHA) * avgQlength + 
	          DISKQLEN_ALPHA * intQlenStat->qlen();
    }

    int curLen = intQlenStat->cur();
    for (int memNum = 0; memNum < curLen; memNum++) intQlenStat->note_exit();
    intQlenStat->reset();
    for (memNum = 0; memNum < curLen; memNum++) intQlenStat->note_entry();
}

///////////////////////////////////////////////////////////////////////////////

int Disk::use(int pageId, int reqType, bool prefetch, Op* op, int pagesLeft,
	       bool bucketRead)
{
    DiskReq* newReq;		// descriptor for this disk request
    DiskReq* nextReq;		// descriptor for next disk request
    int seekDist;		// number of cylinders to seek
    float diskTime;		// total time required for actual use of disk
    int nextCylPage1;		// first page number of next cylinder
    int cylPages;		// number of pages remaining in this cylinder
    int xtraPages;		// number of pages to prefetch in a read
    CacheContext* context;	// cache context for the current request
    float respTime;     	// resp time for this request (incl. queueing)

    // initialize
    xtraPages = 0;

    // note entry and allocate/initialize disk request descriptor

    waitStat->note_entry();
    intQlenStat->note_entry();
    newReq = new DiskReq(pageId);

    // Records disk reads in operator, if present

    if (op != NULL && reqType == DISK_READ) op->query->cnt_diskRead++; 

    if (Flags[TRACE_DISK]) {
        printf("%.3f(%d) i/o REQUEST [%s]: page = %d\n", 
	       clock, identity(), dname, pageId);
    }
    
    // if disk is currently busy, then queue the request and wait

    if (facility::status() == BUSY || pendingHandoff) {
	queueReq(newReq);
        newReq->ioReady->wait();
	pendingHandoff = FALSE;
    }

    // now use the disk facility for the appropriate amount of time,
    // figured using the Bitton/Gray formula, and then dispose of the
    // request and note that it has been served;  also account for a
    // multi-page cache with a write-through protocol for updates (also
    // updating the cached copy of the page, if any)

    facility::reserve();
    
    if (Flags[TRACE_DISK]) {
        printf("%.3f(%d) i/o START [%s]: dir = %s, cyl = %d, new cyl = %d\n", 
                clock, identity(), dname, headDir[curDirection], curCylinder,
		newReq->cylinder);
    }

    if ((reqType == DISK_WRITE) || (DiskCachePages*DiskCacheContexts == 0)) {

	// caching doesn\'t save time for writes or if cache size is zero

        if (Flags[TRACE_DISK]) {
            printf("%.3f(%d) i/o cache unused/invalidated [%s]\n", 
                   clock, identity(), dname);
        }

	seekDist = abs(newReq->cylinder - curCylinder);
        curCylinder = newReq->cylinder;	// note new head position
        diskTime =  (SeekFactor * sqrt((double) seekDist)) +
                    uniform(MinLatency, DiskRotation) + SettleTime + XferTime;
        hold(diskTime);
        servStat->record(diskTime);

	// NOTE: cache contents are left undisturbed in this case!

	// Records disk arm resp time on a job class basis (too expensive to 
        // record at any finer granularity, and still have accurate rollup
        // to higher levels)

	if (op != NULL) {
            op->query->jobClass->armResp->record(diskTime);
            op->query->jobClass->armWriteResp->record(diskTime);
        }

    } else if (cache.whichContext(newReq->page) == NULL) {

	// a cache miss has occurred for a read request

        if (Flags[TRACE_DISK]) {
            printf("%.3f(%d) i/o cache miss [%s], ", clock, identity(), dname);
	    cache.print();
	    printf("\n");
	    
        }

	// first compute how many pages are to be prefetched (as xtraPages)

	if (prefetch == PREFETCH) {
	    // may prefetch up through end of current cylinder
	    nextCylPage1 = (newReq->cylinder + 1) * PagesPerCyl;
	    cylPages = nextCylPage1 - (newReq->page + 1);

            // take the minimum of DiskCachePages, pages left on the cylinder
            //   and pages left in file fragment as the number of extra pages
            //   to be prefetched into the cache
	    xtraPages = (DiskCachePages < cylPages) ? DiskCachePages : cylPages;
	    xtraPages = (xtraPages < pagesLeft) ? xtraPages : pagesLeft;
	} else {
	    // no prefetch requested for this read operation
	    xtraPages = 0;
	}

	// now go ahead and read one plus xtraPages disk pages

	seekDist = abs(newReq->cylinder - curCylinder);
        curCylinder = newReq->cylinder;	// note new head position
        diskTime = (SeekFactor * sqrt((double) seekDist)) +
                    uniform(MinLatency, DiskRotation) + SettleTime +
		    (1 + xtraPages) * XferTime;
        hold(diskTime);
        servStat->record(diskTime);

	// cache must be updated to contain any/all prefetched pages

	if (xtraPages > 0) {
	    context = cache.nextCacheContext();
	    context->firstCachedPage = newReq->page + 1;
	    context->lastCachedPage  = newReq->page + xtraPages;
	}

        if (Flags[TRACE_DISK]) {
            printf("%.3f(%d) i/o [%s] new context, ", clock, identity(), dname);
	    cache.print();
	    printf("\n");
        }

	// Records seek distances on a job class basis (too expensive to 
        // record at any finer granularity, and still have accurate rollup
        // to higher levels)

	if (op != NULL) {
            op->query->jobClass->armResp->record(diskTime);
            op->query->jobClass->armReadResp->record(diskTime);
        }

    } else {

        // a cache hit has occurred for a read request

        diskTime = 0.0;
        if (Flags[TRACE_DISK]) {
            printf("%.3f(%d) i/o cache hit [%s], ", clock, identity(), dname);
	    cache.print();
	    printf("\n");
        }

	// Records disk cache hits in operator, if present
	
	if (op != NULL) op->query->cnt_diskCacheHit++; 

    }

    // Updates service time statistics.

    totDemand += diskTime;
    if (reqType == DISK_WRITE)
      WSvc->record(diskTime);
    else if (prefetch == PREFETCH)
      SRSvc->record(diskTime);
    else 
      RRSvc->record(diskTime);

    if (Flags[TRACE_DISK]) {
        printf("%.3f(%d) i/o DONE [%s], dir = %s, cyl = %d\n",
                clock, identity(), dname, headDir[curDirection], curCylinder);
    }

    facility::release();
    waitStat->note_exit();
    intQlenStat->note_exit();
    if ((simtime() - qlenIntStartTime) > DISKQLEN_INTERVAL) observeQlen();

    // Records total request resp time on a job class basis

    respTime = simtime() - newReq->startTime; // calc request resp time
    if (respTime < 0.0) respTime = 0.0; // patch for CSIM BUG in facil::reserve
    if (op != NULL) {
        op->query->jobClass->diskResp->record(respTime);
        if (reqType == DISK_WRITE) {
	    op->query->jobClass->diskWriteResp->record(respTime);
        } else {
            op->query->jobClass->diskReadResp->record(respTime);
            if (bucketRead) {
                op->query->jobClass->diskBktReadResp->record(respTime);
            } else {
                op->query->jobClass->diskNonBktReadResp->record(respTime);
	    }
        }
    }

    // finally, pass the disk to the next requester, if any

    nextReq = getNextReq();
    if (nextReq != NULL) {
	// notify next requester of disk availability
	pendingHandoff = TRUE;
        nextReq->ioReady->set();
    } else if (highPrQueue.nonempty()) { // service high priority req if any
	nextReq = highPrQueue.dequeue();
	pendingHandoff = TRUE;
	nextReq->ioReady->set();
    }

    delete newReq;

    // return number of pages read/written
    return (1 + xtraPages);
}

///////////////////////////////////////////////////////////////////////////////
// Queues a disks request according to the elevator algorithm.

void Disk::queueReq(DiskReq* newReq) {
    
    if (curCylinder > newReq->cylinder) {
	// inner cylinder requests handled on the way in
	inQueue.enqueueIn(newReq);
    } else if (curCylinder < newReq->cylinder) {
	// outer cylinder requests handled on the way out
	outQueue.enqueueOut(newReq);
    } else if (curCylinder == newReq->cylinder) {
	// requests for current cylinder handled first
	if (curDirection == IN) {
	    inQueue.enqueueIn(newReq);
	} else { // curDirection == OUT
	    outQueue.enqueueOut(newReq);
	}
    }
}

///////////////////////////////////////////////////////////////////////////////
// Selects the next request according to the elevator algorithm.

DiskReq* Disk::getNextReq() {
    DiskReq* nextReq;
    if (curDirection == IN) {
        // disk head is currently moving in
	if (inQueue.nonempty()) {
	    // get next request in current direction
	    nextReq = inQueue.dequeue();
	} else if (outQueue.nonempty()) {
	    // change direction and get next request
	    curDirection = OUT;
	    nextReq = outQueue.dequeue();
	} else {
	    nextReq = NULL;
	}
    } else {  // curDirection == OUT
	// disk head is currently moving out
	if (outQueue.nonempty()) {
	    // get next request in current direction
	    nextReq = outQueue.dequeue();
	} else if (inQueue.nonempty()) {
	    // change direction and get next request
	    curDirection = IN;
	    nextReq = inQueue.dequeue();
	} else {
	    nextReq = NULL;
	}
    }
    return nextReq;
}

///////////////////////////////////////////////////////////////////////////////

// allocate extent of size \'extentsize\' (default size - EXTENT_SIZE)
int Disk::AllocExtent(int extentSize)
{
    int firstPage;			// starting page of new extent
    Extent* ext;			// descriptor for a free extent
    if ((extentSize != EXTENT_SIZE) || (freeExtents == NULL)) {
	// if no free extents or not standard size (for perm files)
	// allocate a brand new extent. free list can be used only for
	// standard sized extents
	firstPage = nextCleanPage;
	nextCleanPage += extentSize;

    } else {
	// take a previously used extent from the free list
	ext = freeExtents;
	freeExtents = freeExtents->nextFree;
	firstPage = ext->firstPage;
	delete ext;
    }
    return(firstPage);
}

///////////////////////////////////////////////////////////////////////////////

void Disk::FreeExtent(int firstPage)
{
    Extent* ext = new Extent;		// get a new extent descriptor,
    ext->firstPage = firstPage;		// note the id of its first page,
    ext->nextFree = freeExtents;	// and add it to the free list
    freeExtents = ext;
}

///////////////////////////////////////////////////////////////////////////////

void Disk::print() 
{
    printf("\tDisk [%s]:\n\t\t curr cyl %d, curr direction %s, next clean pg %d\n\t\twaitStat qtable @  X'%x', freeExtent list @ X'%x'\n", dname, curCylinder, headDir[curDirection], nextCleanPage, waitStat, freeExtents);
    printf("\t\tIn queue:");
    inQueue.print();
    printf("\n\t\tOut queue:");
    outQueue.print();
    printf("\n\t\t");
    cache.print();
    printf("\n");
    fflush(stdout);
}    

//extern "C" double sqrt(double); kbaix

///////////////////////////////////////////////////////////////////////////////

void Disk::printStats()
{
    printf("%5.3f %7.3f %7.3f %7.3f %7.3f %7.3f %5.3f %7d %7d %5.2f %4d\n",
	   util(),  // utilization "at controller"
	   resp(),  // or waitStat->qtime() gives same answer
	   servStat->mean(), // mean svc time "at disk"
	   servStat->min(),
	   servStat->max(),
	   (float) sqrt((double) servStat->var()), // std dev disk svc
	   
	   // completions() counts requests received by the "controller",
	   // servStat->cnt() counts requests that miss the cache
	   // and are serviced by the disk, the difference = # cache hits,
	   // ratio of cache hits to controller reqs = cache hit rate
	   
	   (completions() == 0) ? 0.0 :
	   (float) (completions() - servStat->cnt()) / completions(),
	   
	   completions(),             // "controller" completions
	   servStat->cnt(),     // "disk" completions
	   waitStat->qlen(),   // mean queue len
	   waitStat->max());    // max queue len
//	   RRSvc->mean(),
//	   SRSvc->mean(),
//	   WSvc->mean(),
//	   totDemand);

}	 



