#include "dbsim.h"
#include "object.h"
#include "mailbox.h"
#include "msg.h"
#include "node.h"
#include "bufmgr.h"
#include "split.h"
#include "language.h"
#include "db.h"
#include "op.h"
#include "query.h"
#include "sched.h"
#include "cost.h"

extern "C" {
  double ceil(double x);
  double floor(double x);
  void exit(int);
}

extern int PageSize;
extern int CurTempFileNum;
void writeOutputPage(Node*, Op*, File*, int);
void readPage(Node*, Op*, FileScan*, int&, int&);
int findBucket(Op*);

// routine to simulate a join operation operating on a disk-resident
// relation

void joinOperator(JoinOp* op)
{
  
  Node*       n;
  Msg*	msg;
  CntlMsg     *cMsg;
  DataMsg     *dMsg;
  int	      instCnt;
  int	      tupleCnt, tupleCnt2, fullTuples; // #tuples processed by op
  int         numSites;
  float       tuplesProduced; // running sum of tuples produced
  int         tupNum; // used to model fractional tuples
  int	      inTupCnt;
  TIME	startTime; // time operator starts running
  int	      eofCnt;	  // number of end of stream indicators encountered
  int	      tupsPerNetPageOut;
  int	      buildTupsPerPage, probeTupsPerPage;


  float	sf;		// local value of operator selectivity factor
  int	      splitCnt;
  SplitTable  *splitTable;	
  Mailbox*    inPort;		// mailbox for process to receive messages
  int         bucketNum;      // number of bucket being processed
  int         bNum;           // dummy index variable
  int         pagesReserved;  // actual number of pgs reserved by the BufMgr
  int         hashPages;      // total number of pages avail to the hash tbl
  int         currentPages;   // total amount of memory used by join till now
  int         e ;             // return code from the buffer manager
  int         scanRC;         // return value from readPage
  
  int         tupsOnDisk;     // tuples sent to disk
  int         tupsInHash;     // tuples to be inserted in hash table
  float       overflowRatio;  // % of tuples sent to disk
  int         overflowTups;   // number of overflow tuples
  bool        overflow;       // indicates if overflow has occured
  bool        overflowFile;   // indicates if overflow file being processed
  int         outputBucket;   // bucket to which tuple to be sent
  File        *buildOverflowFile = NULL, *probeOverflowFile = NULL,
      *currBuildFile, *currProbeFile;
  FileScan    *buildFileScan, *probeFileScan;
  int         diskId;
  bool        done;
  BucketRec*  outputArray;
  
  create("join");             // creates CSIM process 
  
  n = op->node;
  // construct mail box for process to receive messages
  inPort = new Mailbox("opMbox", n, op);
  
  // initializations 
  startTime = clock;
  tupsPerNetPageOut = NETPKTSIZE/(op->outTupSize + 15);
  // 15 for padding and alignment ???
  buildTupsPerPage = PageSize / op->inTupSizeL;
  probeTupsPerPage = PageSize / op->inTupSizeR;

  sf = op->selFactor;
  if (Flags[TRACE_OP])
    printf("%f(%d) join started on node %d, sf=%f\n",
	   clock, identity(), n->id,sf);
  
  // begin by reserving as many pages as have been allocated to the join
  pagesReserved = n->bufMgr->ReserveBufferPages(op->pagesAllocated, op);
  if (pagesReserved != op->pagesAllocated)
    printf("BIG PROBLEM!!!!!: operator reserved %d pages allocated (%d)\n"
	   , pagesReserved, op->pagesAllocated);
  
  // send the operators private port to the query 
  // scheduler that started the operator
  cMsg = new CntlMsg;
  cMsg->opCode = NEWTASK;
  cMsg->intData = n->id;  // kind of a kludge to get the node id returned
  cMsg->send (op->initiator, inPort);
  
  // consume some CPU starting the operator
  n->cpu.use(((float) INITIATE_JOIN) / (Mips * 1000)); // milliseconds
  op->query->cnt_CPUInstr += (float)  INITIATE_JOIN;


  if (((JoinPlan*) (op->query))->numBuckets > 1)
    outputArray = new BucketRec[((JoinPlan*) (op->query))->numBuckets];
  
  // make local files for each of the extra buckets
  for(bucketNum = 1; bucketNum < ((JoinPlan*) (op->query))->numBuckets; 
      bucketNum++) {
    outputArray[bucketNum].numTuples = 0;
    
    diskId = random(0, DisksPerNode - 1);
    outputArray[bucketNum].buildFile = 
      new File(CurTempFileNum++, 1, &n, 1, &(diskId), 0, op->inTupSizeL, 
	       TEMPJ);
    diskId = random(0, DisksPerNode - 1);
    outputArray[bucketNum].probeFile = 
      new File(CurTempFileNum++, 1, &n, 1, &(diskId), 0, op->inTupSizeR, 
	       TEMPJ);
  }
  
  // if there are multiple buckets, the join simply loops around after it is 
  // done with the BUILD and PROBE phase for each bucket.
  // for the first bucket the tuples arrive from the child operator directly
  // for the subsequent buckets the tuples are read from local files


  for(bucketNum = 0; bucketNum < ((JoinPlan*) (op->query))->numBuckets; 
      bucketNum++) {

    if (Flags[TRACE_OP])
      printf("%f(%d) Join (node %d) processing bucket %d\n", 
	     clock, identity(), n->id, bucketNum);
    
    overflow = FALSE;
    overflowFile = FALSE; 
    
    // each bucket may produce overflow files so loops over all of them
    do {
      // remove pages needed for the local store/selects to find memory
      // available to the hash table
      if (bucketNum == 0)
	if (overflow)
	  hashPages = pagesReserved;
	else
	  hashPages = pagesReserved - 
	    (((JoinPlan*)op->query)->numBuckets - 1);
      else
	hashPages = ((JoinPlan*)op->query)->bucketKMemory ;
      
      // if overflow or extra buckets set files from which data to be
      // read in build and probe phase

      if (overflow) {
	overflowFile = TRUE;  // tuples from overflow file
	currBuildFile = buildOverflowFile;
	buildFileScan = new FileScan(currBuildFile, n);
	currProbeFile = probeOverflowFile;
	probeFileScan = new FileScan(currProbeFile, n);
	if (Flags[TRACE_OP])
	  printf("%f(%d) Join (node %d) processing overflow for bucket %d\n",
		 clock, identity(), n->id, bucketNum);
      } else {
	overflowFile = FALSE;
	if (bucketNum > 0) {
	  currBuildFile = outputArray[bucketNum].buildFile;
	  buildFileScan = new FileScan(currBuildFile, n);
	  currProbeFile = outputArray[bucketNum].probeFile;
	  probeFileScan = new FileScan(currProbeFile, n);
	}
      }
      
      tupleCnt = 0;
      tupsOnDisk = 0;
      tuplesProduced = 0;
      eofCnt = 0;
      overflowRatio = 0;
      overflow = FALSE;  // reset overflow flag
      overflowTups = 0;
      done = FALSE;
      
      while (!done) {    // with build phase
	inTupCnt = 0;
	tupsInHash = 0;
	// if first bucket and no overflow file
	if ((bucketNum == 0) && (overflowFile == FALSE)) {
	  // read pages off the network
	  msg = inPort->recv(); 
	  if (msg->type == CONTROL)  {
	    printf("join on node %d got illegal control msg opcode=%d\n",
		   n->id, cMsg->opCode);
	    delete cMsg; 
	    exit(NOTOK);
	  } else {
	    // message is another packet of tuples
	    dMsg = (DataMsg*) msg;
	    
	    if (Flags[TRACE_OP]) {
	      printf("%f(%d) join-build (node %d) rcvd netpage (%d tuples) ",
		     clock, identity(), n->id, dMsg->numTuples);
              printf("from source box (%x)\n", dMsg->sourceBox);
            }
	    
	    // calculate amount of CPU time consumed copying 
	    // the tuples from the network page to the hash table
	      if (dMsg->eofFlag == TRUE) {
		eofCnt++;
		if (eofCnt == op->inStreamsL) done = TRUE;
	      }
	    
	    // distribute tuples among all the buckets
	    for (tupNum = 0; tupNum < dMsg->numTuples; tupNum++) {
	      outputBucket = findBucket(op);
	      if (outputBucket == 0)
		inTupCnt++;
	      else {
		outputArray[outputBucket].numTuples++;
		
		// if output page full write to disk
		if (outputArray[outputBucket].numTuples > buildTupsPerPage) {
		  writeOutputPage(n, op, 
				  outputArray[outputBucket].buildFile,
				  buildTupsPerPage);
		  outputArray[outputBucket].numTuples -= 
		    buildTupsPerPage;
		}
	      }
	    }
	    delete dMsg; 
	  }   
	} else {
	  
	  // page to be read from local file (overflow/extra buckets)
	  readPage(n, op,  buildFileScan, scanRC, inTupCnt);
	  if (scanRC == END_OF_SCAN) {
	    done = TRUE;
	    e = n->bufMgr->FlushFile(currBuildFile->fileNo);
	    if (e != OK) printf("Join:buildfile flush returned error %d\n",e);
	    delete buildFileScan;
	    delete currBuildFile;
	  }
	}
	
	// charge cpu for reading the tuples and hashing them
	instCnt = inTupCnt * (INST_TREAD + INST_BUILD); 
	n->cpu.use(((float) instCnt) / (Mips * 1000)); // msecs
	op->query->cnt_CPUInstr += (float)  instCnt;

	// if there is an overflow file, send tuples to disk
	if (overflow) {
	  overflowTups += inTupCnt;
	  if (overflowTups > buildTupsPerPage) {
	    writeOutputPage(n, op, buildOverflowFile, 
			    buildTupsPerPage);
	    overflowTups -= buildTupsPerPage;
	    tupsOnDisk += buildTupsPerPage;
	  }
	  tupsInHash = 0;
	} else {
	  tupsInHash = inTupCnt;
	  
	  // find tuples and pages in hash table
	  tupleCnt += tupsInHash;
	  currentPages = (int) ceil((tupleCnt * op->inTupSizeL * 
				     FUDGEFACTOR * 1.0) / PageSize );
	  
//	  printf("current %d hash %d tupsInHash %d tupleCnt %d\n", 
//		 currentPages, hashPages, tupsInHash, tupleCnt);
	  
	  // an extra page is added to the capacity of the
	  //   hash table to account for minor differences in the
	  //   estimate of what the memory can hold and what is
	  //   actually sent to the join operator
	  if (currentPages > (hashPages + 1))
	    {
	      // mark spool flag so every next packet is processed properly 
	      overflow = TRUE;
	      if (Flags[TRACE_OP])
		printf("%f(%d) join-node %d overflow hashPages %d\n", 
		       clock, identity(), n->id, hashPages);
	      
	      
	      // make overflowFile
	      diskId = random(0, DisksPerNode - 1);
	      buildOverflowFile = 
		new File(CurTempFileNum++, 1, &n, 1, &(diskId), 0, 
			 op->inTupSizeL, TEMPJ);
	    }
	}
      } // while !done with build phase

      if (Flags[TRACE_OP])
	printf("%f(%d) join on node %d has completed build phase HashPages = %d tuples %d\n",
	       clock, identity(), n->id, currentPages, tupleCnt);
      
      if ((bucketNum == 0) && (!overflowFile)) {
	// if this is not an overflow file, send message to controller
	// to indicate that the build phase has completed so the probe
	// phase can start
	  
	cMsg = new CntlMsg;
	cMsg->opCode = TBLSTATUS;
	cMsg->intData = tupleCnt;  // number of tuples built with
	cMsg->queryOp = JOIN;
	cMsg->query = op->query; 
	cMsg->send (op->initiator, inPort);
	
	// wait for the split table message from controller
	cMsg = (CntlMsg*)inPort->recv(); 
	if (cMsg->opCode != NEWSPLIT) {
	  printf("join on node %d got illegal control msg opcode=%d\n",
		 n->id, cMsg->opCode);
	}
	splitTable = cMsg->splitTable; // send made copy of the table
	splitCnt = splitTable->splitCnt;
	cMsg->splitTable = NULL;
	splitTable->tupleSize = op->outTupSize; 
	delete cMsg; 
      }

      if(overflow) {
	writeOutputPage(n, op, buildOverflowFile, overflowTups);
	overflowTups = 0;
	
	//determine ratio of tuples in hash table to tuples on disk
	//  so the appropriate number of probe tuples can be sent to disk
	overflowRatio = (tupsOnDisk * 1.0) / ((tupleCnt + tupsOnDisk) * 1.0) ;
	
	// form probe overflow file
	diskId = random(0, DisksPerNode - 1);
	probeOverflowFile = 
	  new File(CurTempFileNum++, 1, &n, 1, &(diskId), 0, op->inTupSizeR, 
		   TEMPJ);
      }
      
      
      // write left over tups for other buckets (only while processing buck 0)
      if ((bucketNum == 0) && (overflowFile == FALSE))
	for(bNum = 1; bNum < ((JoinPlan*) (op->query))->numBuckets ; 
	    bNum++)  {
	  writeOutputPage(n, op, outputArray[bNum].buildFile,
			  outputArray[bNum].numTuples);
	  outputArray[bNum].numTuples = 0;
	}
      
      // begin the probing phase of the join
	
      tupleCnt = 0;
      tupleCnt2 = 0;
      numSites = 0;
      eofCnt = 0;
      done = FALSE;
      tuplesProduced = 0;
      tupsInHash = 0;
      overflowTups = 0;

      while (!done) {
	inTupCnt = 0;
	tupsInHash = 0;
	if ((bucketNum == 0) && (overflowFile == FALSE)) {
	  // read pages off the network
	  msg = inPort->recv(); 
	  if (msg->type == CONTROL)  {
	    printf("join on node %d got illegal control msg opcode=%d\n",
		   n->id, cMsg->opCode);
	    delete cMsg; 
	    exit(NOTOK);
	  } else {
	    // message is another packet of tuples
	    dMsg = (DataMsg*) msg;
	    
	    if (Flags[TRACE_OP]) {
	      printf("%f(%d) join-probe (node %d) rcvd netpage (%d tuples) overflow %d\n",
		     clock, identity(), n->id, dMsg->numTuples, overflowTups);
//            printf("from source box (%x)\n", dMsg->sourceBox);
          }
	    
	    // calculate amount of CPU time consumed copying 
	    // the tuples from the network page to the hash table
	    if (dMsg->eofFlag == TRUE) {
	      eofCnt++;
	      if (eofCnt == op->inStreamsR) done = TRUE;
	    }
	    
	    // distribute tuples among all the buckets
	    for (tupNum = 0; tupNum < dMsg->numTuples; tupNum++) {
	      outputBucket = findBucket(op);
	      if (outputBucket == 0)
		inTupCnt++;
	      else {
		outputArray[outputBucket].numTuples++;
		
		// if output page full write to disk
		if (outputArray[outputBucket].numTuples > probeTupsPerPage) {
		  writeOutputPage(n, op,  
				  outputArray[outputBucket].probeFile,
				  probeTupsPerPage);
		  outputArray[outputBucket].numTuples -=
		    probeTupsPerPage;
		}
	      }
	    }
	    delete dMsg; 
	  }   
	} else {
	  // page to be read from local file (overflow/extra buckets)
	  readPage(n, op,  probeFileScan, scanRC, inTupCnt);
	  if (scanRC == END_OF_SCAN) {
	    done = TRUE;
	    e = n->bufMgr->FlushFile(currProbeFile->fileNo);
	    if (e != OK) printf("Join:probefile flush returned error %d\n",e);
	    delete probeFileScan;
	    delete currProbeFile;
	  }
	}
	
	// Donovans simulator says that this formula is a kludge
	// as the cost formula should scale with the hash table
	// chain length
	  
	instCnt = inTupCnt * (INST_TREAD + INST_PROBE); 
	n->cpu.use(((float) instCnt) / (Mips * 1000)); // msecs
	  op->query->cnt_CPUInstr += (float)  instCnt;
	
	// if there is an overflow file spool tuples to disk 
	if (overflow) {
	  overflowTups += (int) (inTupCnt * overflowRatio);
	  if (overflowTups > probeTupsPerPage) {
	    writeOutputPage(n, op, probeOverflowFile, probeTupsPerPage);
	    overflowTups -= probeTupsPerPage;
	  }
	  tupsInHash = inTupCnt - (int) (inTupCnt * overflowRatio); 
	}
	else
	  tupsInHash = inTupCnt;
	
	tuplesProduced += tupsInHash * sf;
	
	if (tuplesProduced >= 1.0) {
	  fullTuples = (int) tuplesProduced;
	  	  // charge node for copying tuples to output page
	  instCnt = fullTuples * (INST_SPLIT + op->outTupSize * INST_BCOPY) ;
	  n->cpu.use(((float) instCnt) / (Mips * 1000)); 
          op->query->cnt_CPUInstr += (float) instCnt;

	  // send each output tuple to a (possibly skewed) random node
	  //   and send netpage if page full
	  for(tupNum = fullTuples; tupNum > 0; tupNum--) {
        splitTable->putTuples(n, inPort, FALSE, 1);
      }

	  // update the count of tuples produced so far and 
	  // the tuples produced
	  tupleCnt += fullTuples; 
	  tuplesProduced -= fullTuples;
	}  // if tuplesProduced ...

      } // while (!done) with probe phase
	
      if (overflow) {
	  // write leftover overflow tuples
	  writeOutputPage(n, op, probeOverflowFile, overflowTups);
	  overflowTups = 0;
      }

      if ((bucketNum == 0) && (overflowFile == FALSE))
	// write leftover tuples for other buckets
	for(bNum = 1; bNum < ((JoinPlan*) (op->query))->numBuckets ; 
	    bNum++)  {
	  writeOutputPage(n, op, outputArray[bNum].probeFile,
			  outputArray[bNum].numTuples);
	  outputArray[bNum].numTuples = 0;
	}
      
      if (Flags[TRACE_OP])
	printf("%f(%d) join on node %d has completed probe phase tupCnt = %d\n",
               clock, identity(), n->id, tupleCnt);

    } while (overflow);
    
    if (bucketNum == 0) {
      // release extra pages needed for bucket 0
      e = n->bufMgr->ReleaseReservedPages(op, op->pagesAllocated -
		     ((JoinPlan*) op->query)->bucketKMemory);
      if (e != OK) 
	printf("in Join, bufmgr.ReleaseReservedPages returned error\n");
    }
  } // foreach bucket


  // exhausted all input streams.  flush all output buffers
  // sends Eofs to the subsequent operator
  splitTable->flushNetBufs(n, inPort);

  //release the pages reserved for this join operator
  e = n->bufMgr->ReleaseBufferPages(op);
  if (e != OK) printf("in Join, bufMgr.releasePage returned error\n");
  
  // all finished.  use some cpu time to terminate the operator 
  // cleanly
  n->cpu.use(((float) TERMINATE_JOIN) / (Mips * 1000)); // msecs
  op->query->cnt_CPUInstr += (float)  TERMINATE_JOIN;
  
  if (Flags[TRACE_OP])
    printf("%f(%d) join on node %d finishing time %f\n",
	   clock, identity(), inPort->owningNode->id, clock);
  
  if (((JoinPlan*) op->query)->numBuckets > 1)
    delete [] outputArray;
//    delete [((JoinPlan*) op->query)->numBuckets] outputArray; kbaix
  
  // send opdone message to the scheduler
  cMsg = new CntlMsg;
  cMsg->opCode = OPDONE;
  cMsg->intData = tupleCnt;  // number of tuples written
  cMsg->queryOp = JOIN;
  cMsg->query = op->query;
  cMsg->send (op->initiator, inPort);
  
  delete inPort;
  delete op;
}

//////////////////////////////////////////////////////////////////////////////

// add 'numTuples' to 'file'
void writeOutputPage(Node* n, Op* op, File *file, int numTuples) {
  int e, diskId, pageId;
  
  if (numTuples > 0) {
      diskId = file->storageDisk(0);
      pageId = file->addPage(n,diskId,numTuples);
      
      // Allocate a page in the bufferpool
      e = n->bufMgr->BucketAllocPage(diskId, pageId, file->fileNo);
      if (e != OK) printf("Join:BucketAllocPage returned error %d\n",e);
      
      // Then write into the allocated page
      e = n->bufMgr->BucketWritePage(diskId, pageId, op);
      if (e != OK) printf("Join:BucketWritePage returned error %d\n",e);
  }
}

//////////////////////////////////////////////////////////////////////////////

// read page using 'filescan'
void readPage(Node* n, Op* op, FileScan* fileScan, int& scanRC, int& numTuples){
  int diskId, pageId, e, pagesLeft;
  
  scanRC = fileScan->nextPage(pageId, diskId, numTuples, pagesLeft);
  if (scanRC != END_OF_SCAN) {
    e = n->bufMgr->BucketReadPage(diskId, pageId, PREFETCH, op, pagesLeft);
    if (e != OK) printf("in Select, bufMgr.bucketreadpage returned error\n");
  }
}

//////////////////////////////////////////////////////////////////////////////

// find the bucket to be used to send the tuple. chosen randomly
int findBucket(Op* op) {
  float outputRange;
  int outputBucket;
  
  // hash into range of bucket values. extra range added to send more
  //     tuples to bucket 0
  outputRange = uniform(0,((JoinPlan*) (op->query))->numBuckets +
			((JoinPlan*) (op->query))->weightFactor);
  
  // if value in extra range, send to bucket 0
  if (outputRange >= ((JoinPlan*) (op->query))->numBuckets)
    outputBucket =  0;
  else
    outputBucket = (int) floor((double)outputRange);
  
  return outputBucket;
}
