#include "dbsim.h"
#include "object.h"
#include "node.h"
#include "mailbox.h"
#include "disk.h"
#include "language.h"
#include "db.h"
#include "bufmgr.h"

///////////////////////////////////////////////////////////////////////////////
//
//  Name:        node::node
//
//  Title:       Node class constructor
//
//  Function:    Instantiates node-owned objects and starts node processes.
//               Note that arrays of subobjects whose constructors take 
//               parameters must be initialized by "new", one at a time, so 
//               the constructor parameters can be passed (C++ restriction 
//               on arrays of objects).
//                 
///////////////////////////////////////////////////////////////////////////////

extern void nodeMgr(Node*);
extern char BufMgrType;
extern char DiskType;
extern "C" void exit(int);
extern bool PlotMode;

Node::Node(int nodenum): cpu("nodeCpu", (int) CPUsPerNode) {
    int i;

    id = nodenum;      // remember node ID in state data
    nodeMgrPort = new Mailbox("nodeMgrMbox", this);  // get mailbox for nodemgr
    lddPort = new Mailbox("lldMbox", this);// get mailbox for deadlock detector

    for (i=0;i<DisksPerNode;i++) {
	if (DiskType == 'c') disks[i] = new Disk;
	else {
	    printf("Unknown type of disk %c.\n", DiskType);
	    exit(4);
	}
    }

    cpu.set_servicefunc(rnd_rob); // set cpu sched discipline
    cpu.set_timeslice(5.0);       // 5 msec timeslice for RR

    // Starts the appropriate processes on this node
    nodeMgr(this);    // node manager process
}

///////////////////////////////////////////////////////////////////////////////
// 
// Buffer managers are started after database has been defined because some
// of them may need info about the files in order to properly initialize 
// themselves

void Node::startBufMgr() {
  if      (BufMgrType == '3') bufMgr = new BufMgr(BufSize, this);
  else {
    printf("Unknown type of buffer manager %c.\n", BufMgrType);
    exit(4);
  }
}

///////////////////////////////////////////////////////////////////////////////

Node::~Node() {
    int i;
    for (i=0;i<DisksPerNode;i++) delete disks[i];
    delete bufMgr;
}

///////////////////////////////////////////////////////////////////////////////

void Node::printCPUStats() {
  char plotID[256];

  if (PlotMode) {
      sprintf(plotID, "CPUN%d", id);        // label line for graphgen tool
  } else {
      sprintf(plotID, "");
  }

  printf("%s  %4d  %5.3f  %8.4f  %8.4f  %8.4f  %8.4f  %8d\n",
	 plotID,
	 id,
	 cpu.util(),
	 cpu.tput(),
	 cpu.resp(),
	 cpu.qlen(),
	 cpu.serv(),
	 cpu.completions());
}

///////////////////////////////////////////////////////////////////////////////

void Node::printDiskStats() {
  int i;
  char plotID[256];

  for (i=0; i<DisksPerNode; i++) {
      if (PlotMode) { 
	  sprintf(plotID, "DiskN%dD%d", id, i); // label line for graphgen tool
      } else {
	  sprintf(plotID, "");
      }
    
    printf("%s  %4d %4d ", plotID, id, i);
    disks[i]->printStats();
  }
  if (DisksPerNode > 1) printf("\n");
}


///////////////////////////////////////////////////////////////////////////////

float Node::avgQlen() {
    float sum = 0.0;
    for (int i = 0; i < DisksPerNode; i++) sum += disks[i]->avgQlen();
    return sum / DisksPerNode;
}

