///////////////////////////////////////////////////////////////////////////////
//
//  Name:        nodemgr.c
//
//  Title:       Node manager process.
//
//  Function:    Manages node resources:  right now only operator processes.
//                 
//  Change Activity:
//
///////////////////////////////////////////////////////////////////////////////

#include "dbsim.h"
#include "object.h"
#include "node.h"
#include "msg.h"
#include "mailbox.h"
#include "split.h"
#include "language.h"
#include "db.h"
#include "op.h"
#include "query.h"

extern int CurTempFileNum;

void processSelect(Node*, QueryMsg*);
void processScan(Node*, SelectPlan*, Mailbox*, int);
void processStore(Node*, QueryMsg*);
void processTerm(Node*, QueryMsg*);
void processJoin(Node*, QueryMsg*);
void selectOperator(SelectOp*);
void scanOperator(ScanOp*);
void storeOperator(StoreOp*); 
void terminalOperator(TermOp*); 
void joinOperator(JoinOp*); 
File*  createFile(int, int, int, int);
SplitTable* copySplitTable(SplitTable*);

///////////////////////////////////////////////////////////////////////////////

void nodeMgr(Node* n) 
{
    Msg*	msg;
    QueryMsg*	qmsg;
    Mailbox*	inPort;
  
    create("NodeMgr");    

    if (Flags[TRACE_CONTROL])
      printf("NodeMgr is running\n");
    inPort = n->nodeMgrPort;

    while (SimRunning)
    {
        msg = inPort->recv();     // wait for a message
        switch (msg->type) 
	{               
      	    case QUERYPLAN: 
		qmsg = (QueryMsg*) msg;

		if (Flags[TRACE_QUERY]) {
		  printf("NodeMgr %d, received a query:\n", n->id);
		  qmsg->query->print(0);
		  printf("\n\n");
		}

		// now start a proper process to execute the operation
		switch (qmsg->query->opCode)
		{
		    case (SELECT):  
			processSelect(n, qmsg);
			break;
		    case (STORE):  
			processStore(n, qmsg);
			break;
		    case (TERMOP):  
			processTerm(n, qmsg);
			break;
		    case (JOIN):  
			processJoin(n, qmsg);
		        break;
		    default: 
			printf("unimplemented operation %d\n",
				qmsg->query->opCode);
		}
		delete qmsg;
                break;

	    default:
		printf("NodeMgr %d received an illegal message type\n", n->id);
        }
    }
}

///////////////////////////////////////////////////////////////////////////////

void processSelect(Node* n, QueryMsg* msg)
{
    SelectOp*	selectOp;
    SelectPlan* qry;

    qry = (SelectPlan *) msg->query; 

    selectOp = new SelectOp(n);
    selectOp->query = qry ;
    selectOp->initiator = msg->sourceBox;
    selectOp->outTupSize = qry->outTupSize;
    selectOp->inTupSize = qry->inTupSize;
    selectOp->indexSelectivity = qry->indexSelectivity;
    selectOp->splitTable = copySplitTable(qry->splitTable);
    selectOp->splitTable->tupleSize = qry->outTupSize;
    selectOp->file = qry->file;
    selectOp->index = qry->index;
    selectOp->selFactor = qry->selFactor;

    // start the select operator
    selectOperator(selectOp); 
}

///////////////////////////////////////////////////////////////////////////////

void processScan(Node* n, SelectPlan* sp, Mailbox* inPort)
{
    ScanOp*	scanOp;

    scanOp = new ScanOp(n);
    scanOp->query = sp ;
    scanOp->initiator = inPort;
    scanOp->outTupSize = sp->outTupSize;
    scanOp->inTupSize = sp->inTupSize;
    scanOp->file = sp->file;
    scanOp->index = sp->index;

    // start the select operator
    scanOperator(scanOp); 
}

///////////////////////////////////////////////////////////////////////////////

void processStore(Node* n, QueryMsg* msg)
{
    StoreOp*	storeOp;
    StorePlan*  qry;

    qry = (StorePlan *) msg->query; 
    storeOp = new StoreOp(n);

    storeOp->query = qry;
    storeOp->initiator = msg->sourceBox;
    storeOp->outTupSize = qry->outTupSize;
    storeOp->inTupSize = qry->inTupSize;
    storeOp->inStreams = qry->inStreams;

    //       get file id from the query node plan
    storeOp->file = qry->file;
    storeOp->diskId = qry->diskId;

    // start the storeOperator
    storeOperator(storeOp); 
}

void processTerm(Node* n, QueryMsg* msg)
{
    TermOp*	termOp;
    TermPlan*  qry;

    qry = (TermPlan *) msg->query; 
    termOp = new TermOp(n);

    termOp->query = qry;
    termOp->initiator = msg->sourceBox;
    termOp->inStreams = qry->inStreams;

    // start the terminal operator
    terminalOperator(termOp); 
}

///////////////////////////////////////////////////////////////////////////////

void processJoin(Node* n, QueryMsg* msg)
{
    JoinOp*	joinOp;
    JoinPlan* 	qry;

    qry = (JoinPlan *) msg->query; 
    joinOp = new JoinOp(n);

    joinOp->query = qry;
    joinOp->initiator = msg->sourceBox;
    joinOp->outTupSize = qry->outTupSize;
    joinOp->inTupSizeR = qry->inTupSizeR;
    joinOp->inTupSizeL = qry->inTupSizeL;
    joinOp->inStreamsR = qry->inStreamsR;
    joinOp->inStreamsL = qry->inStreamsL;
    joinOp->selFactor = qry->selFactor;
    joinOp->pagesAllocated = qry->maxPagesNeeded;

    // 15 for padding and alignment ???
    joinOp->tupsPerNetPageOut = NETPKTSIZE/(qry->outTupSize + 15); 

    // start the join operator
    joinOperator(joinOp); 
}

///////////////////////////////////////////////////////////////////////////////

