#include "dbsim.h"
#include "object.h"
#include "msg.h"
#include "mailbox.h"
#include "node.h"
#include "split.h"
#include "language.h"
#include "db.h"
#include "query.h"
#include "sched.h"
#include <math.h>

extern "C" {
  void exit(int);
}

int SelectPlan::numSelectPlans = 0;
int StorePlan::numStorePlans = 0;
int TermPlan::numTermPlans = 0;
int JoinPlan::numJoinPlans = 0;

SelectPlan* SelectPlan::freeSelectPlan = NULL; // no free Sel Plans initially
StorePlan* StorePlan::freeStorePlan = NULL; // no free Store Plans initially
TermPlan* TermPlan::freeTermPlan = NULL; // no free Term Plans initially
JoinPlan* JoinPlan::freeJoinPlan = NULL; // no free Join Plans initially

//////////////////////////////////////////////////////////////////////////////

QueryPlan::QueryPlan() {
    transId = schedType = -1;
    name = NULL;
    jobClass = NULL;
    statsPlan = NULL;
    startTime = clock;
    outTupSize = noActiveDisk = activeCnt = tupleCnt = 0;
    minMemNeeded = 1.0;  minMemUnits = MINREQ_UNITS; 
    maxMemNeeded = 1.0;  maxMemUnits = BUFPOOL_UNITS;
    cnt_bufRef = cnt_bufHit = cnt_bufDirty = cnt_diskRead = 
      cnt_diskCacheHit = cnt_CPUInstr = cnt_msgsSent = cnt_msgsReceived =
      cnt_blockedSends = cnt_blockedRecvs = cnt_bytesSent = 
      cnt_bytesReceived = sendTime = recvTime = estDiskSecs = 
      execTime = 0.0;
    for (int i = 0; i < MAX_NODES; i++) activeSites[i] = FALSE;
};

//////////////////////////////////////////////////////////////////////////////

void QueryPlan::pullStats(QueryPlan* source) 
{
    cnt_bufRef += source->cnt_bufRef;
    cnt_bufHit += source->cnt_bufHit;
    cnt_bufDirty += source->cnt_bufDirty;
    cnt_diskRead += source->cnt_diskRead;
    cnt_diskCacheHit += source->cnt_diskCacheHit;
    cnt_CPUInstr += source->cnt_CPUInstr;
    cnt_msgsSent += source->cnt_msgsSent;     
    cnt_msgsReceived += source->cnt_msgsReceived;
    cnt_blockedSends += source->cnt_blockedSends;
    cnt_blockedRecvs += source->cnt_blockedRecvs;
    cnt_bytesSent += source->cnt_bytesSent;
    cnt_bytesReceived += source->cnt_bytesReceived;
    sendTime += source->sendTime;
    recvTime += source->recvTime;
    estDiskSecs += source->estDiskSecs;
}

//////////////////////////////////////////////////////////////////////////////

void QueryPlan::accumPlanStats() 
{
    statsPlan->bufRefs->record(cnt_bufRef);
    statsPlan->bufHits->record(cnt_bufHit);
    statsPlan->bufDirtys->record(cnt_bufDirty);
    statsPlan->diskReads->record(cnt_diskRead);
    statsPlan->diskCacheHits->record(cnt_diskCacheHit);
    statsPlan->CPUInstr->record(cnt_CPUInstr);
    statsPlan->msgsSent->record(cnt_msgsSent);
    statsPlan->msgsReceived->record(cnt_msgsReceived);
    statsPlan->blockedSends->record(cnt_blockedSends);
    statsPlan->blockedRecvs->record(cnt_blockedRecvs);
    statsPlan->bytesSent->record(cnt_bytesSent);
    statsPlan->bytesReceived->record(cnt_bytesReceived);
    statsPlan->sendTime->record(sendTime / cnt_msgsSent);
    statsPlan->recvTime->record(recvTime / cnt_msgsReceived);
}

//////////////////////////////////////////////////////////////////////////////

void QueryPlan::print(int indent) {
    int i;
    char *opName, *plName;
    if (indent); // make compiler happy

    switch (opCode) {
    case SELECT:  opName = "SELECT"; break;
    case STORE:   opName = "STORE"; break;
    case TERMOP:opName = "TERMINAL"; break;
    case JOIN:    opName = "JOIN"; break;
    default:      opName = "???";
    }

    if (name == NULL) 
      plName = "-";          // unnamed plan
    else
      plName = name;
  
    printf("%s(%s): tran %d, sched %d, out sz %d, #act %d, #tups %d",
	   opName, plName, transId, schedType, outTupSize,
	   activeCnt, tupleCnt);
    NewLine;
    printf(" #outtup %d minMem %f(%d) maxMem %f(%d) statsPlan %x", 
	   numOutTuples, minMemNeeded, minMemUnits, maxMemNeeded, 
	   maxMemUnits, statsPlan);
    NewLine;
    printf("sites = { ");
    for (i = 0; i < MAX_NODES; i++) if (activeSites[i]) printf("%d ", i);
    printf("}");
}

//////////////////////////////////////////////////////////////////////////////

void QueryPlan::printName(int indent) {
    if (indent); // make compiler happy

    switch (opCode) {
    case SELECT:  printf("SELECT"); break;
    case STORE:   printf("STORE"); break;
    case TERMOP:  printf("TERMINAL"); break;
    case JOIN:    printf("JOIN"); break;
    default:      printf("???");
    }
}

//////////////////////////////////////////////////////////////////////////////
//
// find the root of the query tree of this operator is a part
//
QueryPlan* QueryPlan::findRoot()
{
  QueryPlan *qp = this; 
  while (qp->parent != NULL) qp = qp->parent;
  return(qp);
}

//////////////////////////////////////////////////////////////////////////////

SelectPlan::~SelectPlan() {
  if (Flags[TRACE_CONTROL])
    printf("running select plan destructor\n"); 
  if (splitTable != NULL) delete splitTable;
}

//////////////////////////////////////////////////////////////////////////////

void SelectPlan::print(int indent) {
    QueryPlan::print(indent);      // print base class stuff first
    printf(", file %d, sel %f, in sz %d\n", 
	   file->fileNo, selFactor, inTupSize);
}

//////////////////////////////////////////////////////////////////////////////

void SelectPlan::accumStats() {

    // If stats are being maintained for this plan, accumulates them in the
    // appropriate random plan object.

    if (statsPlan != NULL) this->accumPlanStats();
}

//////////////////////////////////////////////////////////////////////////////

void StorePlan::print(int indent) {
    QueryPlan::print(indent);      // print base class stuff first
    printf(", file %d", file->fileNo); 
    printf(", disk %d, #in %d, in sz %d",
	   diskId, inStreams, inTupSize);
    NewSubLine;
    child->print(indent + PRINTINDENT);
}

//////////////////////////////////////////////////////////////////////////////

void StorePlan::accumStats() {

    child->accumStats();           // accum stats into child
    this->pullStats(child);        // pull childs stats into this node

    // If stats are being maintained for this plan, accumulates them in the
    // appropriate random plan object.

    if (statsPlan != NULL) this->accumPlanStats();
}

/////////////////////////////////////////////////////////////////////////////

void TermPlan::print(int indent) {
    QueryPlan::print(indent);      // print base class stuff first
    NewSubLine;
    child->print(indent + PRINTINDENT);
}

//////////////////////////////////////////////////////////////////////////////

void TermPlan::accumStats() {

    child->accumStats();           // accum stats into child
    this->pullStats(child);        // pull childs stats into this node

    // If stats are being maintained for this plan, accumulates them in the
    // appropriate random plan object.

    if (statsPlan != NULL) this->accumPlanStats();
}

//////////////////////////////////////////////////////////////////////////////

JoinPlan::~JoinPlan() {
  if (Flags[TRACE_CONTROL])
    printf("running join plan destructor\n"); 
  if (splitTable != NULL) delete splitTable;
}

//////////////////////////////////////////////////////////////////////////////

void JoinPlan::accumStats() {

    leftChild->accumStats();           // accum stats into left child
    rightChild->accumStats();          // accum stats into right child
    this->pullStats(leftChild);        // pull left child stats into this node
    this->pullStats(rightChild);       // pull right child stats into this node

    // If stats are being maintained for this plan, accumulates them in the
    // appropriate random plan object.

    if (statsPlan != NULL) this->accumPlanStats();
}

//////////////////////////////////////////////////////////////////////////////

void JoinPlan::print(int indent) {
    QueryPlan::print(indent);      // print base class stuff first
    NewLine;
    printf("sel %f, bucks %d, l eofs %d, r eofs %d, l tupsize %d, r tupsize %d, phase %d",
	   selFactor, numBuckets, inStreamsL, inStreamsR, inTupSizeL, 
	   inTupSizeR, phase);
    NewSubLine;
    leftChild->print(indent + PRINTINDENT);
    NewSubLine;
    rightChild->print(indent + PRINTINDENT);
}

//////////////////////////////////////////////////////////////////////////////

//finds the maximum number of pages needed to execute the join in memory
JoinPlan::optMaxPages() {
    int numTuples = Min(leftChild->numOutTuples, rightChild->numOutTuples);
    return(Max((int) ceil(1.0 * numTuples * inTupSizeL * FUDGEFACTOR /
			  (PageSize * noActiveDisk)), 3));
}

//finds the minimum number of pages needed to execute the join in memory
JoinPlan::optMinPages() {
    return((int) ceil(sqrt(optMaxPages())) + 1);
}

//////////////////////////////////////////////////////////////////////////////

