#include "dbsim.h"
#include "object.h"
#include "mailbox.h"
#include "msg.h"
#include "node.h"
#include "bufmgr.h"
#include "language.h"
#include "db.h"
#include "op.h"
#include "query.h"
#include "cost.h"

extern "C" {
    void exit(int);
}

// routine to simulate a scan operation operating on a 
// disk-resident relation

void scanOperator(ScanOp* op)
{
    register int pageId, diskId, numTuples, e, fileId, indexId;
    register Node* n;
    register CntlMsg* cMsg;
    register DataMsg* dMsg;
    TIME	startTime; // time operator starts running
    int         scanType;       // index or file scan
    const int   FILE_SCAN = 1;
    const int   INDEX_SCAN = 2;
    int		prefetchData;	// whether or not to prefetch data pages
    int         scanRC;         // file/index scan return code
    int         pageType;       // index scan page type returned from nextPage
    int         idxLvl;         // index level number returned from nextPage
    int         pagesLeft;      // number of pages left returned from nextPage
    bool        joinParent;     // TRUE if parent of select/scan is a join o
    int         qid;            // bufmgr LRU queue id to place data pages on

    // following four variables are for performance reasons only
    register int tupsPerNetPageOut;
    register IndexScan*	indexScan;	// scan for input file
    register FileScan*	fileScan;	// scan for input file
    register Mailbox*	inPort;		// mbox for process to receive messages

    create("scan");

    n = op->node;
    // construct mail box for process to recieve messages
    inPort = new Mailbox("opMbox", n, op);

    // initializations 
    startTime = clock;

    tupsPerNetPageOut = NETPKTSIZE/(op->outTupSize + 15);

    // Sets flag indicating if parent of select/scan is a join op.  Used for
    // deciding which bufmgr LRU chain to place scanned pages on.
    joinParent = op->query->parent->opCode == JOIN;

    // set up a scan of the input file
    if (op->index == NULL) {     // no index means a file scan
      scanType = FILE_SCAN;
      prefetchData = PREFETCH;
      fileScan = new FileScan(op->file, n);
      pageType = DATA_PAGE;      // always the case for a file scan
      fileId = op->file->fileNo;
      scanRC = fileScan->nextPage(pageId, diskId, numTuples, pagesLeft);
      idxLvl = 0;  // doesn\t apply here				 
    }
    else {
      scanType = INDEX_SCAN;
      if (op->index->indexType == CLUSTERED) {
          prefetchData = PREFETCH;
      } else {
	  prefetchData = NO_PREFETCH;
      }
      // initiate the actual index scan
      indexScan = new IndexScan(op->index, n, 
				((SelectPlan*) (op->query))->indexSelectivity);
      indexId = op->index->indexNo;
      fileId = op->index->data->fileNo;
      scanRC = indexScan->nextPage(pageId, pageType, diskId, numTuples,idxLvl);
    }
    
    if (Flags[TRACE_OP])
    {
      printf("%f(%d) scan started on node=%d, scantype=%d, file=%d, outtupsize=%d\n", clock, identity(),
	     n->id, scanType, fileId, op->outTupSize);
    }
    
    // Note that no CPU instructions are consumed to start/end a scan op, since
    // in a real system, scan processes are started once per disk, and just
    // sit there waiting.  They are not started on demand in response to
    // any query.

    while (scanRC != END_OF_SCAN) // continue until end of file
    {
//      if (Flags[TRACE_OP])
//	  printf("%f(%d) scan node %d, pageId %d, pageType %d, pagedisk %d, numTuples %d\n", clock, identity(),n->id, pageId, pageType, diskId, numTuples);

	// love all the index pages and hate all the data pages
	//   assumes that we do not want to see the same data page twice
        qid = LOVE;
        if (pageType == DATA_PAGE) {
	  if (prefetchData == PREFETCH)
	    qid = HATE;  // hate sequentially read data file pages
	  else
	    qid = LIKE;  // like "randomly" referenced data file pages
	  e = n->bufMgr->ReadPage(diskId, pageId, fileId, qid, prefetchData, 
				  pageType, idxLvl, op, pagesLeft);
	}
	else
	  e = n->bufMgr->ReadPage(diskId, pageId, indexId, LOVE,
				  NO_PREFETCH, pageType, idxLvl, op);
	if (e != OK) printf("in Select, bufMgr.readpage returned error\n");

        if (Flags[TRACE_OP]) {
	    printf("%f(%d) scan on node %d sending netpage %d:%d (%d tuples), qid %d\n",
		   clock, identity(), n->id, diskId, pageId, numTuples, qid); 
	}

        // send page to parent select operator
	dMsg = new DataMsg;  // allocate a free buffer
	dMsg->numTuples = numTuples;
	dMsg->tupleSize = (pageType == INDEX_PAGE) ? 
	  op->index->keySize : op->inTupSize;
	dMsg->pageType = pageType;
	dMsg->eofFlag = FALSE;
        dMsg->cheapMsg = TRUE; // make it a cheap msg to avoid msg copying cost
	dMsg->send(op->initiator, inPort);

        // finished with page, unfix it!
	e = n->bufMgr->UnfixPage(diskId, pageId);
        if (e != OK) printf("in Select, bufMgr.unfixpage returned error\n");

        // get next page to process
        if (scanType == FILE_SCAN) {
            scanRC = fileScan->nextPage(pageId, diskId, numTuples, pagesLeft);
	    idxLvl = 0; // doesn\'t appply here
	} else
            scanRC = 
	      indexScan->nextPage(pageId, pageType, diskId, numTuples, idxLvl);

    } // for all pages
  
    // send the last page of the scan
    dMsg = new DataMsg;  // allocate a free buffer
    dMsg->numTuples = numTuples;
    dMsg->tupleSize = op->inTupSize;
    dMsg->pageType = pageType;
    dMsg->eofFlag = TRUE;
    dMsg->cheapMsg = TRUE;  // make it a cheap msg to avoid msg copying cost
    dMsg->send(op->initiator, inPort);

    // send opdone message to the parent select operator
    cMsg = new CntlMsg;
    cMsg->opCode = OPDONE;
    cMsg->send (op->initiator, inPort);

    if (Flags[TRACE_OP])
      printf("%f(%d) scan on node %d finished\n", 
             clock, identity(), n->id);

    if (scanType == FILE_SCAN) delete fileScan; // delete the file scan
    else delete indexScan; // delete index scan
    delete inPort;
    delete op;	// delete the select operator object

}
