#include "dbsim.h"
#include "object.h"
#include "mailbox.h"
#include "msg.h"
#include "node.h"
#include "split.h"
#include "language.h"
#include "db.h"
#include "op.h"
#include "query.h"
#include "cost.h"

extern "C" {
    void exit(int);
}

void processScan(Node*, SelectPlan*, Mailbox*);

// routine to simulate a selection operation operating on a 
// disk-resident relation

void selectOperator(SelectOp* op)
{
    int         numTuples, fullTuples;
    int         numSites;
    float       tuplesProduced; // running sum of tuples produced
                                // used to model fractional tuples
    Node* 	n;
    Msg         *msg;
    CntlMsg*	cMsg;
    DataMsg*    dMsg;
    int		instCnt;
    int		tupleCnt;  // number of tuples processed by operator
    int		tupleCnt2;  // number of tuples processed by operator
    TIME	startTime; // time operator starts running
    int		tupsPerNetPageOut;

    // following four variables are for performance reasons only
    float	sf;		// local value of operator selectivity factor
    Mailbox*	inPort;		// mailbox for process to receive messages
    int         pageType;
    SplitTable*	splitTable;	// local pointer for performance
    SelectPlan* selPlan;
    bool        eofCnt;

    create("select");

    n = op->node;
    // construct mail box for process to receive messages
    inPort = new Mailbox("opMbox", n, op);

    // initializations 
    tupleCnt = 0;      // total number of tuples processed
    tupleCnt2 = 0;     // total number of tuples processed
    numSites = 0;      // total number of tuples processed
    startTime = clock;
    tupsPerNetPageOut = NETPKTSIZE/(op->outTupSize + 15);
	// 15 for padding and alignment ???
    sf = op->selFactor;
    selPlan = ((SelectPlan*)op->query);

    splitTable = op->splitTable;
    
    if (Flags[TRACE_OP])
    {
      printf("%f(%d) select started on node=%d, sf=%f, outtupsize=%d\n", 
	     clock, identity(), n->id, sf,op->outTupSize);
    }

    // begin by sending the operators private port to the query 
    // scheduler that started the operator
    cMsg = new CntlMsg;
    cMsg->opCode = NEWTASK;
    cMsg->intData = n->id;  // kind of a kludge to get the node id returned
    cMsg->send (op->initiator, inPort);

    // consume some CPU starting the operator
    n->cpu.use(((float) INITIATE_SELECT) / (Mips * 1000)); // milliseconds
    op->query->cnt_CPUInstr += (float) INITIATE_SELECT;

    tuplesProduced = 0.0;

    // start the associated scan process
    processScan(n, selPlan, inPort);

    eofCnt = FALSE;
    while (!eofCnt) // continue until end of file
    {
         // block waiting for the next message
	 msg = inPort->recv(); 
	 if (msg->type == CONTROL)
	 {  
	     printf("select on node %d got illegal control msg opcode=%d\n",
		    n->id, cMsg->opCode);
	     delete cMsg; 
	     exit(NOTOK);
	 }
	 else
	 {
	     // message is another packet of tuples
	     dMsg = (DataMsg*) msg;
	     numTuples = dMsg->numTuples;
	     pageType = dMsg->pageType;

     	     if (Flags[TRACE_OP])
             	  printf("%f(%d) select on node %d received netpage (%d tuples)\n",
			 clock, identity(), n->id, numTuples);

	     // consume some CPU time as appropriate for this page
	     if (pageType == INDEX_PAGE) {
	         instCnt = numTuples * (INST_INDEX_ENTRY_TEST);
		 numTuples = 0; // skip tuple loop
	     }
	     else
                 // data page
		 //
		 // apply cost of applying predicate only if
		 //    a) it is a file scan OR
		 //    b) there is a residual predicate (indicated by 
		 //			      selFactor not being 1)
                 //
		 if ((op->indexSelectivity == 0) || 
		     ((op->indexSelectivity != 0) && (sf != 1.0)))
	             instCnt = numTuples * (INST_TREAD + INST_PRED);
		 else
	             instCnt = numTuples * INST_TREAD;

             n->cpu.use(((float) instCnt) / (Mips * 1000)); //milliseconds
	     op->query->cnt_CPUInstr += (float) instCnt;

	     tuplesProduced += numTuples * sf;

         if (tuplesProduced >= 1.0) {
		 fullTuples = (int) tuplesProduced;

		 // charge node for copying tuples to output page
		 instCnt = fullTuples * (INST_SPLIT + 
					op->outTupSize * INST_BCOPY) ;
		 n->cpu.use(((float) instCnt) / (Mips * 1000)); 
		 op->query->cnt_CPUInstr += (float) instCnt;

         // add another tuple to the split table.
         splitTable->putTuples(n, inPort, FALSE,fullTuples);

	       //update the count of tuples produced so far 
	       //and the tuples produced
	       tupleCnt += fullTuples; 
	       tuplesProduced -= fullTuples;
	    } // if tuplesProduced > 1

	    if (dMsg->eofFlag == TRUE) eofCnt = TRUE;
	    delete dMsg;
         } // if (msg->type )... else 
    }

    // exhausted all input streams.  flush all output buffers
    // sends Eofs to the subsequent operator
    splitTable->flushNetBufs(n, inPort);

    // all finished.  use some cpu time to terminate the operator cleanly
    n->cpu.use(((float) TERMINATE_SELECT) / (Mips * 1000)); //milliseconds
    op->query->cnt_CPUInstr += (float) TERMINATE_SELECT;

    // block waiting for OPDONE message from the scan operator
    msg = inPort->recv(); 
    if (msg->type != CONTROL)
    {  
        printf("expected CONTROL message in getPortIds. got type=%d\n",
	       msg->type);
        exit(NOTOK);
    }
    else 
    {
        cMsg = (CntlMsg*) msg;
        if (cMsg->opCode != OPDONE)
	{
            printf("unexpected CONTROL message. expected OPDONE.got type = %d\n",
		   cMsg->opCode);
	    exit(NOTOK);
	}
	else
	    delete cMsg;
    }

    // now send a done message to the initiator
    cMsg = new CntlMsg;
    cMsg->opCode = OPDONE;
    cMsg->intData = tupleCnt;  // number of tuples written
    cMsg->queryOp = SELECT;
    cMsg->query = op->query;
    cMsg->send (op->initiator, inPort);

    if (Flags[TRACE_OP])
      printf("%f(%d) select on node %d finishedf\n", 
             clock, identity(),n->id);

    delete inPort;  // delete operator's private port
    delete op;	// delete the select operator object
}

