#include <limits.h>
#include "dbsim.h"
#include "object.h"
#include "mailbox.h"
#include "msg.h"
#include "node.h"
#include "query.h"
#include "terminal.h"
#include "language.h"
#include "db.h"
#include "sched.h"
#include "bufmgr.h"
#include <string.h>
#include <math.h>

///////////////////////////////////////////////////////////////////////////////
//
//  Name:        language.c
//
//  Title:       PRPL class implementations
//
//               o ProbDist
//               o RandomReal
//               o RandomFile
//               o RandomPlan
//               o SymTable
// 
//  WARNING!  WARNING!  DANGER WILL ROBINSON!!!!
//               
//               Any new fields which are added to RandomX objects must also
//               be dealt with in the clone methods for each of these objects!
//               The clone methods "replicate" random objects (and in some
//               cases, subobjects) for each terminal instance. "Replicate"
//               is in quotes, because although some elements of these objects
//               must, in fact, be duplicated for each terminal, others must be
//               shared by all terminal instances. 
//
//  Change Activity:
//
//     10/01/91 (KPB) - Initial version.
//
///////////////////////////////////////////////////////////////////////////////

extern "C" {
  void exit(int code);
//  double sqrt(double); kbaix
  extern void qsort(void *, size_t, size_t, int (*)(const void *, const void *));
};

void yyerror(char* s);
void yywhere();
extern Node*    Nodes[];
extern Terminal* Terms[MAX_TERMINALS];
extern int TermIDCounter;
extern File *Relations[MAX_FILES];
extern JobClass *JobClasses[];
extern Index *Indices[MAX_FILES];
extern bool PlanStats;
extern char SearchHeur;
extern int CurTempFileNum;
extern bool PlotMode;

/* The "thingy" type is needed because it is sometimes necessary to deal with
   a void* object which is really a floating point number.  Simply casting a 
   void* to a float is not allowed by the C compiler, since it cannot determine
   whether a run time representation conversion is necessary (although the C++
   compiler doesnt give a hoot).

   The other members are there just to avoid extra ugly type casts, since we 
   have to have this mechanism to deal with floats anyway. */

typedef union thingamabob {
  void* v;
  int   i;
  float f;
  char* s;
  Object* o;
  RandomReal* rr;
  RandomFile* rf;
  RandomPlan* rp;
} thingy;

static char errmsg[80]; // working storage to assemble nice err msgs

///////////////////////////////////////////////////////////////////////////////

// Probability distribution/deterministic sequence constructor

ProbDist::ProbDist(probDist d) {
  dist = d;
  elCount = 0;
  totProb = 0.0;
  firstDistEl = lastDistEl = lastChosenEl = NULL;
}

///////////////////////////////////////////////////////////////////////////////

// Add a name to a probability distribution or sequence

void ProbDist::add(char* name) {
  if (dist == rand_dist) 
    this->add((void*) name, 0.0, 0, name_obj);
  else
    this->add((void*) name, 0.0, 1, name_obj);
}

///////////////////////////////////////////////////////////////////////////////

// Add a number to a probability distribution or sequence

void ProbDist::add(float num) {
  thingy obj;

  obj.f = num;
  if (dist == rand_dist) 
    this->add(obj.v, 0.0, 0, number_obj);
  else
    this->add(obj.v, 0.0, 1, number_obj);
}

///////////////////////////////////////////////////////////////////////////////

// Add a name/prob to a probability distribution list

void ProbDist::add(char* name, float prob) {
  if (dist == rand_dist) 
    this->add((void*) name, prob, 0, name_obj);
  else
    yyerror("random element inside a deterministic sequence list");
}

///////////////////////////////////////////////////////////////////////////////

// Add a number/prob to a probability distribution list

void ProbDist::add(float num, float prob) {
  thingy obj;

  obj.f = num;
  if (dist == rand_dist) 
    this->add(obj.v, prob, 0, number_obj);
  else
    yyerror("random element inside a deterministic sequence list");
}

///////////////////////////////////////////////////////////////////////////////

// Add a name*n to a deterministic sequence

void ProbDist::add(char* name, int repFactor) {
  if (dist == rand_dist) 
    yyerror("deterministic element inside a random sequence distribution");
  else
    this->add((void*) name, 0.0, repFactor, name_obj);
}

///////////////////////////////////////////////////////////////////////////////

// Add a number*n to a deterministic sequence

void ProbDist::add(float num, int repFactor) {
  thingy obj;

  obj.f = num;
  if (dist == rand_dist) 
    yyerror("deterministic element inside a random sequence distribution");
  else
    this->add(obj.v, 0.0, repFactor, number_obj);
}

//////////////////////////////////////////////////////////////////////////////

// add the nodes on which file declustered to probability distribution
void ProbDist::add(File* file) {
  thingy obj;
  int i;

  if (dist == rand_dist) {
      for (i = 0; i < file->numNodes; i++) {
            obj.f = file->storageNode(i)->id;
	    this->add(obj.v, 0.0, 1, number_obj);
      }
  } else { 
      yyerror("random element inside a deterministic sequence");
  }
}

///////////////////////////////////////////////////////////////////////////////

// Internal, generalized, add a name/number to a distribution/sequence

void ProbDist::add(void* obj, float prob, int repFactor, distObjType t) {
  distEl* newEl;

  // Allocs/inits new element

  elCount++;                            // counts elements for prob spreading 
  totProb += prob;                      // accumulates all probs seen so far
  newEl = new distEl;
  newEl->thing = obj;
  newEl->type = t;                      // name_obj or num_obj
  newEl->prob = prob;                   // zero for deterministic sequences
  newEl->cumProb = totProb;             // used for generalized coin flipping 
  newEl->repFactor = repFactor;         // zero for probability distributions
  newEl->repCount = 0;                  // not used for probability dists
  newEl->next = NULL;                   // new elements go at end of list

  // Adds new element to list

  if (firstDistEl == NULL)              // empty list?
    firstDistEl = lastDistEl = lastChosenEl = newEl;
  else {
    lastDistEl->next = newEl;
    lastDistEl = newEl;
  }
}

///////////////////////////////////////////////////////////////////////////////

// Spreads unused probability evenly across all elements in a probability 
// distribution.  Returns TRUE if successful, FALSE otherwise.

#define EPSILON .0000001

bool ProbDist::spread() {
  float portion;                        // equal portion of unused prob for all
  distEl *currEl;

  // Takes care of probabilities that were intended to be 1.0, but are are only
  // some epsilon close to 1.0 because of floating point inaccuracies.

  if (  ((1.0 - totProb) > 0.0) && ((1.0 - totProb) <= EPSILON) ||
        ((1.0 - totProb) < 0.0) && ((totProb - 1.0) <= EPSILON) )
    totProb = 1.0;
  
  // Rejects the distribution if any probabilities were specified and there 
  // is more than 1% unused probability (probably a mistake), or the sum of
  // probabilities is > 1.0

  if (totProb != 0.0) {
    if (totProb > 1.0) {
      yyerror("sum of probabilities is > 1.0");
      return(FALSE);
    }

    if (totProb < .99) {
      yyerror("too much unused probability (> 1%)");
      return(FALSE);
    }
  }

  // If necessary, spreads any unused probability evenly across elements,
  // printing a message telling the user what is happening if there is more
  // than one element in the list.

  portion = (1.0 - totProb) / elCount;  // figure how much each element gets
  if (portion != 0.0) {                 // some unused probability exists
    totProb = 0.0;                      // records cumulative probability
    for (currEl = firstDistEl; currEl != NULL; currEl = currEl->next) {
      currEl->prob += portion;
      totProb += currEl->prob;          // accumulate all prob seen so far
      currEl->cumProb = totProb;        // record cumulative prob in ea element
    }
  }

  // Rounds up the cumulative probability of the last element to insure it 
  // equals 1.0.  Croaks if this is not so.

  lastDistEl->cumProb += .00000005;

  if (((lastDistEl->cumProb > 1) && ((lastDistEl->cumProb - 1.0) > 0.00001))
     || ((lastDistEl->cumProb < 1) && ((1.0 - lastDistEl->cumProb) > 0.00001)))
  {
    yyerror("ProbDist::spread:  totProb != 1.0, some internal error!");
    return(FALSE);
  }

  return(TRUE);
 }

///////////////////////////////////////////////////////////////////////////////

// Verifies that all elements of a distribution/sequence are valid data or
// index files as indicated by caller.  Uses FileOK routine to verify each el.

bool ProbDist::allFiles(fileType ftype){
  thingy lastObj, currObj;
  distEl* currEl;
  bool FileOK(distObjType objType, thingy obj, fileType ftype);
  
  currEl = firstDistEl;
  lastObj.v = (currEl == NULL) ? NULL : currEl->thing;
  lastObj.i++;                          // force it to be different first time
  while (currEl != NULL) {
    currObj.v = currEl->thing;
    if (lastObj.v == currObj.v) continue; // skip if we cked this obj before
      
    if (!FileOK(currEl->type, currObj, ftype))
      return(FALSE);
    
    lastObj.v = currObj.v;
    currEl = currEl->next;
  }

  return(TRUE);  // Theyre all files of the same type if we made it here
}

///////////////////////////////////////////////////////////////////////////////

// Verifies that a single distr/sequence el is a valid data or index file

bool FileOK(distObjType objType, thingy obj, fileType ftype) {
  RandomFile* randFile;
  char *expected, *got;
  
  // Set up for a possible informative error msg

  if (ftype == data_file) {
    expected = "data";
    got = "index";
  } else {
    expected = "index";
    got = "data";
  }

  // Handles number type objects:

  if (objType == number_obj) {
    
    if (notafile((int) obj.f)) {
      sprintf(errmsg, "undefined file:  %d", (int) obj.f);
      yyerror(errmsg);
      return(FALSE);
    }
    
    if (ftype != ConstFiles[(int) obj.f]->type) {
      sprintf(errmsg, "expected all %s files, but %d is a %s file", 
	      expected, (int) obj.f, got);
      yyerror(errmsg);
      return(FALSE);
    }
  
  // Handles name type objects:

  } else {
    randFile = (RandomFile*) FileSyms.lookup(obj.s);
    if (randFile == NULL) {
      sprintf(errmsg, "undefined file:  %s", obj.s);
      yyerror(errmsg);
      return(FALSE);
    }
    
    if (ftype != randFile->type) {
      sprintf(errmsg, "expected all %s files, but %s is a %s file", 
	      expected, obj.s, got);
      yyerror(errmsg);
      return(FALSE);
    }
  }
  
  return(TRUE);
}

///////////////////////////////////////////////////////////////////////////////

// Verifies that all elements of a distribution/sequence are real numbers 

bool ProbDist::allReals(){
  thingy lastObj;
  distEl* currEl;
  
  currEl = firstDistEl;
  lastObj.v = (currEl == NULL) ? NULL : currEl->thing;
  lastObj.i++;                        // force it to be different first time
  while (currEl != NULL) {
    if (lastObj.v == currEl->thing) continue; // skip if cked this obj before
    if (currEl->type == name_obj)     // if type == number_obj, its OK by def
      if (RealSyms.lookup((char*) currEl->thing) == NULL) {
	sprintf(errmsg, "undefined real number:  %s", (char*) currEl->thing);
	yyerror(errmsg);
	return(FALSE);
      }
    lastObj.v = currEl->thing;
    currEl = currEl->next;
  }
  
  return(TRUE);                         // Theyre all reals if we made it here
}

///////////////////////////////////////////////////////////////////////////////

// Verifies that all elements of a distribution/sequence are plans

bool ProbDist::allPlans(){
  thingy lastObj;
  distEl* currEl;
  
  currEl = firstDistEl;
  lastObj.v = (currEl == NULL) ? NULL : currEl->thing;
  lastObj.i++;                        // force it to be different first time
  while (currEl != NULL) {
    if (lastObj.v == currEl->thing) continue; // skip if cked this obj before
    if (currEl->type == number_obj) {
      yyerror("plans cannot be referred to by a number");
      return(FALSE);
    }
    if (PlanSyms.lookup((char*) currEl->thing) == NULL) {
      sprintf(errmsg, "undefined plan:  %s", (char*) currEl->thing);
      yyerror(errmsg);
      return(FALSE);
    }
    lastObj.v = currEl->thing;
    currEl = currEl->next;
  }
  
  return(TRUE);                         // Theyre all plans if we made it here
}

///////////////////////////////////////////////////////////////////////////////

// Chooses an element of a random or deterministic sequence

void* ProbDist::choose(distObjType& objType, bool& reset){
  float probeProb, prevProb;
  distEl *currEl, *chosenEl = NULL;
  
  // Random sequence -- uses generalized coin flipping technique:  a probing
  // probability is used to "index" into the cumulative probabilities of each
  // element.  The larger the difference between an elements cumulative 
  // probability and its predecessor, the larger the chance it has of being
  // chosen.

  if (dist == rand_dist) {
    probeProb = uniform(0.0, totProb);
    prevProb = -1.0;
    for (currEl = firstDistEl; currEl != NULL; currEl = currEl->next) { 
      if (probeProb > prevProb && probeProb <= currEl->cumProb) {
	chosenEl = currEl;               // probe probability falls in current
	break;                           // elements range
      }
      prevProb = currEl->cumProb;       
    }
    
    // If we exited loop thru list without finding an element, we have a bug,
    // since the last elements cumProb must = 1.0 !

    if (chosenEl == NULL) {
      printf("ProbDist::choose (%x) couldnt choose a random seq element!\n",
	     this);
      exit(4);
    }

    objType = chosenEl->type;            // return type of thing (name or num)
    reset = FALSE;                       // meaningless for random sequences
    wrapped = FALSE;
    return(chosenEl->thing);             // return thing itself

  // Deterministic sequence

  } else {
    reset = FALSE;  
    wrapped = FALSE;
    if (lastChosenEl->repCount >= lastChosenEl->repFactor) { //enuf reps?
      lastChosenEl->repCount = 0;        // reset for next visit 
      lastChosenEl = lastChosenEl->next; // advance to next sequence element
      if (lastChosenEl == NULL) {        // end of list?
        reset = TRUE;                    // let caller know we reset
	wrapped = TRUE;
	lastChosenEl = firstDistEl;      // back to first element
      }
    }
    objType = lastChosenEl->type;
    lastChosenEl->repCount++;            // we gave this element out once more
    return(lastChosenEl->thing);			     
  }
}

///////////////////////////////////////////////////////////////////////////////

// Replicates a probability distribution/deterministic sequence 

ProbDist* ProbDist::clone() {
  ProbDist* c;
  distEl *prevEl, *cEl, *pEl;

  // Random sequences have no private state that changes, thus they can be 
  // shared since they are never deleted during the entire simulation run.  
  // Only deterministic sequences need to be cloned, since they each need their
  // own state data that changes after every choose() call.

  if (dist == determ_dist) {
    c = new ProbDist(determ_dist);
    
    prevEl = cEl = NULL;
    for (pEl = firstDistEl; pEl != NULL; pEl = pEl->next) {
      cEl = new distEl;
      cEl->thing = pEl->thing;
      cEl->type = pEl->type;
      cEl->repFactor = pEl->repFactor;
      cEl->repCount = pEl->repCount;
      cEl->next = NULL;   // assume this is the last one   

      if (prevEl == NULL) 
	c->firstDistEl = cEl; // store curr element as first one
      else 
	prevEl->next = cEl;  // store curr element hooked off prev element
	  
      prevEl = cEl;
    }
    c->lastDistEl = cEl;
    c->lastChosenEl = c->firstDistEl;     // "choose" state is NOT cloned
  }

  else 
    c = this;                // random sequences can be shared w/o problems
  
  return(c);
}

///////////////////////////////////////////////////////////////////////////////

// Prints a probability distribution/deterministic sequence for debugging.

void ProbDist::print(int indent){
  int i = indent, prob = 0;             // get rid of "indent not used" warning
  thingy obj;
  distEl* currEl;
  
  // Random sequence

  if (dist == rand_dist) {
    printf("[ ");
    for (currEl = firstDistEl; currEl != NULL; currEl = currEl->next) {
      obj.v = currEl->thing;
      if (currEl->type == name_obj)
	printf("%s/%4.2f(%4.2f) ", obj.s, currEl->prob, currEl->cumProb);
      else 
	printf("%f/%4.2f(%4.2f) ", obj.f, currEl->prob, currEl->cumProb);
    }
    printf("]");

  // Deterministic sequence

  } else {
    printf("< ");
    for (currEl = firstDistEl; currEl != NULL; currEl = currEl->next) {
      obj.v = currEl->thing;
      if (currEl->type == name_obj)
	printf("%s*%d ", obj.s, currEl->repFactor);
      else 
	printf("%f*%d ", obj.f, currEl->repFactor);
    }
    printf(">");
  }
}

///////////////////////////////////////////////////////////////////////////////

// Chooses a specific real number according to the specified distribution of
// the RandomReal number.

float RandomReal::choose(){
  distObjType type;
  thingy obj;
  bool reset;                       // we dont care if sequence is repeated
  
  switch (dist) {
    case const_real:   return(val);
    case uniform_real: return(uniform(uniformLow, uniformHigh));
    case rand_real:  
      return((float) random((int) uniformLow, (int) uniformHigh));
    case expntl_real:  return(expntl(mean));
    case nrm_real:     return(normal(mean, variance));
    case userDist_real: {
      obj.v = userDist->choose(type, reset);
      if (type == number_obj)      // object stored as a real constant
        return(obj.f);
      else {                       // object stored as a name of a real object
        obj.o = RealSyms.lookup(obj.s); // lookup number via symb name
        if (obj.v == NULL) {       // this shouldnt happen!
          printf("RandomReal::choose found undef real symb %s in userDist\n",
		 obj.s);
          exit(4);
        }
        return(obj.rr->choose());  // recursively choose a real
      }
    }
  }
  return 999.999; // kbaix
}

///////////////////////////////////////////////////////////////////////////////

// Replicates a RandomReal object.

RandomReal* RandomReal::clone(){
  RandomReal* c;

  // Only need to clone reals with user distributions.  Others have state data
  // which does not change, and since random real objects are never deleted 
  // during a simulation, they can be shared.

  if (dist == userDist_real) {
    c = new RandomReal;
    c->dist = userDist_real;
    c->userDist = userDist->clone(); // recursively clone the distribution obj
  }
  else 
    c = this;                        // can share all other types of reals

  return(c);
}

///////////////////////////////////////////////////////////////////////////////

// Prints a RandomReal object for debugging.

void RandomReal::print(int indent){
  switch (dist) {
    case const_real: {
      printf("%f", val);
      break;
    }
    case uniform_real: {
      printf("uniform(%f, %f)", uniformLow, uniformHigh);
      break;
    }
    case rand_real: {
      printf("rand(%f, %f)", uniformLow, uniformHigh);
      break;
    }
    case expntl_real: {
      printf("expntl(%f)", mean);
      break;
    }
    case nrm_real: {
      printf("normal(%f,%f)", mean, variance);
      break;
    }
    case userDist_real: {
      userDist->print(indent);
      break;
    }
  }  
}

///////////////////////////////////////////////////////////////////////////////

// Stamps out a specific instance of a RandomSet object (returns an array
// of integers which represent the set).  For some combinations of element
// and cardinality distributions, this may be impossible for a reasonable
// number of attempts.  Prints an error message if this is the case.

int* RandomSet::choose(int& card){
  int *result, i, j, numChooses = 0;
  
  card = (int)(cardinality->choose());
  result = new int[card];

  for (i = 0; i < card; i++) {
    do {

      // Keeps choosing until a unique element value results.  If 5*max 
      // attempts fail to produce a unique value, writes a message and 
      // terminates the simulation.  The distribution for the cardinality is 
      // probably much too diffuse relative the the distribution for the 
      // element values to produce enough unique elements.

      result[i] = (int)(elements->choose()); // choose an element value

      // Check if element was chosen already

      for (j = 0; j < i; j++) 
	if (result[j] == result[i])
	  break;

    } while( (j < i) && (++numChooses < (500 * max)) );
    
    if (numChooses >= (500 * max)) {
      printf("RandomSet::choose couldnt find a unique element in %d tries.\n  Cardinality distribution too diffuse relative to element distribution\n", 
	     500 * max);
      exit(4);
    }
  }

  return(result);
}  

///////////////////////////////////////////////////////////////////////////////

// Prints a RandomSet object for debugging.

void RandomSet::print(int indent) {
  printf("{ ");
  elements->print(indent);
  printf(", ");
  cardinality->print(indent);
  printf(" }");
}

///////////////////////////////////////////////////////////////////////////////

// Replicates a RandomSet object.

RandomSet* RandomSet::clone() {
  RandomSet* c;

  c = new RandomSet(max);
  c->elements = elements->clone();
  c->cardinality = cardinality->clone();
  return(c);
}

///////////////////////////////////////////////////////////////////////////////

// Stamps out a specific instance of a file or index according to the
// distribution of the RandomFile object.  Returns a pointer to the file/index
// and sets the ftype argument to indicate file or index.

void* RandomFile::choose(fileType& ftype){  
  distObjType distType;
  thingy obj;
  bool reset;                        // we dont care if sequence is repeated
  
  switch (dist) {
    case const_file:
    ftype = type;
    if (type == data_file) 
      return(file);
    else
      return(index);

    case random_file:               // recursively choose from uniform dist
    return(ConstFiles[random(randomLow, randomHigh)]->choose(ftype));

    case userDist_file: 
      obj.v = userDist->choose(distType, reset);
      if (distType == number_obj)        // object stored as a file ID
	return(ConstFiles[(int) obj.f]->choose(ftype)); // recursively choose
      else {                        // object stored as a name of a file object
        obj.o = FileSyms.lookup(obj.s); // lookup file via symb name
        if (obj.v == NULL) {       // this shouldnt happen!
          printf("RandomFile::choose found undefined file symb in userDist\n");
          exit(4);
        }
        return(obj.rf->choose(ftype));  // recursively choose a file
      }
  }
  return (void*) 0; // kbaix
}

///////////////////////////////////////////////////////////////////////////////

// Destructor for a RandomFile.  Carefully insures that all appropriate
// supporting data structures are cleaned up for constant files (must be 
// careful cause someone may have dipped below the interface and already
// done complete or partial cleanup).

RandomFile::~RandomFile() {

  // If this is a valid constant file, insures that everything is cleaned up

  if (dist == const_file && fileNo > 0) {
    if (type == index_file) {
      if (Indices[fileNo] != NULL) {
	delete Indices[fileNo];
	Indices[fileNo] = NULL;
      }
    } else {
      if (Relations[fileNo] != NULL && Relations[fileNo]->fileNo == fileNo) {
	delete Relations[fileNo];
	Relations[fileNo] = NULL;
      }
    }

    if (ConstFiles[fileNo] != NULL && ConstFiles[fileNo]->fileNo == fileNo)
      ConstFiles[fileNo] = NULL;
  }
}

///////////////////////////////////////////////////////////////////////////////

// Replicates a RandomFile object.

RandomFile* RandomFile::clone(){
  RandomFile* c;

  // Only need to clone files with user distributions.  Others have state data
  // which does not change, and since random file objects are never deleted 
  // during a simulation, they can be shared.

  if (dist == userDist_file) {
    c = new RandomFile;
    c->dist = userDist_file;
    c->type = type;
    c->userDist = userDist->clone(); // replicate the distribution object
  }
  else 
    c = this;

  return(c);
}
///////////////////////////////////////////////////////////////////////////////

// Prints a RandomFile object for debugging.

void RandomFile::print(int indent){
  switch (dist) {
    case const_file: {
      if (type == data_file) {
	printf("REL %d", fileNo);
//	NewLine;
//	file->print();
      } else {
	printf("INDEX %d", fileNo);
//	NewLine;
//	index->print();
      }
      break;
    }
    case random_file: {
      printf("random(%d, %d)", randomLow, randomHigh);
      break;
    }
    case userDist_file: {
      userDist->print(indent);
      break;
    }
  }  
}

///////////////////////////////////////////////////////////////////////////////

// RandomPlan constructor.  Sets data members to detectably invalid values.

RandomPlan::RandomPlan() {
  name = NULL;
  opCode = schedType = -1;
  sites = NULL;
  sizeSites = 0;
  siteSet = NULL;
  memReq = NULL; 
  statsPlan = NULL;
  completions = 0;
  respTimes = execTimes = memAlloc = bufRefs = bufHits = bufDirtys = 
    diskReads = diskCacheHits = CPUInstr = msgsSent = msgsReceived = 
      blockedSends = blockedRecvs = sendTime = recvTime = bytesSent = 
	bytesReceived = NULL;
  maxMPL = INT_MAX;
  actualMaxMPL = currentMPL = execMPL = readyMPL = 0;
  timeWeightedMPL = prevMPLChange = 0.0;
}

///////////////////////////////////////////////////////////////////////////////

// Stamps out a specific instance of a QueryPlan according to the structure
// and distributions indicated in the RandomPlan object.

QueryPlan* RandomPlan::choose(QueryPlan* parent, JobClass* jc){
  distObjType type;
  thingy obj;
  bool reset;                        // we dont care if sequence is repeated
  QueryPlan* qp;
  
  switch (dist) {
  case const_plan: 
    qp = this->fillin(parent, jc); // flesh out const plan
    break;

  case userDist_plan: 
    obj.v = userDist->choose(type, reset);
    if (type == number_obj) {      // shouldnt happen! Plans are always named
      printf("RandomPlan::choose found number_obj plan in userDist\n");
      exit(4);
    } else {                       // object stored as a name of a plan object
      obj.o = PlanSyms.lookup(obj.s); // lookup plan via symb name
      if (obj.v == NULL) {         // this shouldnt happen!
        printf("RandomPlan::choose found undefined plan symb in userDist\n");
        exit(4);
      }
      qp = obj.rp->choose(parent, jc); // recursively choose plan
    }
    break;
  }

  // If statistics are to be gathered on a plan basis, stores the place to
  // record those stats in the query plan.  If the QueryPlan statsPlan field 
  // is non-NULL, QueryPlan::accumStats() will do the actual recording of 
  // statistics in CSIM tables (located in a RandomPlan object).
  // Otherwise leaves the statsPlan field defaulted to NULL.

  if (statsPlan != NULL)
    qp->statsPlan = statsPlan;
  
  // if node is a store or a termop then save a pointer back to the
  //   original random plan. needed only for the root of the query plan
  //   tree.
  if ((qp->opCode == STORE) || (qp->opCode == TERMOP)) {
      if (((RootPlan*) qp)->randPlan != NULL) {
          // as this choose may be nested, change maxmpl only if it is 
          // less than the maxmpl found so far
          if (((RootPlan*) qp)->randPlan->maxMPL > this->maxMPL)
              ((RootPlan*) qp)->randPlan = this;
      }              
      else {
        ((RootPlan*) qp)->randPlan = this;
      }
  }

  return(qp);

}

///////////////////////////////////////////////////////////////////////////////

// Called by RandomPlan::choose to fillin the details of a QueryPlan object

QueryPlan* RandomPlan::fillin(QueryPlan* parent, JobClass* jc) {
  SelectPlan* sp;
  StorePlan* stp;
  TermPlan* tp;
  JoinPlan* jp;
  QueryPlan* qp;
  int i, cnt = 0;
  void* indexOrDataFile;
  fileType ftype;
  int disks[MAX_DISKSPERNODE];
  Node*	relNodes[MAX_NODES]; 

  // If a random site set was used, stamps out a real set and puts it in the
  // slot for a real one.

  if (siteSet != NULL)
    sites = siteSet->choose(sizeSites);

  switch (opCode) {
  case SELECT: 
    sp = new SelectPlan;
    sp->opCode = opCode;            // SelectPlan constructor should do this!?
    sp->selFactor = sel->choose();

    indexOrDataFile = file->choose(ftype);
    if (ftype == data_file) {
      sp->file = (File*) indexOrDataFile;
      sp->indexSelectivity = 0.0;   // doesnt make sense for a file scan
    } else {
      sp->index = (Index*) indexOrDataFile;
      sp->file = sp->index->data;
      sp->indexSelectivity = isel->choose();

      if (sp->indexSelectivity == 0.0) {
	// special case: index code returns one tuple which should be passed on
        //               by the select
	sp->selFactor = 1.0;
      } else if (sp->indexSelectivity < sp->selFactor) {
	// index selectivity should be higher than the operator selectivity
        //   for it to make sense
	printf("ERROR: index selectivity (%d) should be higher than operator selecitivity (%d)",
	       sp->indexSelectivity, sp->selFactor);
	sp->indexSelectivity = sp->selFactor;
      } else { 
	// Selection predicate for select.c is changed to just the non-index 
	// matching portion of the total operator selectivity, such that
	// operator selectivity = index selectivity * non-index selectivity

	sp->selFactor = sp->selFactor / sp->indexSelectivity;
      }
    }

    sp->inTupSize = sp->file->tupleSize;
    sp->outTupSize = (int)(sp->inTupSize * proj->choose());
    if (sp->outTupSize == 0) sp->outTupSize = 1; // cant be < 1
    sp->parent = parent;

    // If user did not explicitly specify sites, runs the operator only where 
    // source file is declustered.

    if (sites == NULL) {
	sp->noActiveDisk = sp->file->numNodes;
	for (i = 0; i < NumDiskNodes; i++) sp->activeSites[i] = FALSE;
	for (i = 0; i < sp->file->numNodes; i++) 
	  sp->activeSites[sp->file->storageNode(i)->id] = TRUE;
    }

    qp = sp;
    break;
    
  case STORE: 
    stp = new StorePlan;
    stp->opCode = opCode;           // StorePlan constructor should do this!?
    stp->child = lchild->choose(stp, jc);
    stp->inTupSize = stp->child->outTupSize;
    stp->inStreams = stp->child->noActiveDisk;
    stp->outTupSize = stp->inTupSize;
    stp->diskId = (int)(disk->choose());// sets disk number for result rel
    disks[0] = stp->diskId;

    if (Flags[TRACE_QUERY]) printf("store assigned disk %d\n", stp->diskId);
    
    // If the user did not explicitly specify sites, runs op at all nodes, and
    // creates a temporary-store on all nodes as well.

    if (sites == NULL) {
      for (i = 0; i < NumDiskNodes; i++) stp->activeSites[i] = TRUE;
      stp->noActiveDisk = NumDiskNodes;
      stp->file = new File(CurTempFileNum++, NumDiskNodes, Nodes,
			   1, disks, 0, stp->outTupSize, TEMPS);
    } else {

    // If user specified sites for the store, creates a temporary-store file
    // on those sites (activeSites array is set after switch statement)

      for(i = 0; i < sizeSites; i++) relNodes[i] = Nodes[sites[i]];
      stp->file = new File(CurTempFileNum++, sizeSites, relNodes, 1, disks,
			   0, stp->outTupSize, TEMPS);
    }

    qp = stp;
    break;
    
  case TERMOP:
    tp = new TermPlan;
    tp->opCode = opCode;           // TermPlan constructor should do this!?
    tp->child = lchild->choose(tp, jc);
    tp->inStreams = tp->child->noActiveDisk;

    // If user did not explicitly specify sites, runs on all nodes

    if (sites == NULL) {
      for (i = 0; i < NumDiskNodes; i++) tp->activeSites[i] = TRUE;
      tp->activeSites[NumDiskNodes] = FALSE;  // or NumNodes-1, same thing
      tp->noActiveDisk = NumDiskNodes;
    }

    qp = tp;
    break;
    
  case JOIN:
    jp = new JoinPlan;
    jp->leftChild = lchild->choose(jp, jc);
    jp->rightChild = rchild->choose(jp, jc);
    jp->inTupSizeL = jp->leftChild->outTupSize;
    jp->inTupSizeR = jp->rightChild->outTupSize;
    jp->inStreamsL = jp->leftChild->noActiveDisk;
    jp->inStreamsR = jp->rightChild->noActiveDisk;
    jp->selFactor = sel->choose();
    jp->outTupSize = (int) ceil(proj->choose() *
      (jp->leftChild->outTupSize + jp->rightChild->outTupSize)); 
    if (jp->outTupSize == 0) jp->outTupSize = 1; // cant be < 1
    jp->parent = parent;
    jp->phase = BUILD;

    // If user did not explicitly specify sites, then runs the join only at 
    // the nodes being used to process its left input - KLUDGE !!!

    if (sites == NULL) {
	for (i = 0, cnt = 0; i < NumDiskNodes; i++) {
	    if (jp->leftChild->activeSites[i] == TRUE) {
		jp->activeSites[i] = TRUE;
		cnt++;
	    } else
	      jp->activeSites[i] = FALSE;    
	}
	jp->noActiveDisk = cnt;
    }
   
    qp = jp;
    break;
    
  default:
    printf("ERROR:  unknown/unsupported op code in random plan\n");
    return(NULL);
  }

  qp->name = name;                // for debugging (NULL for unnamed plan)
  qp->jobClass = jc;              // for statistics on a job class basis
  qp->statsPlan = statsPlan;      // for statistics on a plan basis(if != NULL)
  qp->schedType = schedType;      // the way to schedule this plan

  // If the user explicitly specified sites to run op at, then use those sites.

  if (sites != NULL) {
    for (i = 0; i < NumDiskNodes; i++) qp->activeSites[i] = FALSE;
    for (i = 0; i < sizeSites; i++) 
      qp->activeSites[sites[i]] = TRUE;
    qp->noActiveDisk = sizeSites;
    if (siteSet != NULL) delete [] sites;
  }

  // If memory requirements were specified, chooses them and copies to query
  // The supremely intelligent scheduler will make sense of them.
 
  if (memReq != NULL) {
    qp->minMemNeeded = memReq->min->choose();
    qp->maxMemNeeded = memReq->max->choose();
    qp->minMemUnits = memReq->minUnits;
    qp->maxMemUnits = memReq->maxUnits;
  }


  return(qp);
}


///////////////////////////////////////////////////////////////////////////////

RandomPlan* RandomPlan::clone(){
  RandomPlan* c;

  c = new RandomPlan;

  // Static data members can simply be copied/shared

  c->dist = dist;
  c->name = name;                // plan name can be shared
  c->schedType = schedType;
  c->sizeSites = sizeSites;
  c->sites = sites;              // site list can be shared
  c->statsPlan = statsPlan;      // all clones must record stats in same place
  c->bufRefs = bufRefs;          // stats tables must be shared
  c->bufHits = bufHits;          // stats tables must be shared
  c->bufDirtys = bufDirtys;      // stats tables must be shared
  c->diskReads = diskReads;      // stats tables must be shared
  c->diskCacheHits = diskCacheHits;  // stats tables must be shared
  c->CPUInstr = CPUInstr;        // stats tables must be shared
  c->memAlloc = memAlloc;        // stats tables must be shared
  c->msgsSent = msgsSent;        // stats tables must be shared
  c->msgsReceived = msgsReceived;        // stats tables must be shared
  c->blockedSends = blockedSends;        // stats tables must be shared
  c->blockedRecvs = blockedRecvs;        // stats tables must be shared
  c->sendTime = sendTime;        // stats tables must be shared
  c->recvTime = recvTime;        // stats tables must be shared
  c->bytesSent = bytesSent;        // stats tables must be shared
  c->bytesReceived = bytesReceived;        // stats tables must be shared

  // Random data members must be cloned, if not null.

  if (dist == userDist_plan) 
    c->userDist = userDist->clone();
  else {
    c->opCode = opCode;
    if (siteSet != NULL) c->siteSet = siteSet->clone();
    if (sel != NULL) c->sel = sel->clone();
    if (isel != NULL) c->isel = isel->clone();
    if (proj != NULL) c->proj = proj->clone();
    if (cost != NULL) c->cost = cost->clone();
    if (memReq != NULL) {
      c->memReq = new MemReq;
      c->memReq->minUnits = memReq->minUnits;
      c->memReq->maxUnits = memReq->maxUnits;
      c->memReq->min = memReq->min->clone();
      c->memReq->max = memReq->max->clone();
    }
    if (file != NULL) c->file = file->clone();
    if (lchild != NULL) c->lchild = lchild->clone();
    if (rchild != NULL) c->rchild = rchild->clone();
  }
    
  return(c);
}
///////////////////////////////////////////////////////////////////////////////

void RandomPlan::print(int indent) {
  switch (dist) {

  case const_plan:
    printf("%s(%x)", name, this);
    
    switch (opCode) {
    case SELECT: 
      printf(", SELECT, sel ");
      sel->print(indent);
      printf(", proj ");
      proj->print(indent);
      printf(", file ");
      file->print(indent);
      break;
      
    case STORE:
      printf(", STORE, disk ");
      disk->print(indent);
      NewSubLine;
      lchild->print(indent+PRINTINDENT);
      break;

    case TERMOP:
      printf(", TERMOP");
      NewSubLine;
      lchild->print(indent+PRINTINDENT);
      break;

    case JOIN:
      printf(", HJOIN, sel ");
      sel->print(indent);
      printf(", proj ");
      proj->print(indent);
      NewSubLine;
      lchild->print(indent+PRINTINDENT);
      NewSubLine;
      rchild->print(indent+PRINTINDENT);
      break;

    default:
      printf("RandomPlan::print:  unknown/unsupported op code in rand plan\n");
    }

    NewLine;
    printf("schedType %d, statsPlan %x", schedType, statsPlan);
    if (memReq != NULL) {
      printf(", minMem ");
      memReq->min->print(indent);
      printf("(%d)", memReq->minUnits);
      printf(", maxMem ");
      memReq->max->print(indent);
      printf("(%d)", memReq->maxUnits);
    }
    if (siteSet != NULL) {
      printf(", siteSet ");
      siteSet->print(indent);
    }

    break;
    
  case userDist_plan:
    userDist->print(indent);
    break;

  default:
    printf("RandomPlan::print:  unknown distribution in random plan\n");
  }  
}

///////////////////////////////////////////////////////////////////////////////
//
//    Name:       RandomPlan::activateStats
// 
//    Function:   Allocates CSIM tables to hold plan related stats
//
//    Change Activity:
//
//      12/91 - Original version.  KPB
//
///////////////////////////////////////////////////////////////////////////////

void RandomPlan::activateStats(RandomPlan* flaggedPlan) {

  // Indicates that plan stats need to be written out at the end of the 
  // simulation.

  PlanStats = TRUE;

  // If this is the plan node which is flagged as the stats gathering point,
  // and it has not been flagged already (under another name), then  allocates 
  // the CSIM tables to store the statistics.

  if (this == flaggedPlan && statsPlan == NULL) {

    max_tables(max_tables(0)+16); // Bumps the CSIM limits for tables

    bufRefs = new table("pln-bufRefs");
    bufHits = new table("pln-bufHits");
    bufDirtys = new table("pln-bufDirtys");
    diskReads = new table("pln-diskReads");
    diskCacheHits = new table("pln-cacheHits");
    CPUInstr = new table("pln-CPUInstr");
    memAlloc = new table("pln-memAlloc");
    msgsSent = new table("pln-msgsSent");
    msgsReceived = new table("pln-msgsRcv");
    blockedSends = new table("pln-blkSnd");
    blockedRecvs = new table("pln-blkRcv");
    sendTime = new table("pln-sendTime");
    recvTime = new table("pln-recvTime");
    bytesSent = new table("pln-bytesSent");
    bytesReceived = new table("pln-bytesRecv");
    execTimes = new table("pln-execTimes"), // root nodes of a query only
    respTimes = new table("pln-respTimes"); // root nodes of a query only

  }

  // Sets the pointer to the RandomPlan object which is flagged as the stats
  // gathering point.  This pointer will be copied to all real QueryPlan 
  // objects stamped from this random plan.  When this pointer is non-NULL in
  // a QueryPlan object, the QueryPlan::accumStats() methods will do the 
  // actual recording of statistics in the CSIM tables which reside in the 
  // flagged RandomPlan object.

  statsPlan = flaggedPlan;
    
}
    
///////////////////////////////////////////////////////////////////////////////

void RandomPlan::writeStats() {
  this->writeStats(name);
}

///////////////////////////////////////////////////////////////////////////////

void RandomPlan::writeStats(char* symbol) {
  char* plotID;

  if (statsPlan == NULL) return;  // no stats for this plan

  plotID = PlotMode ? symbol : "";

  printf("\n  %s:\n", symbol);
  printf("%s                                mean             min              max            std dev\n", plotID);
  printf("%s                          ---------------  ---------------  ---------------  ---------------\n", plotID);
  if (respTimes->cnt() > 0) 
    printf("%s    Response times:       %15.4f  %15.4f  %15.4f  %15.4f\n",
           plotID,
	   respTimes->mean(),
	   respTimes->min(),
	   respTimes->max(),
	   (float) (sqrt((double)respTimes->var()))
	   );
  if (execTimes->cnt() > 0) 
    printf("%s    Execution times:      %15.4f  %15.4f  %15.4f  %15.4f\n",
           plotID,
	   execTimes->mean(),
	   execTimes->min(),
	   execTimes->max(),
	   (float) (sqrt((double)execTimes->var()))
	   );
  printf("%s    Buffer refs:          %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 bufRefs->mean(),
	 bufRefs->min(),
	 bufRefs->max(),
	 (float) (sqrt((double)bufRefs->var()))
	 );
  printf("%s    Buffer hits:          %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 bufHits->mean(),
	 bufHits->min(),
	 bufHits->max(),
	 (float) (sqrt((double)bufHits->var()))
	 );
  printf("%s    Buffer dirties:       %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 bufDirtys->mean(),
	 bufDirtys->min(),
	 bufDirtys->max(),
	 (float) (sqrt((double)bufDirtys->var()))
	 );
  printf("%s    Memory alloc:         %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 memAlloc->mean(),
	 memAlloc->min(),
	 memAlloc->max(),
	 (float) (sqrt((double)memAlloc->var()))
	 );
  printf("%s    CPU inst (thousands): %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 CPUInstr->mean() / 1000.0,
	 CPUInstr->min() / 1000.0,
	 CPUInstr->max() / 1000.0,
	 sqrt(CPUInstr->var()) / 1000.0
	 );
  printf("%s    Disk reads:           %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 diskReads->mean(),
	 diskReads->min(),
	 diskReads->max(),
	 (float) (sqrt((double)diskReads->var()))
	 );
  printf("%s    Disk cache hits:      %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 diskCacheHits->mean(),
	 diskCacheHits->min(),
	 diskCacheHits->max(),
	 (float) (sqrt((double)diskCacheHits->var()))
	 );
  printf("%s    Messages sent         %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 msgsSent->mean(),
	 msgsSent->min(),
	 msgsSent->max(),
	 (float) (sqrt((double)msgsSent->var()))
	 );
  printf("%s    Messages received     %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 msgsReceived->mean(),
	 msgsReceived->min(),
	 msgsReceived->max(),
	 (float) (sqrt((double)msgsReceived->var()))
	 );
  printf("%s    Blocked sends         %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 blockedSends->mean(),
	 blockedSends->min(),
	 blockedSends->max(),
	 (float) (sqrt((double)blockedSends->var()))
	 );
  printf("%s    Blocked receives      %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 blockedRecvs->mean(),
	 blockedRecvs->min(),
	 blockedRecvs->max(),
	 (float) (sqrt((double)blockedRecvs->var()))
	 );
  printf("%s    Time/send             %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 sendTime->mean(),
	 sendTime->min(),
	 sendTime->max(),
	 (float) (sqrt((double)sendTime->var()))
	 );
  printf("%s    Time/receive          %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 recvTime->mean(),
	 recvTime->min(),
	 recvTime->max(),
	 (float) (sqrt((double)recvTime->var()))
	 );
  printf("%s    Tot MBytes sent       %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 bytesSent->mean() / (1024.0 * 1024.0),
	 bytesSent->min() / (1024.0 * 1024.0),
	 bytesSent->max() / (1024.0 * 1024.0),
	 (float) (sqrt((double)bytesSent->var())) / (1024.0 * 1024.0)
	 );
  printf("%s    Tot MBytes received   %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 bytesReceived->mean() / (1024.0 * 1024.0),
	 bytesReceived->min() / (1024.0 * 1024.0),
	 bytesReceived->max() / (1024.0 * 1024.0),
	 (float) (sqrt((double)bytesReceived->var())) / (1024.0 * 1024.0)
	 );
  printf("\n%s    Throughput:           %15.4f queries/sec (%d completions)\n",
         plotID,
	 completions / (simtime() / 1000.0),
	 (int) completions
	 );
  printf("%s    Disk cache avg hit%%:  %15.4f\n",
         plotID,
	 (diskReads->mean() == 0.0) ? 0.0 :
	 diskCacheHits->mean() / diskReads->mean());
  printf("%s    Buffer avg hit%%:      %15.4f\n",
         plotID,
	 (bufRefs->mean() == 0.0) ? 0.0 :
	 bufHits->mean() / bufRefs->mean());
  printf("%s    Avg MPL:              %15.4f\n",
         plotID,
         (timeWeightedMPL / simtime()));
  printf("%s    Max MPL achieved:     %15d\n", 
         plotID,
         actualMaxMPL);
}

///////////////////////////////////////////////////////////////////////////////

JobClass::JobClass() {
  population = -1;
  thinksecs = NULL;
  interArrivalSecs = NULL;
  plans = NULL;
  waitSecs = NULL;
  statWaitSecs = 0.0;
  completions = 0;
  maxMPL = INT_MAX;
  execMPL = readyMPL = currentMPL = actualMaxMPL = 0;

  stopCount = -1;
  maxCompletions = -1;
  mplQueue = new MPLQueue(statWaitSecs);
  memoryWaitTable = new Dqtable(statWaitSecs, "memory wait queue");
  timeWeightedMPL = prevMPLChange = firstMPLEntry = 0.0;
  usedMem = 0;

  allocTables();
}

///////////////////////////////////////////////////////////////////////////////

void JobClass::allocTables() {
  respTimes = new Dtable(statWaitSecs, "jc-respTimes");
  execTimes = new Dtable(statWaitSecs, "jc-execTimes");
  bufRefs = new Dtable(statWaitSecs, "jc-bufRefs");
  bufHits = new Dtable(statWaitSecs, "jc-bufHits");
  bufDirtys = new Dtable(statWaitSecs, "jc-bufDirtys");
  memAlloc = new Dtable(statWaitSecs, "jc-memAlloc");
  buckets = new Dtable(statWaitSecs, "jc-buckets");
  diskReads = new Dtable(statWaitSecs, "jc-diskReads");
  diskCacheHits = new Dtable(statWaitSecs, "jc-cacheHits");
  CPUInstr = new Dtable(statWaitSecs, "jc-CPUInstr");
  diskResp = new Dtable(statWaitSecs, "jc-diskResp");
  diskReadResp = new Dtable(statWaitSecs, "jc-diskRResp");
  diskNonBktReadResp = new Dtable(statWaitSecs, "jc-diskNonBktRResp");
  diskBktReadResp = new Dtable(statWaitSecs, "jc-diskBktRResp");
  diskWriteResp = new Dtable(statWaitSecs, "jc-diskWResp");
  armResp = new Dtable(statWaitSecs, "jc-armResp");
  armReadResp = new Dtable(statWaitSecs, "jc-armRResp");
  armWriteResp = new Dtable(statWaitSecs, "jc-armWResp");
  msgsSent = new Dtable(statWaitSecs, "jc-msgsSent");
  msgsReceived = new Dtable(statWaitSecs, "jc-msgsRcv");
  blockedSends = new Dtable(statWaitSecs, "jc-blkSnd");
  blockedRecvs = new Dtable(statWaitSecs, "jc-blkRcv");
  sendTime = new Dtable(statWaitSecs, "jc-sendTime");
  recvTime = new Dtable(statWaitSecs, "jc-recvTime");
  bytesSent = new Dtable(statWaitSecs, "jc-bytesSent");
  bytesReceived = new Dtable(statWaitSecs, "jc-bytesRecv");
  mplQueue = new MPLQueue(statWaitSecs);
  memoryWaitTable = new Dqtable(statWaitSecs, "memory wait queue");
}

///////////////////////////////////////////////////////////////////////////////

void JobClass::deleteTables() {
  delete mplQueue;
  delete memoryWaitTable;
  delete respTimes;
  delete execTimes;
  delete bufRefs;
  delete bufHits;
  delete bufDirtys;
  delete memAlloc;
  delete buckets;
  delete diskReads;
  delete diskCacheHits;
  delete CPUInstr;
  delete diskResp;
  delete diskReadResp;
  delete diskNonBktReadResp;
  delete diskBktReadResp;
  delete diskWriteResp;
  delete armResp;
  delete armReadResp;
  delete armWriteResp;
  delete msgsSent;
  delete msgsReceived;
  delete blockedSends;
  delete blockedRecvs;
  delete sendTime;
  delete recvTime;
  delete bytesSent;
  delete bytesReceived;
}

///////////////////////////////////////////////////////////////////////////////

void JobClass::start() {

  // For an open job class, just starts a single arrival source process.

  if (interArrivalSecs != NULL) {
    Source(this);
    return;
  }
  
  // Fixes up the stopCount value to equal the class population, if this is
  // a "primary" job class.  Terminals decrement this count when they run out
  // of queries, and if it hits zero, the simulation is stopped.  Non-primary
  // job classes have a stopCount that is always negative (constructor 
  // inits to negative number).
  // population added to stop count to take shift changes into account
  if (stopCount >= 0) // primary job class?
    stopCount += population;

  int lastID = TermIDCounter + population - 1;
  while (TermIDCounter <= lastID) {
      Terms[TermIDCounter] = new Terminal(TermIDCounter++, this);
  }
}


///////////////////////////////////////////////////////////////////////////////
int fcompare(float *num1, float *num2) {
    if ((*num1) < (*num2)) return -1;
    else if ((*num1) == (*num2)) return 0;
    else return 1;
}

///////////////////////////////////////////////////////////////////////////////

void JobClass::writeStats() {
  char* plotID;
  qtable *tempTable;

  plotID = PlotMode ? name : "";

  printf("\n  %s:\n", name);
  printf("%s                                mean             min              max            std dev\n", plotID);
  printf("%s                          ---------------  ---------------  ---------------  ---------------\n", plotID);
  printf("%s    Response times:       %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 respTimes->mean(),
	 respTimes->min(),
	 respTimes->max(),
	 (float) (sqrt((double)respTimes->var()))
	 );
  printf("%s    Execution times:      %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 execTimes->mean(),
	 execTimes->min(),
	 execTimes->max(),
	 (float) (sqrt((double)execTimes->var()))
	 );
  printf("%s    Buffer refs:          %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 bufRefs->mean(),
	 bufRefs->min(),
	 bufRefs->max(),
	 (float) (sqrt((double)bufRefs->var()))
	 );
  printf("%s    Buffer hits:          %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 bufHits->mean(),
	 bufHits->min(),
	 bufHits->max(),
	 (float) (sqrt((double)bufHits->var()))
	 );
  printf("%s    Buffer dirties:       %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 bufDirtys->mean(),
	 bufDirtys->min(),
	 bufDirtys->max(),
	 (float) (sqrt((double)bufDirtys->var()))
	 );
  printf("%s    Memory alloc:         %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 memAlloc->mean(),
	 memAlloc->min(),
	 memAlloc->max(),
	 (float) (sqrt((double)memAlloc->var()))
	 );
  printf("%s    Hash join buckets:    %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 buckets->mean(),
	 buckets->min(),
	 buckets->max(),
	 (float) (sqrt((double)buckets->var()))
	 );
  printf("%s    Disk req resp time:   %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 diskResp->mean(),
	 diskResp->min(),
	 diskResp->max(),
	 (float) (sqrt((double)diskResp->var()))
	 );
  printf("%s    Disk read resp time:  %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 diskReadResp->mean(),
	 diskReadResp->min(),
	 diskReadResp->max(),
	 (float) (sqrt((double)diskReadResp->var()))
	 );
  printf("%s    Disk read resp (!bkt):%15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 diskNonBktReadResp->mean(),
	 diskNonBktReadResp->min(),
	 diskNonBktReadResp->max(),
	 (float) (sqrt((double)diskNonBktReadResp->var()))
	 );
  printf("%s    Disk read resp (bkt) :%15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 diskBktReadResp->mean(),
	 diskBktReadResp->min(),
	 diskBktReadResp->max(),
	 (float) (sqrt((double)diskBktReadResp->var()))
	 );
  printf("%s    Disk write resp time: %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 diskWriteResp->mean(),
	 diskWriteResp->min(),
	 diskWriteResp->max(),
	 (float) (sqrt((double)diskWriteResp->var()))
	 );
  printf("%s    Disk arm resp time:   %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 armResp->mean(),
	 armResp->min(),
	 armResp->max(),
	 (float) (sqrt((double)armResp->var()))
	 );
  printf("%s    Arm resp (reads):     %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 armReadResp->mean(),
	 armReadResp->min(),
	 armReadResp->max(),
	 (float) (sqrt((double)armReadResp->var()))
	 );
  printf("%s    Arm resp (writes):    %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 armWriteResp->mean(),
	 armWriteResp->min(),
	 armWriteResp->max(),
	 (float) (sqrt((double)armWriteResp->var()))
	 );
  printf("%s    Disk reads:           %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 diskReads->mean(),
	 diskReads->min(),
	 diskReads->max(),
	 (float) (sqrt((double)diskReads->var()))
	 );
  printf("%s    Disk cache hits:      %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 diskCacheHits->mean(),
	 diskCacheHits->min(),
	 diskCacheHits->max(),
	 (float) (sqrt((double)diskCacheHits->var()))
	 );
  printf("%s    CPU inst (thousands): %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 CPUInstr->mean() / 1000.0,
	 CPUInstr->min() / 1000.0,
	 CPUInstr->max() / 1000.0,
	 (float) (sqrt((double)CPUInstr->var())) / 1000.0
	 );
  printf("%s    Messages sent         %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 msgsSent->mean(),
	 msgsSent->min(),
	 msgsSent->max(),
	 (float) (sqrt((double)msgsSent->var()))
	 );
  printf("%s    Messages received     %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 msgsReceived->mean(),
	 msgsReceived->min(),
	 msgsReceived->max(),
	 (float) (sqrt((double)msgsReceived->var()))
	 );
  printf("%s    Blocked sends         %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 blockedSends->mean(),
	 blockedSends->min(),
	 blockedSends->max(),
	 (float) (sqrt((double)blockedSends->var()))
	 );
  printf("%s    Blocked receives      %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 blockedRecvs->mean(),
	 blockedRecvs->min(),
	 blockedRecvs->max(),
	 (float) (sqrt((double)blockedRecvs->var()))
	 );
  printf("%s    Time/send             %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 sendTime->mean(),
	 sendTime->min(),
	 sendTime->max(),
	 (float) (sqrt((double)sendTime->var()))
	 );
  printf("%s    Time/receive          %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 recvTime->mean(),
	 recvTime->min(),
	 recvTime->max(),
	 (float) (sqrt((double)recvTime->var()))
	 );
  printf("%s    Tot MBytes sent       %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 bytesSent->mean() / (1024.0 * 1024.0),
	 bytesSent->min() / (1024.0 * 1024.0),
	 bytesSent->max() / (1024.0 * 1024.0),
	 (float) (sqrt((double)bytesSent->var())) / (1024.0 * 1024.0)
	 );
  printf("%s    Tot MBytes received   %15.4f  %15.4f  %15.4f  %15.4f\n",
         plotID,
	 bytesReceived->mean() / (1024.0 * 1024.0),
	 bytesReceived->min() / (1024.0 * 1024.0),
	 bytesReceived->max() / (1024.0 * 1024.0),
	 (float) (sqrt((double)bytesReceived->var())) / (1024.0 * 1024.0)
	 );
  printf("\n%s    Throughput:           %15.4f queries/sec (%d completions)\n",
         plotID,
	 completions / (simtime() / 1000.0),
	 (int) completions
	 );
  printf("%s    Disk cache avg hit%%:  %15.4f\n",
         plotID,
	 (diskReads->mean() == 0.0) ? 0.0 :
	 diskCacheHits->mean() / diskReads->mean());
  printf("%s    Buffer avg hit%%:      %15.4f\n",
         plotID,
	 (bufRefs->mean() == 0.0) ? 0.0 :
	 bufHits->mean() / bufRefs->mean());
  printf("%s    Avg MPL:              %15.4f\n",
         plotID,
         timeWeightedMPL / (clock - firstMPLEntry));
  printf("%s    Max MPL achieved:     %15d\n", 
         plotID,
         actualMaxMPL);

  tempTable = mplQueue->queryQueueStat;
  printf("\n    Wait Queue Statistics        MPL Wait Queue");
  printf("                Memory\n");
  printf("\n%s\tTotal\t\t\t\t%d\t\t\t%d\n%s\tMeanLength\t\t\t%5.3f\t\t\t%5.3f",
	 plotID, tempTable->cnt(), memoryWaitTable->cnt(), 
         plotID, tempTable->qlen(), memoryWaitTable->qlen()); 
  printf("\n%s\tMaxLength\t\t\t%d\t\t\t%d\n",
	 plotID, tempTable->max(), memoryWaitTable->max());
  printf("%s\tAvgWaitingTime(secs)\t\t%5.3f\t\t\t%5.3f\n", 
         plotID, (tempTable->qtime() / 1000.0), 
         (memoryWaitTable->qtime() / 1000.0));
}

///////////////////////////////////////////////////////////////////////////////

void JobClass::print(int indent) {
  printf("pop = %d, max MPL = %d", population, maxMPL);
  if (thinksecs != NULL) {
    printf(", thinksecs = ");
    thinksecs->print(indent);
  }
  if (interArrivalSecs != NULL) {
    printf(", interArrivalSecs = ");
    interArrivalSecs->print(indent);
  }
  printf(", plans = ");
  plans->print(indent);
}

///////////////////////////////////////////////////////////////////////////////

SymTable::SymTable(){
  head = NULL;
}

///////////////////////////////////////////////////////////////////////////////

bool SymTable::add(char* n, Object* v){
  symTabEntry* entry;

  if (this->lookup(n) != NULL) {      // name already used?
    sprintf(errmsg, "second definition of symbol %s ignored", n);
    yyerror(errmsg);
    return(FALSE);
  }
  entry = new symTabEntry;
  entry->name = n;
  entry->thing = v;
  entry->next = head;
  head = entry;
  return(TRUE);
}

///////////////////////////////////////////////////////////////////////////////

Object* SymTable::lookup(char* n){
  symTabEntry* entry = head;

  while (entry != NULL) 
    if (strcmp(entry->name, n) == 0)
      break;  // found it
    else 
      entry = entry->next;      // advance to next symbol table entry
	
  if (entry != NULL)
    return(entry->thing);
  else
    return(NULL);
}

///////////////////////////////////////////////////////////////////////////////

void SymTable::writeStats() {
  symTabEntry* entry = head;

  while (entry != NULL) {
    entry->thing->writeStats(entry->name);
    entry = entry->next;
  }
}

///////////////////////////////////////////////////////////////////////////////

void SymTable::print(int indent){
  symTabEntry* entry = head;

  while (entry != NULL) {
    NewLine;
    printf("%s:  ", entry->name);
    NewSubLine;
    entry->thing->print(indent+PRINTINDENT);
    entry = entry->next;
  }
}
