
///////////////////////////////////////////////////////////////////////////////
//
//  Name:        language.h
//
//  Title:       Parallel Relational Plan Language (PRPL) class definitions
//
//               Defines data objects and types that are shared between 
//               the PRPL parser and other modules (specifically the terminals)
//
//               o ProbDist
//               o RandomReal
//               o RandomSet
//               o RandomFile
//               o RandomPlan
//               o SymTable
//
//  WARNING!  WARNING!  DANGER WILL ROBINSON!!!!
//               
//               Any new fields which are added to RandomX objects must also
//               be dealt with in the clone methods for each of these objects!
//               The clone methods "replicate" random objects (and in some
//               cases, its subobjects) for each terminal instance. "Replicate"
//               is in quotes, because although some elements of these objects
//               must, in fact, be duplicated for each terminal, others must be
//               shared by all terminal instances. 
//
//  Change Activity:
//
//     10/01/91 (KPB) - Initial version.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef languageh
#define languageh 

extern "C" {
  double ceil(double);
};

class QueryQueue;
class MPLQueue;
class PlanDef;
class File;
class Index;
class QueryPlan;

///////////////////////////////////////////////////////////////////////////////

typedef enum probDist {determ_dist, rand_dist};
typedef enum distObjType {name_obj, number_obj}; 
typedef enum fileType {data_file, index_file};

typedef struct distel {   // an element of a distribution or sequence
  void* thing;            // the object itself
  distObjType type;       // type of object 
  float prob;             // probability of this element for a distribution
  float cumProb;          // cumulative probability from first el to this one
  int repFactor;          // repetition factor for a deterministic seq element
  int repCount;           // number of times this element has been emitted
  distel* next;           // next element in list
} distEl;

// A user specified probability distribution([...]) or deterministic sequence
// (<...>)

class ProbDist : public Object {
  private:
    int elCount;          // number of elements in distribution/sequence
    float totProb;        // sum of the probabilities specified till now
    distEl* firstDistEl;  // pointer to a first dist or sequence element
    distEl* lastDistEl;   // pointer to a last dist or sequence element
    distEl* lastChosenEl; // pointer to last chosen sequence element
    bool wrapped;        // this invocation just wrapped back seq start

    // internal, generalized add element to distribution list
    void add(void* thing, float prob, int repFactor, distObjType t);

  public:
    probDist dist;        // random or deterministic sequence?
    ProbDist(probDist d);
    ~ProbDist(){};
    void add(char* name); // adds a name to a dist or a sequence
    void add(float num);  // adds a number to a dist or a sequence
    void add(char* name, float prob); // adds a name/prob to a distribution
    void add(float num, float prob);  // adds a number/prob to a distribution
    void add(char* name, int repFactor); // adds a name*n to a determ seq
    void add(float num, int repFactor);  // adds a num*n to a deterministic seq
    void add(File* file);  // adds file sites to a random sequence
    bool spread();        // Spreads unused probability evenly across elements
    bool justWrapped() { return wrapped; };
    bool atEnd() {
      return ((dist == determ_dist) && (lastChosenEl == lastDistEl) &&
	      (lastChosenEl->repCount == lastChosenEl->repFactor));
    };

    /* Note the following constraint checking functions are meant to be called
       AFTER the ENTIRE distribution list has been recognized.  They cannot
       be done while building the list, since the parser is bottom up, and
       doesn't know the constraints until higher up in the parse tree. */

    bool allFiles(fileType ftype); // TRUE if all objects files of type ftype
    bool allReals();      // TRUE if all objects check out as real numbers
    bool allPlans();      // TRUE if all objects check out as plans

    // choose picks an object, sets t to its type and sets reset to TRUE if
    // a deterministic sequence has wrapped back to the first element.

    void* choose(distObjType& t, bool& reset); 

    ProbDist* clone();   // replicates a probdist object
    void print(int indent);
};

///////////////////////////////////////////////////////////////////////////////

typedef enum realDist 
  {const_real, uniform_real, rand_real, expntl_real, nrm_real, userDist_real};

class RandomReal : public Object { // represents a fixed or random real number
  public:
    realDist dist;           
    float val;            // its value if a constant
    float uniformLow;     // low value if uniform/random distributed
    float uniformHigh;    // high value if uniform/random distributed
    float mean;           // mean if exponential or normally distributed
    float variance;       // variance if normally distributed
    ProbDist* userDist;   // ptr to user specified distribution 

    RandomReal(){};
    ~RandomReal(){};
    float choose();       // returns a randomly selected real num
    void print(int indent);
    RandomReal* clone();  // replicates a random real object
};

///////////////////////////////////////////////////////////////////////////////

class RandomSet : public Object { // represents a random set of integers
  public:
    int max;              // max allowed cardinality and element value
    RandomReal* elements; // distribution to choose elements
    RandomReal* cardinality; // distribution to choose cardinality

    RandomSet(int mx) { max = mx; };
    ~RandomSet(){};
    int* choose(int& card); // returns a set of integers of size card
    void print(int indent);
    RandomSet* clone();   // replicates a random set object
};

///////////////////////////////////////////////////////////////////////////////

typedef enum fileDist {const_file, random_file, userDist_file};

class RandomFile : public Object {  // represents a fixed or random relation
  public:
    fileType type;        // index_file or data_file
    fileDist dist;
    int fileNo;           // index/data file number for constant files
    File* file;           // ptr to file object if a "constant" data/index file
    Index* index;         // ptr to index object if a "constant" index file 
    int randomLow;        // low value if uniformly distributed
    int randomHigh;       // high value if uniformly distributed
    ProbDist* userDist;   // ptr to user specified distribution 

    RandomFile(){};
    ~RandomFile();
    void* choose(fileType& ftype);  // returns a randomly selected file 
    void print(int indent);
    RandomFile* clone();  // replicates a random file object
};

///////////////////////////////////////////////////////////////////////////////
// 
// "Delayed" table -- just waits until some time passes before really recording
// anything in the base class CSIM table.

class Dtable : public table {
  float delay;              // msecs to wait b4 recording anything

public:
  Dtable(float secs, char* name) : table(name) { delay = secs * 1000.0; }
  void record(float x) { if (clock > delay) table::record(x); };
  void forcedRecord(float x) { table::record(x); }; // ignore delay!
};

///////////////////////////////////////////////////////////////////////////////

class MemReq {             // place to store [min, max] memory req construct
  public:
    RandomReal* min;
    int minUnits;			      
    RandomReal* max;
    int maxUnits;			      

    MemReq(){ min = NULL; max = NULL; };
    ~MemReq(){};
};

class JobClass;
typedef enum planDist {const_plan, userDist_plan};

class RandomPlan : public Object { // represents a fixed or random query plan
  private:
    QueryPlan* fillin(QueryPlan* parent, JobClass* jc); // fill in a const plan

  public:
    planDist dist;
    char* name;           // valid if this is the root on rhs of a plan stmt
    RandomPlan* statsPlan;// plan object to record stats in if stats == TRUE 
    int completions;    // used to calculate throughput for this plan
    int schedType;        // technique to schedule these types of plans
    int opCode;           // physical algebra operator
    RandomReal* sel;      // operator selectivity, if applicable
    RandomReal* isel;     // index selectivity, if applicable
    RandomReal* proj;     // % projection, if applicable
    RandomReal* cost;     // per tuple operator cost (# instr), if applicable
    RandomReal* disk;     // disk # for query result file 
    RandomFile* file;     // file, if applicable
    RandomPlan* lchild;   // left child, if applicable (unary ops use lchild) 
    RandomPlan* rchild;   // right child, if applicable
    ProbDist* userDist;   // ptr to user specified distribution 
    MemReq* memReq;       // memory requirements of operator
    table *bufRefs;       // table of buffer references for this plan
    table *bufHits;       // table of buffer hits for this plan
    table *bufDirtys;     // table of buffer pages dirtied for this plan
    table *diskReads;     // table of disk reads by this plan
    table *diskCacheHits; // table of disk cache hits by this plan
    table *CPUInstr;      // table of CPU instructions used by this plan
    table *respTimes;     // tbl of resp times for this plan (root nodes only)
    table *execTimes;     // tbl of resp times for this plan (root nodes only)
    table *memAlloc;      // memory allocation for this plan
    table *msgsSent;      // tot number of msgs sent
    table *msgsReceived;  // tot number of msgs received
    table *blockedSends;  // number of sends blocked because receiver too slow
    table *blockedRecvs;  // number of recvs blocked because sender too slow
    table *sendTime;      // time waiting for n/w latency or blocking
    table *recvTime;      // time waiting for n/w latency or blocking
    table *bytesSent;     // tot msg bytes sent
    table *bytesReceived; // tot msg bytes received
    int   maxMPL;           // maximum MPL allowed for the plan
    int   actualMaxMPL;     // actual max MPL observed in the system
    int   execMPL;          // num of queries actually executing 
    int   readyMPL;         // execMPL + num of queries waiting for memory
    int   currentMPL;       // readyMPL + num of queries in MPL queue
    double timeWeightedMPL; // maintained by scheduler
    double prevMPLChange;   // last time a query admitted/departed

    // Set of sites can to run the op at can be specified as a constant or
    // random set.  Both fields NULL mean take the default site assignment.

    int* sites;
    int sizeSites;				   
    RandomSet* siteSet;

    RandomPlan();
    ~RandomPlan(){};
    QueryPlan* choose(QueryPlan* parent, JobClass* jc); //stamp out a real plan
    void print(int indent);
    RandomPlan* clone();  // replicates a random plan object
    void writeStats();    // write out plan related statistics
    void writeStats(char* symbol);  // write out plan related statistics
    void activateStats(RandomPlan* flaggedPlan); // turn on plan based stats
}; 

typedef enum joinTypes {hash_join, ir_join, irs_join, is_join, iss_join, 
			nl_join, merge_join};
typedef enum scanTypes {index_scan, file_scan};

///////////////////////////////////////////////////////////////////////////////

class JobClass : public Object { // represents a class of terminals
  void deleteTables();
  void allocTables();

  public:
    int id;               // job class ID (index into JobClasses array)
    char* name;           // class name (symbol table entry)
    RandomReal* thinksecs;// think time in seconds (closed class)
    RandomReal* interArrivalSecs; // inter-arrival time in seconds (open class)
    RandomReal* waitSecs; // secs to do nothing before submitting queries
    float statWaitSecs;   // secs to wait before collecting any statistics
    int population;       // number of terminals in current shift (closed cls)
    ProbDist* plans;      // distribution of plans this terminal class submits
    int completions;      // used to calculate throughput for this class
    Dtable *respTimes;    // table of response time stats for this class
    Dtable *execTimes;    // table of exec times (resp time - wait time)
    Dtable *bufRefs;      // table of buffer references for this job class
    Dtable *bufHits;      // table of buffer hits for this job class
    Dtable *bufDirtys;    // table of buffer pages dirtied for this job class
    Dtable *diskReads;    // table of disk reads by this class
    Dtable *diskCacheHits;// table of disk cache hits by this class
    Dtable *CPUInstr;     // table of CPU instructions used by this class
    Dtable *diskResp;     // disk req resp time (including queueing)
    Dtable *armResp;      // disk arm resp time (no queueing - average)
    Dtable *armReadResp;  // disk arm resp time (no queueing - reads only)
    Dtable *armWriteResp; // disk arm resp time (no queueing - writes only)
    Dtable *memAlloc;     // memory allocation for queries of this class
    int    usedMem;       // amount of memory being used by executing queries
    Dtable *buckets;      // number of hjoin buckets for this class
    Dtable *diskReadResp; // disk read resp time (including queueing)
    Dtable *diskNonBktReadResp; // disk read resp time (including queueing) 
                                // for non-bucket reads
    Dtable *diskBktReadResp; // disk read resp time (including queueing) for
                             //  bucket reads
    Dtable *diskWriteResp;// disk write resp time (including queueing)
    Dtable *msgsSent;     // tot number of msgs sent
    Dtable *msgsReceived; // tot number of msgs received
    Dtable *blockedSends; // number of sends blocked because receiver too slow
    Dtable *blockedRecvs; // number of recvs blocked because sender too slow
    Dtable *sendTime;     // time waiting for n/w latency or blocking
    Dtable *recvTime;     // time waiting for n/w latency or blocking
    Dtable *bytesSent;    // tot msg bytes sent
    Dtable *bytesReceived;// tot msg bytes received
    int stopCount;        // cnt of terms that must complete b4 simulation is
                          // terminated.  Ignored if < 0 (non stopping class)
    int maxCompletions;   // used by primary classes to stop simulation 
  
    // Scheduler maintained data...

    int maxMPL;           // maximum MPL allowed for this job class
    int actualMaxMPL;     // actual max MPL observed in the system
    int execMPL;          // num of queries executing in the class
    int readyMPL;         // execMPL + num of queries waiting for memory
    int currentMPL;       // readyMPL + num of queries in MPL queue
    MPLQueue* mplQueue;   // job class MPL waiting queue
    qtable *memoryWaitTable; // table to collect stats on memory waitingQ
    double timeWeightedMPL;  // maintained by scheduler
    double prevMPLChange;    // last time a query admitted/departed
    double firstMPLEntry;    // time of first entry into timeWeightedMPL sum

    JobClass();
    ~JobClass() { /* deleteTables(); CSIM v16 bug?? */}; 
    void updateDelay() { /* deleteTables(); CSIM v16 bug?? */ allocTables(); };
    void start();         // instatiates/starts all instances of a job class
    void writeStats();    // write out job class statistics
    void writeStats(char* symbol) {
      if (symbol);        // make compiler happy
      this->writeStats();
    };
    void print(int indent);
    void classify();     // set class type as db/wsSmall/wsMed/wsLarge
}; 

///////////////////////////////////////////////////////////////////////////////

// Symbol table stuff.  Note that a particular object may have multiple names.

typedef struct symEntry { // a symbol table entry
    char* name;           // name of the symbol
    Object* thing;        // object that symbol refers to
    symEntry* next;       // ptr to next symbol or NULL for end of list
} symTabEntry;
  
class SymTable {          // symbol table entry
  private:
    symTabEntry* head;    // single head, single thead list for now

  public:
    SymTable();
    ~SymTable(){};
    bool add(char* name, Object* value); // TRUE if insert sucessful
    Object* lookup(char* name); // returns an object, given its name
    void print(int indent);
    void writeStats();    // calls a writeStats method on every object in table
};

extern SymTable RealSyms; // real num symbol table
extern SymTable FileSyms; // file symbol table
extern SymTable PlanSyms; // plan symbol table
extern SymTable JobClassSyms; // terminal symbol table
extern RandomFile *ConstFiles[MAX_FILES]; 

#define    notafile(fID)  ((fID) >= MAX_FILES || (fID) < 0 || ConstFiles[(fID)] == NULL || ConstFiles[(fID)]->fileNo != (fID))

#define    notadisk(dID)  ((dID) >= DisksPerNode ||(dID) < 0)

#define    notanode(nID)  ((nID) >= NumNodes || (nID) < 0 || Nodes[(nID)] == NULL || Nodes[(nID)]->id != (nID))

#endif
