// four basic message types
#define QUERY 1
#define QUERYPLAN  2
#define CONTROL  3
#define DATAPKT  4

// message pool buffer types
#define NO_POOL         0
#define GLOBAL_MSG_POOL 1
#define LOCAL_MSG_POOL  2

#define SIZE_CONTROL 40
#define SIZE_DATAPKT 40
#define SIZE_QUERY 40
#define SIZE_QUERYPLAN 40

#define NETPKTSIZE 8192

#define PROTOCOLCOST 1000
#define TUPLEOVERHEAD 2

// control message types
#define OPDONE      1
#define NEWTASK     2
#define TBLSTATUS   3
#define NEWSPLIT    4
#define NOTIFYSCHED 5 
#define HTREQ       6 
#define ABORT       7 
#define COMMIT      8
#define PAGEDONE    9

class QueryPlan;
class SplitTable;
class File;
class Index;
class Mailbox;

//////////////////////////////////////////////////////////////////////////////

class Msg {
  private:
    static int nextSeqNum;      // class wide sequence number counter
  public:
    int type;                   // message type
    int numBytes;               // total number of bytes in msg packet
    Mailbox* destBox;           // destination mailbox
    Mailbox* sourceBox;         // return mailbox of sender
    int id;                     // message identifier (for debugging)
    float timestamp;            // time when the message was sent
    short pool;                 // buffer pool msg was pulled from
    bool cheapMsg;              // models a cheap msg which does not have any
                                //    copying cost
    Msg() {
      id = ++nextSeqNum;        // assigns a seq number to the Msg
      sourceBox = destBox = (Mailbox*) NULL;
      type =  numBytes = -1;
      timestamp = 0.0;
      pool = NO_POOL;
      cheapMsg = FALSE;
    };

    void print(char* traceString);
    void send(Mailbox* dest,  Mailbox* source);
};

//////////////////////////////////////////////////////////////////////////////

class QueryMsg: public Msg {
  private:
    static int numQMsg;
    static QueryMsg* freeMsg;
    QueryMsg    *nextMsg;	// use to link messages on the free list only
  public:
    QueryPlan*  query;          // pointer to the query plan
    File*       file;           // sent along with the local select operator
    Index*      index;          // sent along with the local select operator
    SplitTable* splitTable;     // sent along with the local select operator

    QueryMsg() {
      query = NULL;
      file = NULL;
      index = NULL;
      splitTable = NULL;
      type = QUERYPLAN;
      nextMsg = NULL;
    };
    void print(char* traceString);
    void send(Mailbox* dest,  Mailbox* source);
    inline void* operator new(size_t size);       // new is overloaded
    inline void  operator delete(void* oldMsg);   // delete is overloaded
};

//////////////////////////////////////////////////////////////////////////////

class DataMsg: public Msg {
  private:
    static int numDMsg;
    static class DataMsg* freeMsg;
    DataMsg     *nextMsg;	// use to link messages on the free list only
  public:
    int 	numTuples;
    int 	tupleSize;
    int         pageType;
    bool 	eofFlag;

    DataMsg() {
      numTuples = tupleSize = -1;
      type = DATAPKT;
      pageType = 0;
      eofFlag = FALSE;
      nextMsg = NULL;
    };
    static DataMsg* alloc();    // allocate an empty message
    void static free(DataMsg*);	// release a message buffer
    void print(char* traceString);
    void send(Mailbox* dest,  Mailbox* source);
    inline void* operator new(size_t size);       // new is overloaded
    inline void  operator delete(void* oldMsg);   // delete is overloaded
};

//////////////////////////////////////////////////////////////////////////////

class CntlMsg: public Msg {
  private:
    static int numCMsg;
    static class CntlMsg* freeMsg;
    CntlMsg     *nextMsg;	// use to link messages on the free list only
  public:
    int          opCode;		// one of the control message types
    int          intData;
    SplitTable*  splitTable;    
    QueryPlan*   query;
    int 	 queryOp;       // SELECT, STORE, JOIN
    int		 numBuckets;    // valid iff opCode is NEWSPLIT
    int          myNodeNum;     // used by op to send node number to scheduler
//    File*        fileName;    // used by xmed store to send back file name

    CntlMsg() {
      type = CONTROL;
      opCode = queryOp = -1;
      intData = 0;
      splitTable = NULL;
      nextMsg = NULL;
    };
    void print(char* traceString);
    void send(Mailbox* dest,  Mailbox* source);
    inline void* operator new(size_t size);       // new is overloaded
    inline void  operator delete(void* oldMsg);   // delete is overloaded
};

//////////////////////////////////////////////////////////////////////////////

inline void* QueryMsg::operator new(size_t size) {
    QueryMsg *newBox;

    if (freeMsg == NULL) {
	newBox = (QueryMsg *) new char[size];
	if (Flags[TRACE_LEAK])
	  printf("new QMSg %d\n", ++numQMsg);
    } else {
        newBox = freeMsg;
	freeMsg = freeMsg->nextMsg;
    }
    return newBox;
};

//////////////////////////////////////////////////////////////////////////////

inline void QueryMsg::operator delete(void* oldBox) {
    ((QueryMsg *) oldBox)->nextMsg = freeMsg;
    freeMsg = (QueryMsg *) oldBox;
};

//////////////////////////////////////////////////////////////////////////////

inline void* DataMsg::operator new(size_t size) {
    DataMsg *newBox;

    if (freeMsg == NULL) {
	newBox = (DataMsg *) new char[size];
	if (Flags[TRACE_LEAK])
	  printf("new DMSg %d\n", ++numDMsg);
    } else {
        newBox = freeMsg;
	freeMsg = freeMsg->nextMsg;
    }
    return newBox;
};

//////////////////////////////////////////////////////////////////////////////

inline void DataMsg::operator delete(void* oldBox) {
    ((DataMsg *) oldBox)->nextMsg = freeMsg;
    freeMsg = (DataMsg *) oldBox;
};

//////////////////////////////////////////////////////////////////////////////

inline void* CntlMsg::operator new(size_t size) {
    CntlMsg *newBox;

    if (freeMsg == NULL) {
	newBox = (CntlMsg *) new char[size];
	if (Flags[TRACE_LEAK])
	  printf("new CMSg %d\n", ++numCMsg);
    } else {
        newBox = freeMsg;
	freeMsg = freeMsg->nextMsg;
    }
    return newBox;
};

//////////////////////////////////////////////////////////////////////////////

inline void CntlMsg::operator delete(void* oldBox) {
    ((CntlMsg *) oldBox)->nextMsg = freeMsg;
    freeMsg = (CntlMsg *) oldBox;
};
