///////////////////////////////////////////////////////////////////////////////
//
//  Name:        mailbox.c
//
//  Title:       UW database simulator mailbox class implementations.
//
//               The mbox class is an enhancement to the CSIM mailbox class 
//               which allows us to charge CSIM CPU facilities for message 
//               receive protocol pathlength costs.  It is a generic class and
//               is not intended for external use outside of its subclasses.
//
//               The MailBox and TermMailBox classes are futher refinements
//               which just indicates whether the owner is a terminal or a
//               node (send/receive processing is different for each case).
//
//  Change Activity:
//
//     08/25/91 (KPB) - Initial version.
//
///////////////////////////////////////////////////////////////////////////////

#include "dbsim.h"
#include "object.h"
#include "node.h"
#include "query.h"
#include "op.h"
#include "msg.h"
#include "mailbox.h"

extern facility *Network;
extern storage  *GlobMsgPool[];
extern int LocKBytes;
int Mailbox::numMbox = 0;
Mailbox* Mailbox::freeMailbox = NULL;	// no free Mailboxes initially

///////////////////////////////////////////////////////////////////////////////
//
//  mbox::recv         Issues a CSIM receive on the mailbox, and if the mailbox
//                     owner is a node (versus a terminal), then charges the 
//                     owning nodes CPU for the message protocol pathlength.
//
///////////////////////////////////////////////////////////////////////////////

Msg* mbox::recv() { 
  Msg* m;

  // Waits for an incoming message
  mailbox::receive((int *) &m);
  if (Flags[TRACE_MSG]) 
    	printf("%f(%d) mbox::recv got msg(%d), %d bytes\n", 
	       clock, identity(), m->id, m->numBytes);

  return(m);
}

///////////////////////////////////////////////////////////////////////////////
// 
//  Mailbox::Mailbox   Associates a mailbox with a node.
// 
///////////////////////////////////////////////////////////////////////////////

Mailbox::Mailbox(char* name, Node* n, Op* o): mbox(name) { // kbaix
  owningNode = n;
  owningOp = o;
  type = NODE;
  if (LocKBytes > 0)
    msgPool = new storage ("LocMsgPool", LocKBytes * 1024);
  else
    msgPool = NULL;

  if (Flags[TRACE_MSG]) 
    printf("%f(%d) Mailbox::Mailbox - instantiating Mailbox %x (%s)\n", 
	   clock, identity(), this, name);
}

///////////////////////////////////////////////////////////////////////////////
//
//  Mailbox::recv     Issues a CSIM receive on the mailbox associated with
//                      a node (versus a terminal) and then charges the 
//                     owning nodes CPU for the message protocol pathlength.
//
///////////////////////////////////////////////////////////////////////////////

Msg* Mailbox::recv() { 
  Msg* m;
  float netCost, rcvTime;
  int numInstr;

  // Waits for an incoming message

  if (msg_cnt() == 0) // count blocked receives
    if (owningOp != NULL) owningOp->query->cnt_blockedRecvs++;
  rcvTime = simtime();
  mailbox::receive((int *) &m);
  rcvTime = simtime() - rcvTime;
  if (rcvTime < 0.0) rcvTime = 0.0; // work around CSIM bug
  if (Flags[TRACE_MSG]) 
    	printf("%f(%d) Mailbox::recv got msg(%d), %d bytes\n", 
	       clock, identity(), m->id, m->numBytes);

  // If the message was not short circuited, then simulates an appropriate
  // wire delay for the receiver.  This delay is on the receivers end so
  // that the sender is not blocked, and can overlap processing with its 
  // receivers (pipelining).

  if (m->sourceBox->owningNode != m->destBox->owningNode) {
    if (m->numBytes <= 0) {
      printf("mailbox::receive - < 0 bytes (%d), msg id %d\n", 
              m->numBytes, m->id);
      netCost = 0.55;
    }
    else if (m->numBytes == 0) netCost = 0;    // empty message
    else if (m->numBytes < 65) netCost = 0.60; // milliseconds
    else netCost = DelayPerByte * m->numBytes + MinDelay;

    // if appropriate delay time has not passed make the reciver wait
    if ((clock - m->timestamp) < netCost)
        Network->use(netCost - (clock - m->timestamp)); //consume n/w bandwidth
  }

  // Do protocol and msg buffer deallocation only if not a "cheap" message
  // (i.e. a private intra-node msg where sender and receiver share the same
  // msg buffer -- used by scan and select processes, for example).

  if (!(m->cheapMsg)) {
      // Accounts for protocol costs on the receivers CPU
      
      numInstr = (int) (MinRecvProtocol + RecvProtocolPerByte * m->numBytes);
      owningNode->cpu.use(((float) numInstr) / (Mips * 1000.0)); // msecs
      if (owningOp != NULL) owningOp->query->cnt_CPUInstr += (float) numInstr;
      
      // Deallocates mem from the local or global msg pools, as appropriate.
      
      if (m->pool == NO_POOL) {
	  // do nothing
      } else if (m->pool == GLOBAL_MSG_POOL) {
	  GlobMsgPool[owningNode->id]->deallocate(m->numBytes);
      } else if (m->pool == LOCAL_MSG_POOL) {
	  msgPool->deallocate(m->numBytes);
      }
  }
  m->pool = NO_POOL;  // mark pool as deallocated
      
  // Updates msg receiving stats
  if (owningOp != NULL) {
    owningOp->query->cnt_msgsReceived++;
    owningOp->query->cnt_bytesReceived += m->numBytes;
    owningOp->query->recvTime += rcvTime;
  }

  return(m);
}

///////////////////////////////////////////////////////////////////////////////
// 
//  TermMailbox::TermMailbox   Associates a mailbox with a terminal
// 
///////////////////////////////////////////////////////////////////////////////

TermMailbox::TermMailbox(char* name, Terminal* t): mbox(name) {
  owningTerm = t;
  type = TERMINAL;
  if (Flags[TRACE_MSG]) 
    printf("%f(%d) TermMailbox::TermMailbox - instantiating TermMailbox %x (%s)\n", 
	   clock, identity(), this, name);
}
