///////////////////////////////////////////////////////////////////////////////
//
//  Name:        msg.c
//
//  Function:    Contains the message class and subclass (DataMsg, QueryMsg,
//               CntlMsg) method implementations.
//                 
//  Change Activity:
//
//     08/25/91 (KPB) - Initial version.
//
///////////////////////////////////////////////////////////////////////////////

#include "dbsim.h"
#include "object.h"
#include "node.h"
#include "op.h"
#include "query.h"
#include "msg.h"
#include "mailbox.h"
#include "terminal.h"
#include <math.h>

extern "C" {
  double floor(double x);
  void exit(int);
}

extern storage  *GlobMsgPool[];
extern float TotNWBytesSent, TotNWBytesShort, TotSends, TotBlockedSends;

int DataMsg::numDMsg = 0;
int QueryMsg::numQMsg = 0;
int CntlMsg::numCMsg = 0;


int Msg::nextSeqNum = 0;
QueryMsg* QueryMsg::freeMsg = NULL; // no free query msgs initially
DataMsg* DataMsg::freeMsg = NULL; // no free data msgs initially
CntlMsg* CntlMsg::freeMsg = NULL; // no free control msgs initially

///////////////////////////////////////////////////////////////////////////////
//
//  Msg::send      Superclass send.  If the sender is a terminal, just does
//                 a CSIM send.  If the sender is a node, charges the
//                 sending nodes CPU for the message protocol pathlength.
//                 If the message short circuits the network, also charges 
//                 the sending nodes for the mem to mem copy cost.  The 
//                 receiver is charged for any wire delay in Mailbox::receive.
//
///////////////////////////////////////////////////////////////////////////////

void Msg::send(Mailbox* dest, Mailbox* source) {
    Node* sourceNode;
    int numInstr;
    float sendTime;

    // Copies addessing info to state data

    sourceBox = source;
    destBox = dest;
 
    if (sourceBox->type == TERMINAL) {
	// message is from a terminal; just send it
	
    	destBox->send((int) this);
	if (Flags[TRACE_MSG]) 
	  printf("%f(%d) Msg::send id %d, src %x, dest %x, %d bytes - from terminal\n",
		 clock, identity(), id, sourceBox, destBox, numBytes);
	
    	return;
    } else {
	// message is from a node
	sourceNode = sourceBox->owningNode;
 
  	// Accounts for protocol costs on the senders CPU (in milliseconds)
	numInstr = (int) (MinSendProtocol + SendProtocolPerByte * numBytes);
  	sourceNode->cpu.use(((float) numInstr) / (Mips * 1000.0)); 
        if (sourceBox->owningOp != NULL) 
          sourceBox->owningOp->query->cnt_CPUInstr += (float) numInstr;

        // Updates msg sending stats (exclude msgs to terminals)
        if ((sourceBox->owningOp != NULL) && (destBox->type != TERMINAL)) {
          sourceBox->owningOp->query->cnt_msgsSent++;
          sourceBox->owningOp->query->cnt_bytesSent += numBytes;
	  TotSends++;
	}

	sendTime = simtime(); // save start time of send (excluding CPU time)

	// If msg is not being sent to a terminal, and its not a "cheap"
	// private message (e.g. from scan to select) then allocates memory 
        // from the local or global message pools, as appropriate.
	// If pools are too low, then allocate method will hang up this
	// process until enough message pool bytes have been deallocated 
	// to allow this process to run.

	if ((destBox->type != TERMINAL) && (!cheapMsg)) {
	    // If there are no message pools (not modelling send blocking)...
	    if ((GlobKBytes == 0) && (LocKBytes == 0)) { 
		// do nothing
		
		// If there is only a global pool, and no local pool...
	    } else if ((GlobKBytes > 0) && (LocKBytes == 0)) { 
		
		printf("BUG:  LocKBytes is zero!  Deadlock will occur\n");
		exit(4);  // sim.c should have prevented this!
		
		// If there is only a local pool, and no global pool...
	    } else if ((LocKBytes > 0) && (GlobKBytes == 0)) {
		if (destBox->msgPool->avail() < numBytes) { // gonna block?
		    TotBlockedSends++;
		    sourceBox->owningOp->query->cnt_blockedSends++;
		}
		destBox->msgPool->allocate(numBytes);
		pool = LOCAL_MSG_POOL;
		
		// If there is both a local and a global pool...
	    } else if ((LocKBytes > 0) && (GlobKBytes > 0)) {

		// Allocate at local pool only if available or if global 
		// pool will block anyway, else from global.  VERY IMPORTANT
		// to wait behind local pool if both local and global are busy,
		// otherwise deadlock can occur because global pool is a 
		// shared resource and can be overrun by a fast process.

		if ((destBox->msgPool->avail() >= numBytes) ||
		    (GlobMsgPool[destBox->owningNode->id]->avail() 
		     < numBytes)) {
		    if (destBox->msgPool->avail() < numBytes) { // gonna block?
			TotBlockedSends++;
			sourceBox->owningOp->query->cnt_blockedSends++;
		    }
		    destBox->msgPool->allocate(numBytes);
		    pool = LOCAL_MSG_POOL;
		} else {
		    // gonna block?
		    if (GlobMsgPool[destBox->owningNode->id]->avail() < numBytes) {
			TotBlockedSends++;
			sourceBox->owningOp->query->cnt_blockedSends++;
		    }
		    GlobMsgPool[destBox->owningNode->id]->allocate(numBytes);
		    pool = GLOBAL_MSG_POOL;
		}
	    }
	}

  	// If packet shorts circuits the network, simulates a mem2mem copy cost
  	if (sourceNode == destBox->owningNode) {

	  // For short circ, send time is just msg pool queueing time
	  sendTime = simtime() - sendTime; 
	  if (sendTime < 0.0) sendTime = 0.0; // work around CSIM bug!
	  if (sourceBox->owningOp != NULL) 
	    sourceBox->owningOp->query->sendTime += sendTime;
	  
	  if (numBytes <= 0) {
	    printf("Msg::send - bad msg size (%d)\n", numBytes);
	    numInstr = 800;
	  }
	  else
	    // charge cost of copying page only if not a cheap message
	    if (!cheapMsg) {
	        numInstr = (int)(MinShortInstr + ShortInstrPerByte * numBytes);
	        sourceNode->cpu.use(((float) numInstr) / (Mips * 1000.0));
	        if (sourceBox->owningOp != NULL) 
		  sourceBox->owningOp->query->cnt_CPUInstr += (float) numInstr;
	    }

	    if (Flags[TRACE_MSG]) 
	      printf("%f(%d) Msg::send - msg(%d) short circuited, %d bytes, %d instr\n",
		     clock, identity(), id, numBytes, numInstr);
	    
	    if (destBox->type != TERMINAL) TotNWBytesShort += (float) numBytes;

  	} else {
	  // Otherwise, the packet will actually need to use the network.
          // In this case, the receiver must experience the wire delay, since
          // otherwise the message send operation will be synchronous and
          // may prevent pipelining between a sending and receiving process.
          // The wire delay code is in Mailbox::receive.
	    
          // timestamp the message so the reciever can find out
          // if it needs to wait before processing the message

	  timestamp = clock;

	  sendTime = simtime() - sendTime; 
	  if (sendTime < 0.0) sendTime = 0.0; // work around CSIM bug!
	  if (sourceBox->owningOp != NULL) 
	    sourceBox->owningOp->query->sendTime += sendTime;

	  if (Flags[TRACE_MSG]) 
	  printf("%f(%d) Msg::send - msg(%d) via network, %d bytes, %d instr\n",
		 clock, identity(), id, numBytes, numInstr);
	    
	  if (destBox->type != TERMINAL)
	    TotNWBytesSent += (float) numBytes;
        }

        // In either case (short circuit or not), we must pass the message 
        // to the appropriate CSIM process to pick up.
 
        destBox->send((int) this);    // actually send msg to csim process
    }
}

///////////////////////////////////////////////////////////////////////////////
//
//  Msg::print     Superclass debug print.  Prints the common message data.
//
///////////////////////////////////////////////////////////////////////////////

void Msg::print(char* traceString) {

  if (Flags[TRACE_MSG]) {
    printf("%s msg %d, type = %d, %d bytes:\n", traceString, 
	   id, type, numBytes);

    if ((int) sourceBox != NULL) 
    {    // if message is addressed
        if (sourceBox->type == TERMINAL)       // from a terminal
	{
	    printf("source is some terminal\n");
/*
	    printf("source terminal = %d\n", 
		((TermMailBox *)sourceBox)->owningTerm->id);
*/
	}
        else  // from a node
	    printf("source node = %d, ", sourceBox->owningNode->id);

        if (destBox->type == TERMINAL)         // to a terminal
	{
	    printf("dest is some terminal\n");
/*
	    printf("dest terminal = %d\n", 
		((TermMailBox*)destBox)->owningTerm->id);
*/
	}
        else  // to a node
		printf("dest node = %d,\n", destBox->owningNode->id);
    }
    else printf("    -- unaddressed --\n");
  };
};

///////////////////////////////////////////////////////////////////////////////
//
//  QueryMsg::send      Calculates message size and calls superclass send.
//
///////////////////////////////////////////////////////////////////////////////

void QueryMsg::send(Mailbox* dest, Mailbox* source) {
  numBytes = SIZE_QUERY;
  Msg::send(dest, source);
}

///////////////////////////////////////////////////////////////////////////////
//
//  QueryMsg::print     Debug print.  Prints QueryMsg specific data, then 
//                      calls superclass print.
//
///////////////////////////////////////////////////////////////////////////////

void QueryMsg::print(char* traceString) {
  if (Flags[TRACE_MSG]) {
    Msg::print(traceString);
    //query.print(); ??
  };
}

///////////////////////////////////////////////////////////////////////////////
//
//  CntlMsg::send       Calculates message size and calls superclass send.
//
///////////////////////////////////////////////////////////////////////////////

void CntlMsg::send(Mailbox* dest, Mailbox* source) {
  numBytes = SIZE_CONTROL;
  Msg::send(dest, source);
}

///////////////////////////////////////////////////////////////////////////////
//
//  CntlMsg::print      Debug print.  Prints CntlMsg specific data, then 
//                      calls superclass print.
//
///////////////////////////////////////////////////////////////////////////////

void CntlMsg::print(char* traceString) {
  if (Flags[TRACE_MSG]) {
    Msg::print(traceString);
    printf("    cntl type = %d,  int data = %d, q op = %d\n",
	   opCode, intData, queryOp);
    //splitTable.print();
  };
}

///////////////////////////////////////////////////////////////////////////////
//
//  DataMsg::print       Debug print.  Prints DataMsg specific data, then 
//                       calls superclass print.
//
///////////////////////////////////////////////////////////////////////////////

void DataMsg::print(char* traceString) {
  if (Flags[TRACE_MSG]) {
    Msg::print(traceString);
    printf("    num tuples = %d, tuple size = %d, eofFlag = %d\n", 
	   numTuples, tupleSize, eofFlag);
  };
}

///////////////////////////////////////////////////////////////////////////////
//
//  DataMsg::send        Calculates message size and calls superclass send.
//
///////////////////////////////////////////////////////////////////////////////

void DataMsg::send(Mailbox* dest, Mailbox* source) {
  numBytes = SIZE_DATAPKT + (numTuples * (tupleSize + TUPLEOVERHEAD));
  if ((numBytes > 20000) || (numBytes < 0)) {
      printf("datamsg.send being asked to set a message of size %d bytes %d numTuples %d tupleSize\n",
	      numBytes, numTuples, tupleSize);
      printf("source node %d dest node %d\n", source->owningNode->id, 
              dest->owningNode->id);
      printf("source Query\n");
      source->owningOp->query->print(5);
      printf("dest Query\n");
      dest->owningOp->query->print(5);
      exit(NOTOK);
  }
  Msg::send(dest, source);
}
