// class definitions for Select, Store, and Join operators

class SplitTable;
class Node;
class Mailbox;
class QueryPlan;
class File;
class Index;

class Op {
  public:
    Node*	node;           // node this operator is running on
    Mailbox*	initiator;	// initiator's mail box
    int		outTupSize;	// in bytes
    int		inTupSize;	// in bytes
    QueryPlan*  query;          // pointer to node in operator tree

    Op(Node*);
    ~Op(){};
};

class SelectOp: public Op {
  private:
    static int numSelectOps;
    static class SelectOp* freeSelectOp;
    SelectOp *nextOp;
  public:
    File*	file;		   // input file descriptor
    Index*      index;             // if != NULL, index to use for scan of file
    float	selFactor;	   // selectivity factor of operation
    float	indexSelectivity;  // index selectivity (used only for iscan)
    int         diskId;            // for parallel selects, disk this op is on
    SplitTable*	splitTable;	   // split table 

    SelectOp(Node*);
    ~SelectOp();
    inline void* operator new(size_t size);  // new is overloaded
    inline void operator delete(void* oldPlan);  // delete is overloaded too   
};
    
class ScanOp: public Op {
  private:
    static int numScanOps;
    static class ScanOp* freeScanOp;
    ScanOp *nextOp;
  public:
    File*	file;		   // input file descriptor
    Index*      index;             // if != NULL, index to use for scan of file
    int         diskId;            // for parallel selects, disk this op is on

    ScanOp(Node*);
    ~ScanOp(){};
    inline void* operator new(size_t size);  // new is overloaded
    inline void operator delete(void* oldPlan);  // delete is overloaded too   
};
    
class StoreOp: public Op {
  private:
    static int numStoreOps;
    static class StoreOp* freeStoreOp;
    StoreOp *nextOp;
  public:
    File*	file;		// output file descriptor
    int		diskId;
    int		inStreams;	// number of input streams from child op

    StoreOp(Node*);
    inline void* operator new(size_t size);  // new is overloaded
    inline void operator delete(void* oldPlan);  // delete is overloaded too   
};

class TermOp: public Op {
  private:
    static int numTermOps;
    static class TermOp* freeTermOp;
    TermOp *nextOp;
  public:
    int		inStreams;	// number of input streams from child op
    TermOp(Node* n) : Op(n) { inStreams = 0; };
    inline void* operator new(size_t size);  // new is overloaded
    inline void operator delete(void* oldPlan);  // delete is overloaded too   
};

class JoinOp: public Op {
  private:
    static int numJoinOps;
    static class JoinOp* freeJoinOp;
    JoinOp *nextOp;
  public:
    float	selFactor;	// selectivity factor of operation
    int		inTupSizeL;
    int		inTupSizeR;
    int		tupsPerNetPageOut;
    int		inStreamsL;	// number of input streams from left child
    int		inStreamsR;	// number of input streams from right child
    int         pagesAllocated; // total number of pages allocated to the join

    JoinOp(Node*);
    ~JoinOp();
    inline void* operator new(size_t size);  // new is overloaded
    inline void operator delete(void* oldPlan);  // delete is overloaded too   
};

// structure to store information about buckets - used only by joins
typedef struct {
    int   numTuples;  // number of tuples on the page
    File  *probeFile, *buildFile;
} BucketRec;

//////////////////////////////////////////////////////////////////////////////

inline void* SelectOp::operator new(size_t size)
{
    SelectOp *newPlan;

    if (freeSelectOp == NULL) {
	// allocate a new join plan
	newPlan = (SelectOp *) new char[size];
	if (Flags[TRACE_LEAK])
	 printf("new SelectOp %d\n", ++numSelectOps);
    } else {
	// reuse an existing join plan
        newPlan = freeSelectOp;
	freeSelectOp = freeSelectOp->nextOp;
    }
    return newPlan;
};

//////////////////////////////////////////////////////////////////////////////
  
inline void SelectOp::operator delete(void* oldPlan)
{
    // put the join plan on the free list
    ((SelectOp *) oldPlan)->nextOp = freeSelectOp;
    freeSelectOp = (SelectOp *) oldPlan;
};



//////////////////////////////////////////////////////////////////////////////

inline void* ScanOp::operator new(size_t size)
{
    ScanOp *newPlan;

    if (freeScanOp == NULL) {
	// allocate a new join plan
	newPlan = (ScanOp *) new char[size];
	if (Flags[TRACE_LEAK])
	  printf("new ScanOp %d\n", ++numScanOps);
    } else {
	// reuse an existing join plan
        newPlan = freeScanOp;
	freeScanOp = freeScanOp->nextOp;
    }
    return newPlan;
};

//////////////////////////////////////////////////////////////////////////////
  
inline void ScanOp::operator delete(void* oldPlan)
{
    // put the join plan on the free list
    ((ScanOp *) oldPlan)->nextOp = freeScanOp;
    freeScanOp = (ScanOp *) oldPlan;
};

//////////////////////////////////////////////////////////////////////////////

inline void* StoreOp::operator new(size_t size)
{
    StoreOp *newPlan;

    if (freeStoreOp == NULL) {
	// allocate a new join plan
	newPlan = (StoreOp *) new char[size];
	if (Flags[TRACE_LEAK])
	  printf("new StoreOp %d\n", ++numStoreOps);
    } else {
	// reuse an existing join plan
        newPlan = freeStoreOp;
	freeStoreOp = freeStoreOp->nextOp;
    }
    return newPlan;
};

//////////////////////////////////////////////////////////////////////////////
  
inline void StoreOp::operator delete(void* oldPlan)
{
    // put the join plan on the free list
    ((StoreOp *) oldPlan)->nextOp = freeStoreOp;
    freeStoreOp = (StoreOp *) oldPlan;
};

//////////////////////////////////////////////////////////////////////////////

inline void* TermOp::operator new(size_t size)
{
    TermOp *newPlan;

    if (freeTermOp == NULL) {
	// allocate a new join plan
	newPlan = (TermOp *) new char[size];
	if (Flags[TRACE_LEAK])
	  printf("new TermOp %d\n", ++numTermOps);
    } else {
	// reuse an existing join plan
        newPlan = freeTermOp;
	freeTermOp = freeTermOp->nextOp;
    }
    return newPlan;
};

//////////////////////////////////////////////////////////////////////////////
  
inline void TermOp::operator delete(void* oldPlan)
{
    // put the join plan on the free list
    ((TermOp *) oldPlan)->nextOp = freeTermOp;
    freeTermOp = (TermOp *) oldPlan;
};

//////////////////////////////////////////////////////////////////////////////

inline void* JoinOp::operator new(size_t size)
{
    JoinOp *newPlan;

    if (freeJoinOp == NULL) {
	// allocate a new join plan
	newPlan = (JoinOp *) new char[size];
	if (Flags[TRACE_LEAK])
	  printf("new JoinOp %d\n", ++numJoinOps);
    } else {
	// reuse an existing join plan
        newPlan = freeJoinOp;
	freeJoinOp = freeJoinOp->nextOp;
    }
    return newPlan;
};

//////////////////////////////////////////////////////////////////////////////
  
inline void JoinOp::operator delete(void* oldPlan)
{
    // put the join plan on the free list
    ((JoinOp *) oldPlan)->nextOp = freeJoinOp;
    freeJoinOp = (JoinOp *) oldPlan;
};
