// various opcodes
#define SELECT       1
#define STORE        2
#define JOIN         3
#define SPOOL        4
#define DELAYOP      5
#define TERMOP       6

// various Scheduler type classifications
#define LEFTDEEP 1
#define RIGHTDEEPONE 2
#define RIGHTDEEPALL 3
#define RIGHTDEEPSTATIC1 4
#define RIGHTDEEPSTATICa 5
#define RIGHTDEEP4 6
#define RIGHTDEEP5 7
#define RIGHTDEEP6 8
#define BUSHYMAX 9
#define BUSHY1 10
#define BUSHY2 11
#define BUSHY3 12
#define BUSHY4 13
#define BUSHY5 14
#define BUSHY6 15
#define BUSHY7 16
#define BUSHY8 17                 /* don't forget to update MAX_SCHEDTYPES! */

#define MAX_SCHEDTYPES 18         /* don't forget to update this! */

// Array of plan name prefixes which correspond to the previous scheduling
// disciplines.  Their position in the array must correspond to the value
// of the scheduling discipline as defined above.  Need something explicit
// and more general than this to specify scheduling in a plan!  KLUDGE

#ifdef isparser 
  static char PlanPrefixes[MAX_SCHEDTYPES][10] = {
    "1reserved",
    "ld",
    "rd1",
    "rda",
    "rds1",
    "rdsa",
    "rd4",
    "rd5",
    "rd6",
    "bsm",
    "bs1",
    "bs2",
    "bs3",
    "bs4",
    "bs5",
    "bs6",
    "bs7",
    "bs8"
  };
#else
  extern char PlanPrefixes[MAX_SCHEDTYPES][10];
#endif

// various JOIN phases, operator initialized to 1
#define BUILD 1
#define PROBE 2

// Codes which indicate how to interpret minMem and maxMemNeeded fields

#define BUFPOOL_UNITS 1
#define MINREQ_UNITS  2
#define BYTE_UNITS    3
#define KBYTE_UNITS   4
#define MBYTE_UNITS   5
#define MAXREQ_UNITS  6

class SplitTable;
class Mailbox;
class File;
class Index;
class JobClass;
class RandomPlan;

class QueryPlan : public Object {
  public:
    int		transId;
    char*       name;           // valid if this node was stamped from the root
                                // node of a PRPL "plan x = ..." stmt
    JobClass*   jobClass;       // pointer to job class object
    RandomPlan* statsPlan;      // random plan in which to store plan oriented 
                                // stats.  Stats are ignored (tossed) if NULL
    int         schedType;      // type of scheduling for this query
    int		opCode;		// an opcode
    QueryPlan*	parent;
    TIME	startTime;
    int		outTupSize;	// size of result tuples in bytes
    int		noActiveDisk;	// number of nodes operator will be active on
    bool	activeSites[MAX_NODES]; // true if operator is to run on a node
    Mailbox*	sites[MAX_NODES]; // port of the process running the operator
    int		activeCnt;	// number of operators currently active
    int		tupleCnt;	// number of tuples actually processed ???
    double      execTime;       // actual execution time of the operator
    int         numOutTuples;   // number of estimated output tuples of the op
      
    int         maxPagesNeeded; // maximum memory needed by the operator
    int         minPagesNeeded; // minimum memory needed by the operator

    float       minMemNeeded;   // minimum memory requested by the operator
    int         minMemUnits;    // how to interpret mimMemNeeded value
    float       maxMemNeeded;   // maximum memory requested by the operator
    int         maxMemUnits;    // how to interpret maxMemNeeded value

    // The following statistics are located via the particular operators 
    // that are implementing this particular query plan node.
    // The accumStats() method on QueryPlan objects recursively accumulates
    // statistics for child plan nodes into the parent, as well as rolling
    // up statistics of child plan nodes into a parent plan node flagged as
    // a statistics collection point.  Upon completion of the query, a terminal
    // takes responsibility for accumulating the statistics from the root node
    // of a query plan into the job class the query was issued under.

    float     cnt_bufRef;       // count of buffer references by this plan node
    float     cnt_bufHit;       // count of buffer hits by this plan node
    float     cnt_bufDirty;     // count of buf pages dirtied by this plan node
    float     cnt_diskRead;     // count of disk reads by this plan node
    float     cnt_diskCacheHit; // count of disk cache hits by this plan node
    float     cnt_CPUInstr;     // count of CPU instructions used by this node
    float     cnt_msgsSent;     // tot number of msgs sent
    float     cnt_msgsReceived; // tot number of msgs received
    float     cnt_blockedSends; // number of sends blocked 
    float     cnt_blockedRecvs; // number of recvs blocked 
    float     cnt_bytesSent;    // tot msg bytes sent
    float     cnt_bytesReceived; // tot msg bytes received
    float     sendTime;         // time waiting for n/w latency or blocking
    float     recvTime;         // time waiting for n/w latency or blocking
    float     estDiskSecs;      // estimated seconds of disk demand

    QueryPlan();
    void print(int indent);
    void printName(int indent);
    virtual void accumStats(){}; // rollup stats from child nodes into this one
    void accumPlanStats();      // accumulates plan-based stats in randomPlan
    void pullStats(QueryPlan* source); // pull stats from source to this node
    QueryPlan* findRoot();  // return the root of the query tree
    virtual void assignSplitTbl(SplitTable* st) {
      if (st); // eliminate "not used" msg from compiler
      printf("BARF!  QueryPlan::assignSplitTbl called!\n");
      int a = 0, b = 1, c = b/a;  // please blow up!
    };

};

//////////////////////////////////////////////////////////////////////////////

class SelectPlan: public QueryPlan {
private:
    static int numSelectPlans;
    static class SelectPlan* freeSelectPlan;	// list of free requests
  public:
    SelectPlan* nextPlan;
    File*	file;		
    Index*      index;             // if != NULL, index to use for scan on file
    float	selFactor;	   // selectivity factor of operator
    float	indexSelectivity;  // sel factor of the index (iscans only)
    int		inTupSize;
    SplitTable*	splitTable;	
    
    SelectPlan() {
      opCode = SELECT;
      file = NULL;
      index = NULL;
      inTupSize = -1;
      selFactor = 0.0;
      indexSelectivity = 0.0;
      splitTable = NULL;
    };
    ~SelectPlan();
    void print(int indent);
    inline void* operator new(size_t size);  // new is overloaded
    inline void operator delete(void* oldPlan);  // delete is overloaded too   
    void accumStats();          // rollup stats from child nodes into this one
    void assignSplitTbl(SplitTable* st) { splitTable = st; };
};

//////////////////////////////////////////////////////////////////////////////

class JoinPlan: public QueryPlan {
private:
    static int numJoinPlans;
    static class JoinPlan* freeJoinPlan;	// list of free requests
  public:
    JoinPlan*   nextPlan;
    float	selFactor;	// selectivity factor of operator
    int		inStreamsL;	// number of eof's expected from left child
    int		inStreamsR;	// number of eof's expected from right child
    int		inTupSizeL;	// size of tuples (in bytes) from left child
    int		inTupSizeR;	// size of tuples (in bytes) from right child
    SplitTable*	splitTable;	
    QueryPlan*	leftChild;
    QueryPlan*	rightChild;
    int         phase;          // indicates join phase (BUILD or PROBE)
    int         numBuckets;     // number of buckets. KLUDGE for now assume a
                                // fixed number of buckets for the join
    float       weightFactor;   // extra entries to be added for sending more
                                //    tuples to bucket 0
    int         bucketKMemory;  //  memory needed for buckets 2..n
                                //  in split table (set in makesched.c)

    JoinPlan() {
      opCode = JOIN;
      selFactor =  weightFactor = 0.0;
      inStreamsL = inStreamsR = inTupSizeL = inTupSizeR = bucketKMemory = 0;
      splitTable = NULL;
      numBuckets = 1;
    };
    ~JoinPlan();
    void print(int indent);
    inline void* operator new(size_t size);   // new is overloaded
    inline void operator delete(void* oldPlan);  // delete is overloaded too   
    void accumStats();         // rollup stats from child nodes into this one
    void assignSplitTbl(SplitTable* st) { splitTable = st; };
    optMaxPages();  // returns the max number of pages needed to run the join
    optMinPages();  // returns the min number of pages needed to run the join

};

/////////////////////////////////////////////////////////////////////////////


class RootPlan: public QueryPlan {
 public:
    int       totalMinPagesNeeded[MAX_NODES];  // min. numer of pages needed
    int       totalMaxPagesNeeded[MAX_NODES];  // max. number of pages needed
    RandomPlan* randPlan;     // pointer to parent random plan
    QueryPlan*	child;
    int		inStreams;	// number of eofs expected from child
    int		inTupSize;	// size of tuples (in bytes) from child

    TIME      arrivalTime;    // time recieved by global scheduler process
    TIME      planRespTime;   // completion time - arrival time (no delay incl)
    
    RootPlan() { randPlan = NULL; inStreams = 0; inTupSize = -1; };
};

//////////////////////////////////////////////////////////////////////////////

class StorePlan: public RootPlan {
  private:
    static int numStorePlans;
    static class StorePlan* freeStorePlan;	// list of free requests
  public:
    StorePlan*  nextPlan;
    File*	file;		
    int		diskId;		// which disk the store should use

    StorePlan() {
      opCode = STORE;
      file = NULL;
      diskId = 0;
    };
    void print(int indent);
    inline void* operator new(size_t size);  // new is overloaded
    inline void operator delete(void* oldPlan);  // delete is overloaded too   
    void accumStats();          // rollup stats from child nodes into this one
};


//////////////////////////////////////////////////////////////////////////////
// Simulates a user terminal - just takes tuples as input with no action

class TermPlan: public RootPlan {
  private:
    static int numTermPlans;
    static class TermPlan* freeTermPlan;	// list of free requests
  public:
    TermPlan*   nextPlan;

    TermPlan() {
      opCode = TERMOP;
      child = NULL;
    };
    void print(int indent);
    inline void* operator new(size_t size);  // new is overloaded
    inline void operator delete(void* oldPlan);  // delete is overloaded too   
    void accumStats();          // rollup stats from child nodes into this one
};

//////////////////////////////////////////////////////////////////////////////

inline void* SelectPlan::operator new(size_t size)
{
    SelectPlan *newPlan;

    if (freeSelectPlan == NULL) {
	// allocate a new select plan
	newPlan = (SelectPlan *) new char[size];
	if (Flags[TRACE_LEAK])
	  printf("new selectplan %d\n", ++numSelectPlans);
    } else {
	// reuse an existing select plan
        newPlan = freeSelectPlan;
	freeSelectPlan = freeSelectPlan->nextPlan;
    }
    return newPlan;
};

//////////////////////////////////////////////////////////////////////////////

inline void SelectPlan::operator delete(void* oldPlan)
{
    // put the select plan on the free list
    ((SelectPlan *) oldPlan)->nextPlan = freeSelectPlan;
    freeSelectPlan = (SelectPlan *) oldPlan;
};

//////////////////////////////////////////////////////////////////////////////

inline void* StorePlan::operator new(size_t size)
{
    StorePlan *newPlan;

    if (freeStorePlan == NULL) {
	// allocate a new store plan
	newPlan = (StorePlan *) new char[size];
	if (Flags[TRACE_LEAK])
	  printf("new storeplan %d\n", ++numStorePlans);
    } else {
	// reuse an existing store plan
        newPlan = freeStorePlan;
	freeStorePlan = freeStorePlan->nextPlan;
    }
    return newPlan;
};

//////////////////////////////////////////////////////////////////////////////

inline void StorePlan::operator delete(void* oldPlan)
{
    // put the store plan on the free list
    ((StorePlan *) oldPlan)->nextPlan = freeStorePlan;
    freeStorePlan = (StorePlan *) oldPlan;
};

//////////////////////////////////////////////////////////////////////////////

inline void* TermPlan::operator new(size_t size)
{
    TermPlan *newPlan;

    if (freeTermPlan == NULL) {
	// allocate a new term plan
	newPlan = (TermPlan *) new char[size];
	if (Flags[TRACE_LEAK])
	  printf("new termplan %d\n", ++numTermPlans);
    } else {
	// reuse an existing term plan
        newPlan = freeTermPlan;
	freeTermPlan = freeTermPlan->nextPlan;
    }
    return newPlan;
};

//////////////////////////////////////////////////////////////////////////////

inline void TermPlan::operator delete(void* oldPlan)
{
    // put the term plan on the free list
    ((TermPlan *) oldPlan)->nextPlan = freeTermPlan;
    freeTermPlan = (TermPlan *) oldPlan;
};

//////////////////////////////////////////////////////////////////////////////

inline void* JoinPlan::operator new(size_t size)
{
    JoinPlan *newPlan;

    if (freeJoinPlan == NULL) {
	// allocate a new join plan
	newPlan = (JoinPlan *) new char[size];
	if (Flags[TRACE_LEAK])
	  printf("new joinplan %d\n", ++numJoinPlans);
    } else {
	// reuse an existing join plan
        newPlan = freeJoinPlan;
	freeJoinPlan = freeJoinPlan->nextPlan;
    }
    return newPlan;
};

//////////////////////////////////////////////////////////////////////////////
  
inline void JoinPlan::operator delete(void* oldPlan)
{
    // put the join plan on the free list
    ((JoinPlan *) oldPlan)->nextPlan = freeJoinPlan;
    freeJoinPlan = (JoinPlan *) oldPlan;
};

