//////////////////////////////////////////////////////////////////////////////
//
//  Name:        sched.h
//
//  Title:       Simulator scheduler class definition.
//
//  Change Activity:
//
//     08//91 (MM) - Initial version.
//
//////////////////////////////////////////////////////////////////////////////

// fudge factor - used for calculating the size of join hash tables
#define FUDGEFACTOR 1.2

// maximum number of joins allowed in a query
#define MAXJOINS 15

///////////////////////////////////////////////////////////////////////////////
// 
// "Delayed" qtable -- just waits until time passes before really recording
// anything in the base class CSIM qtable.

class Dqtable : public qtable {
  float delay;              // msecs to wait b4 recording anything
  bool statWaitExpired;

public:
  Dqtable(float secs, char* name) : qtable(name) 
   { delay = secs * 1000.0; 
     statWaitExpired = (delay != 0.0) ? FALSE : TRUE;
   }
  void note_entry();
  void note_exit();
  void forcedNote_entry() { qtable::note_entry(); }; // ignore delay!
  void forcedNote_exit() { qtable::note_exit(); }; // ignore delay!
};

//////////////////////////////////////////////////////////////////////////////

class Operator {
 private:
  static int numOper;
  static class Operator* freeOperator;
 public:
  QueryPlan*  oprPlan;   // operator in this dependency group
  Operator*  nextOpr;    // pointer to the next operator in the row
  Operator*  nextReq;    // linked list of free operators
  int phase;             // used to the separate the build and the probe 
                         //    operator nodes for a join Plan
  int bucketNum;         // number of the bucket being processed by this 
                         //     operator

  Operator() {
    oprPlan = NULL;
    nextOpr = NULL;
  };
  void print();
  inline void* operator new(size_t size);      	       // new is overloaded
  inline void  operator delete(void* oldReq);          // delete is overloaded
};

//////////////////////////////////////////////////////////////////////////////

class Schedule {

 private:
  static int numSched;
  static class Schedule* freeSchedule;
 public:
  Operator*   oprRow;     // list of dependency groups
  Schedule*   nextRow;    // pointer to next dependency group
  Schedule*  nextReq;    // linked list of free operators

  Schedule() {
    oprRow = NULL;
    nextRow = NULL;
  };
  ~Schedule();
  void print();
  inline void* operator new(size_t size);       // new is overloaded
  inline void  operator delete(void* oldReq);   // delete is overloaded
};

//////////////////////////////////////////////////////////////////////////////

class QueryReq {
  private:
    static int numQReq;
    static class QueryReq* freeQueryReq;
  public:
    Node *node;                    // pointer to scheduling node 
    QueryMsg *reqMsg;              // query request message
    Schedule *schedule;
    Schedule *waitSchedule;
    int minPagesNeeded[MAX_NODES]; // mini number of pages needed at each site 
                                   //      to run the query
    QueryReq *nextReq;      // pointer to next request in queue
    QueryReq *nextFreeReq;

  QueryReq() {
    node = NULL;
    reqMsg = NULL;
    schedule = NULL;
    waitSchedule = NULL;
    nextReq = NULL;
    nextFreeReq = NULL;
  };
  inline void* operator new(size_t size);       // new is overloaded
  inline void  operator delete(void* oldReq);   // delete is overloaded
};

//////////////////////////////////////////////////////////////////////////////

class QueryQueue {
  public:

	QueryReq *qhead;
	QueryReq *qtail;
	Dqtable* queryQueueStat;

    QueryQueue(float delay) {
            qhead = NULL;
	    qtail = NULL;
	    queryQueueStat = new Dqtable(delay, "query queue wait");
        }

    // see if queue is empty
    int nonempty()
        {
    	    return(qhead != NULL);
        };

    // insert into query queue
    void enqueue(Node*, QueryMsg*, Schedule*, Schedule*);  

    // remove query from the queue, default = remove head of queue
    QueryReq *dequeue(QueryReq*& query);

    // remove next eligible query from the queue
    QueryReq *findNext();

    // print queue contents
    void print();
	
    // print queue statistics
    void printStats();

};

//////////////////////////////////////////////////////////////////////////////

class MemoryQueue: public QueryQueue {
  private:
      // check if any site clash present in the two memory arrays
      bool siteClash(int queryPagesNeeded[MAX_NODES],
		     int currentPagesNeeded[MAX_NODES]);

  public:
      MemoryQueue(float delay) : QueryQueue(delay) {};

      // check if any of the waiting queries can run given the present number
      //     of pages in BufPages and fork off a query scheduler if they can
      //     the query does not run if it has a site conflict with any of the
      //     waiting queries
      void checkQueue();
      
      // check if any of the queries waiting have a site clash with
      //    the new query whose memory request is in currentPagesNeeded
      bool checkWaiters(int currentPagesNeeded[MAX_NODES]);

      // these methods same as in QueryQueue except that they also
      // update the job class memory wait statistics
      void enqueue(Node*, QueryMsg*, Schedule*, Schedule*);  
      QueryReq *dequeue(QueryReq*& query);
};

//////////////////////////////////////////////////////////////////////////////


class MPLQueue : public QueryQueue {
  public:
      MPLQueue(float delay) : QueryQueue(delay) {};

      // remove next eligible query from the queue
      // checks if class mpl is acceptable 
      QueryReq *findNext();
};    


//////////////////////////////////////////////////////////////////////////////

inline void* Schedule::operator new(size_t size) {
    Schedule *newReq;

    if (freeSchedule == NULL) {
	// allocate a new Schedule
	newReq = (Schedule *) new char[size];
	if (Flags[TRACE_LEAK])
	  printf("new Sched %d\n", ++numSched);
    } else {
	// reuse an existing Schedule
        newReq = freeSchedule;
	freeSchedule = freeSchedule->nextReq;
    }
    return newReq;
};

//////////////////////////////////////////////////////////////////////////////

inline void Schedule::operator delete(void* oldReq) {
    // put the Schedule on the free list
    ((Schedule *) oldReq)->nextReq = freeSchedule;
    freeSchedule = (Schedule *) oldReq;
};

//////////////////////////////////////////////////////////////////////////////

inline void* Operator::operator new(size_t size) {
    Operator *newReq;

    if (freeOperator == NULL) {
	// allocate a new operator
	newReq = (Operator *) new char[size];
	if (Flags[TRACE_LEAK])
	  printf("new oper %d\n", ++numOper);
    } else {
	// reuse an existing operator
        newReq = freeOperator;
	freeOperator = freeOperator->nextReq;
    }
    return newReq;
};

//////////////////////////////////////////////////////////////////////////////

inline void Operator::operator delete(void* oldReq) {
    // put the operator on the free list
    ((Operator *) oldReq)->nextReq = freeOperator;
    freeOperator = (Operator *) oldReq;
};

//////////////////////////////////////////////////////////////////////////////

inline void* QueryReq::operator new(size_t size) {
    QueryReq *newReq;

    if (freeQueryReq == NULL) {
	// allocate a new QueryReq
	newReq = (QueryReq *) new char[size];
	if (Flags[TRACE_LEAK])
	  printf("new qreq %d\n", ++numQReq);
    } else {
	// reuse an existing QueryReq
        newReq = freeQueryReq;
	freeQueryReq = freeQueryReq->nextFreeReq;
    }
    return newReq;
};

//////////////////////////////////////////////////////////////////////////////

inline void QueryReq::operator delete(void* oldReq) {
    // put the QueryReq on the free list
    ((QueryReq *) oldReq)->nextFreeReq = freeQueryReq;
    freeQueryReq = (QueryReq *) oldReq;
};
