#include <string.h>
#include "dbsim.h"
#include "object.h"
#include "node.h"
#include "msg.h"
#include "mailbox.h"
#include "split.h"
#include "language.h"
#include "db.h"
#include "query.h"
#include "sched.h"
#include "bufmgr.h"

///////////////////////////////////////////////////////////////////////////////
//
//  Name:        gen_sched.c
//
//  Function:    a generic query scheduler
//                 
//  Change Activity:
//
//
///////////////////////////////////////////////////////////////////////////////

extern Node* Nodes[];
extern Mailbox* GlobalSchedPort;
extern int NumDiskNodes;
extern JobClass *JobClasses[];

extern void startQuery(Node*, QueryMsg*);
extern int actualResourceAllocation(QueryMsg*, Schedule*&, Schedule*&);
void DeleteSched(Schedule*&);
extern MemoryQueue *GlobalMemoryWaitQueue;
extern MemoryQueue *GlobalMPLWaitQueue;
extern double TimeWeightedMPL;
extern double FirstMPLEntry;
extern int MaxMPL;
extern int BufPages[];
extern double PrevMPLChange;
extern int CurrentMPL;
extern float GlobalStatWaitSecs;

int allocateResources(QueryMsg*, Schedule*&, Schedule*&);
void QueryScheduler(Node* , QueryMsg*, Schedule*, Schedule*);
void startQuery(Node*, QueryMsg*);
void processOperatorDone(CntlMsg*, int&);
void genericScheduler(Schedule*, Schedule*, Mailbox*, int transId);
int activate(QueryPlan* , Mailbox*, int);
void startOperator(QueryPlan* , Mailbox*, int);
int getPortIds(QueryPlan*, Mailbox*);
void fillSplitAddrs (QueryPlan*, SplitTable*, int);
SplitTable* copySplitTable(SplitTable*);
void sendSplitTblMsg(QueryPlan* , SplitTable* , Mailbox* );


int allocateResources(QueryMsg* reqMsg, Schedule*& schedule, 
		      Schedule*& waitSchedule) {
  int e;
  
  // form the schedules. also does the resource allocation
  e = actualResourceAllocation(reqMsg, schedule, waitSchedule);
  
  // if the query not scheduled properly, decrease class mpl
  if (e != OK) {
    reqMsg->query->jobClass->currentMPL--;
    ((RootPlan*) (reqMsg->query))->randPlan->currentMPL--;
    
    // Send "the results" back to the terminal, reusing the request pkt
      reqMsg->send(reqMsg->sourceBox, GlobalSchedPort);
    
    // destroy the old schedules
    DeleteSched(schedule);
    DeleteSched(waitSchedule);
  }
  return(e);
}

//////////////////////////////////////////////////////////////////////
//
// Generic Query Scheduler
//   In - 1) node - 
//        2) reqMsg - scheduler request message, contains pointer to 
//                      query tree  
//   Action
//     calls the appropriate query schedule generator and then calls 
//     the generic scheduler to actually schedule all the operators
//
///////////////////////////////////////////////////////////////////////
void QueryScheduler(Node* node, QueryMsg* reqMsg, 
		    Schedule* schedule, Schedule* waitSchedule) 
{

    Mailbox* 	qschedMbox;
    int nodeNo, numSites, totalMemory;
    QueryReq* nextQuery;
    register JobClass* jc;
    RandomPlan* statsPlan;
    RootPlan* root;
    float queryStartTime = clock;

    create("QueryScheduler");
    
    if (Flags[TRACE_SCHED]) {
    	printf("query scheduler process started\n");
    	reqMsg->print("query scheduler got");
    };

    root = (RootPlan*) (reqMsg->query);
    jc = root->jobClass;
    
    // check the global wait queue to see if another query can run
    // done here because only after resource allocation do we know 
    // the actual number of pages left in the buffer pool
    //  this check enables multiple queries to start after one query completes
    //  as each starting query after doing its own resource allocation will
    //  also check the next waiting query

    GlobalMemoryWaitQueue->checkQueue();

    // Updates global, jobclass, and plan related MPL statistics.

    if ((GlobalStatWaitSecs != 0.0) && 
	(clock > GlobalStatWaitSecs) && (PrevMPLChange == 0.0))
      PrevMPLChange = FirstMPLEntry = clock;
    if ((GlobalStatWaitSecs == 0.0) || (clock > GlobalStatWaitSecs)) {
      TimeWeightedMPL += (clock - PrevMPLChange) * CurrentMPL;
      PrevMPLChange = clock;  
      if ((CurrentMPL + 1) > MaxMPL) MaxMPL = CurrentMPL + 1;
    }
    CurrentMPL++;

    if ((jc->statWaitSecs != 0.0) &&
	(clock > (jc->statWaitSecs * 1000.0)) && 
	(jc->prevMPLChange == 0.0)) {
	jc->prevMPLChange = jc->firstMPLEntry = clock;
    }

    if ((jc->statWaitSecs == 0.0) || 
	(clock > (jc->statWaitSecs * 1000.0))) {
      jc->timeWeightedMPL += (clock - jc->prevMPLChange) 
	* jc->execMPL;
      jc->prevMPLChange = clock;
      if ((jc->execMPL + 1) > jc->actualMaxMPL)
	jc->actualMaxMPL = jc->execMPL + 1;
    }

    jc->execMPL++;

    if (root->statsPlan != NULL) {
       statsPlan = root->statsPlan;
       statsPlan->timeWeightedMPL += (clock - statsPlan->prevMPLChange)
                * statsPlan->execMPL;
       statsPlan->prevMPLChange = clock;
       statsPlan->execMPL++;
       if (statsPlan->execMPL > statsPlan->actualMaxMPL)
                statsPlan->actualMaxMPL = statsPlan->execMPL;
    }

    // get a mail mailbox for operators to communicate with
    qschedMbox = new Mailbox("qschedMbox", node);
    
    float startExecTime = clock;

    //
    // call the generic query scheduler
    //
    genericScheduler(schedule, waitSchedule, qschedMbox, root->transId);

    root->execTime = clock - startExecTime;
    root->planRespTime = clock - root->arrivalTime;

    //after the query completes update the bufPages array
    //also update memory usage statistics by calculating the
    //average amount of memory given to the query over all the sites

    totalMemory = numSites = 0;
    for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++) {
        BufPages[nodeNo] += root->totalMaxPagesNeeded[nodeNo];
        totalMemory += root->totalMaxPagesNeeded[nodeNo];
        if (root->totalMaxPagesNeeded[nodeNo] > 0) {
	  numSites++;
	  // If buffer manager needs to be notified of mem release, does it.
	}
    }

    // Updates class memory usage statistics

    jc->usedMem -= (int) (totalMemory * 1.0 / numSites);
    if (root->statsPlan != NULL)
       root->statsPlan->memAlloc->record((int)(totalMemory * 1.0 / numSites));
    jc->memAlloc->record((int) (totalMemory * 1.0 / numSites));

    RandomPlan* savedStatsPlan = root->statsPlan;
    RandomPlan* savedRandPlan = root->randPlan;
    float savedArrivalTime = root->arrivalTime;

    // Send "the results" back to the terminal, reusing the request pkt
    reqMsg->send(reqMsg->sourceBox, GlobalSchedPort);

    // check the global wait queue to see if another query can run now
    //    after buffers have been given back
    GlobalMemoryWaitQueue->checkQueue();

    // Updates global, jobclass, and plan related MPL statistics.

    if ((GlobalStatWaitSecs != 0.0) && 
	(clock > GlobalStatWaitSecs) && (PrevMPLChange == 0.0))
      PrevMPLChange = FirstMPLEntry = clock;
    if ((GlobalStatWaitSecs == 0.0) || (clock > GlobalStatWaitSecs)) {
      TimeWeightedMPL += (clock - PrevMPLChange) * CurrentMPL;
      PrevMPLChange = clock;  
    }
    CurrentMPL--;

    if ((jc->statWaitSecs != 0.0) &&
	(clock > jc->statWaitSecs) && 
	(jc->prevMPLChange == 0.0)) {
	jc->prevMPLChange = jc->firstMPLEntry = clock;
    }
    if ((jc->statWaitSecs == 0.0) || 
	(clock > jc->statWaitSecs)) {
      jc->timeWeightedMPL += (clock - jc->prevMPLChange) 
	* jc->execMPL;
      jc->prevMPLChange = clock;

    }
    jc->currentMPL--;
    jc->execMPL--;

    jc->readyMPL--;

    savedRandPlan->currentMPL--;
    if (savedStatsPlan != NULL) {
       savedStatsPlan->timeWeightedMPL += 
         (clock - savedStatsPlan->prevMPLChange) * savedStatsPlan->execMPL;
       savedStatsPlan->prevMPLChange = clock;
       savedStatsPlan->execMPL--;
       savedStatsPlan->readyMPL--;
    }

    // if queries waiting in the job class queue start the next query
    if (GlobalMPLWaitQueue->nonempty())
    {
       nextQuery = GlobalMPLWaitQueue->findNext();
       if (nextQuery != NULL) {
           if (Flags[TRACE_SCHED])
               printf("next job class query started");
           startQuery(nextQuery->node, nextQuery->reqMsg);
           delete nextQuery;
       }
    }

    DeleteSched(schedule);
    DeleteSched(waitSchedule);
    delete qschedMbox;
}

/////////////////////////////////////////////////////////////////////////////
//
// Starts queries from classes until they are at their maxMPLs
// Called when MPLs are changed to see if we can admit any waiters.
//
/////////////////////////////////////////////////////////////////////////////
void attemptAdmission() {
    QueryReq* nextQuery;
    int classNum;

    for (classNum = 0; classNum < NumClasses; classNum++) {
      while (JobClasses[classNum]->mplQueue->nonempty() &&
	 ((nextQuery = JobClasses[classNum]->mplQueue->findNext()) != NULL)) {
	  startQuery(nextQuery->node, nextQuery->reqMsg);
	  delete nextQuery;
      }
    }
}


/////////////////////////////////////////////////////////////////////////////
//
// process the completion message received from one of the running operators
//     increment the numOpsDone counter if the operator has completed
//
/////////////////////////////////////////////////////////////////////////////
void processOperatorDone(CntlMsg *cMsg, int& numOpsDone) {
  
  JoinPlan* joinPlan;
  StorePlan* storePlan;
  
  // decrement number of active instances of operator
  cMsg->query->activeCnt--;
  cMsg->query->tupleCnt += cMsg->intData;
  
  if (cMsg->query->opCode == STORE) {
      storePlan = (StorePlan*) cMsg->query;
  }
  
//    if (Flags[TRACE_SCHED])
//    {
//        cMsg->query->printName(5);
//        printf(" instance has completed %d\n", cMsg->query->activeCnt );
//    }
  
  // if no more active instances of the operator
  if (cMsg->query->activeCnt == 0) {
    switch(cMsg->query->opCode) {

    case JOIN:
      joinPlan = (JoinPlan *) cMsg->query;
      if (Flags[TRACE_SCHED]) 
	printf(" JOIN operator has completed phase %d at %f\n", 
	       joinPlan->phase, clock );

      if (joinPlan->phase == BUILD)
	joinPlan->phase = PROBE;     // transition from BUILD to PROBE
      else {
	// set the active count back to the original
	joinPlan->activeCnt = joinPlan->noActiveDisk;
	
	// if attached stats plan, update completions and response time stats
	if (joinPlan->statsPlan != NULL) {
	  joinPlan->statsPlan->completions++;
	  joinPlan->statsPlan->respTimes->record(clock - joinPlan->startTime);
	}
      }
      break;

    case SELECT:
      if (Flags[TRACE_SCHED]) 
	printf(" SELECT operator has completed at %f\n", clock);
      
      // if attached stats plan, update completions and response time stats
      if (cMsg->query->statsPlan != NULL) {
	cMsg->query->statsPlan->completions++;
	cMsg->query->statsPlan->respTimes->record
	  (clock - cMsg->query->startTime);
      }
      break;

    case STORE:
      if (Flags[TRACE_SCHED])
	printf(" STORE operator has completed at %f\n", clock);
      break;

    case TERMOP:
      if (Flags[TRACE_SCHED]) 
	printf(" TERMOP operator has completed at %f\n", clock);
      break;
    }
    
    // increase number of operators done
    numOpsDone++;
  }
}

/////////////////////////////////////////////////////////////////////
//
//  genericScheduler
//     In - 1) schedule
//          2) qschedBox - mailbox for communication
//          3) transId - transaction ID
//     Action - traverses the schedule L->R and top->bottom and
//           activates all the operators. once all the operators in
//           a row have been activates waits for them to finish and
//           processes the next row
//
//           Q - change row-wise synchronization?
//
/////////////////////////////////////////////////////////////////////
void genericScheduler(Schedule* schedule, Schedule* waitSchedule,
		      Mailbox* qschedMbox, int transId)
{

    int totalOprs, numOpsDone ;
    Schedule *scheduleRow, *waitRow;
    Operator* opr;
    CntlMsg *cMsg ;


    //
    // for each row in the schedule
    //
    for(scheduleRow = schedule, waitRow = waitSchedule;
        scheduleRow != NULL, waitRow != NULL; 
        scheduleRow = scheduleRow->nextRow, waitRow = waitRow->nextRow)
    {


       //
       // for each operator in the pipeline group activate the operator
       //
       for(opr = scheduleRow->oprRow; opr != NULL; opr = opr->nextOpr)
            // activate the operator
	    activate(opr->oprPlan, qschedMbox, transId);

       //
       // count total number of operators to wait for
       // have to check for active cnts because the operator may have 
       // already finished while starting some other operators
       //
       totalOprs = 0 ;
       for(opr = waitRow->oprRow; opr != NULL; opr = opr->nextOpr)
           if (opr->oprPlan->activeCnt > 0)
	       totalOprs++;

       //
       // wait for all the operators to finish
       //
       numOpsDone = 0;

       // while all operators not finished
       while (numOpsDone < totalOprs)
       {
       
           // block waiting for next message
           cMsg = (CntlMsg*) qschedMbox->recv();
           if (cMsg->type != CONTROL)
	       printf("expected CONTROL message in getPortIds. got type=%d\n",
		      cMsg->type);
	   else
	   {
	       switch (cMsg->opCode)
	       {
	           case OPDONE:
	           case TBLSTATUS:
		       processOperatorDone(cMsg, numOpsDone);
		       break;
		   default:
		       printf("Unexpected control msg type received %d", 
			      cMsg->opCode);
		       printf("from node %d\n", 
			      cMsg->queryOp, cMsg->sourceBox->owningNode->id);

		       break;
	       }
	       delete cMsg;  // release message buffer
	   }
       }
    }
}



////////////////////////////////////////////////////////////////////////
//
// activate
//
//   Action - activate a node in a query tree
//
////////////////////////////////////////////////////////////////////////
activate(QueryPlan* qpNode, Mailbox* qschedMbox, int transId)
{
    SplitTable  *splitTable;
    int		cnt;

    switch(qpNode->opCode) {
    case JOIN:
      
      if (((JoinPlan*)qpNode)->phase == BUILD) {
	if (Flags[TRACE_SCHED]) 
	  printf("Starting BUILD phase of JOIN\n");
	qpNode->startTime = clock; // record start time of opr
	
	// update the num buckets statistics if this is the first bucket
	((JoinPlan*) qpNode)->jobClass->buckets->record
	  (((JoinPlan*) qpNode)->numBuckets);
	
	startOperator(qpNode, qschedMbox, transId); 
	
	// get the private ports of the various processes involved
	cnt = getPortIds (qpNode, qschedMbox);
	
	// build/init/assign split tables for left & right children

	splitTable = new SplitTable 
	  (cnt, ((JoinPlan*) qpNode)->inTupSizeL, FALSE); 
	fillSplitAddrs (qpNode, splitTable, cnt);
	((JoinPlan*) qpNode)->leftChild->assignSplitTbl(splitTable);

	splitTable = new SplitTable 
	  (cnt, ((JoinPlan*) qpNode)->inTupSizeR, FALSE);
	fillSplitAddrs (qpNode, splitTable, cnt);
	((JoinPlan*) qpNode)->rightChild->assignSplitTbl(splitTable);
	
      } else { // PROBE phase

	if (Flags[TRACE_SCHED]) 
	  printf("Starting PROBE phase of JOIN\n");
	
	// send new split tables all sites
	sendSplitTblMsg(qpNode, ((JoinPlan*) qpNode)->splitTable, qschedMbox);
      }
      break;
      
    case SELECT:
      if (Flags[TRACE_SCHED]) printf("Starting SELECT \n");
      qpNode->startTime = clock; // record start time of opr
      
      // start the select operator
      startOperator(qpNode, qschedMbox, transId); 
      
      // get the private ports of the various operators involved
      cnt = getPortIds (qpNode, qschedMbox);
      break;
      
    case TERMOP:
      if (Flags[TRACE_SCHED]) printf("Starting TERMOP \n");
      
      // start the terminal operator
      startOperator(qpNode, qschedMbox, transId); 
      
      // get the private ports of the various processes involved
      cnt = getPortIds (qpNode, qschedMbox);
      
      // build/init/assign a split table for the child query
      splitTable = new SplitTable 
	(cnt, ((TermPlan*) qpNode)->child->outTupSize, FALSE);
      fillSplitAddrs (qpNode, splitTable, cnt);
      ((TermPlan*) qpNode)->child->assignSplitTbl(splitTable);
      break;
      
    case STORE:
      if (Flags[TRACE_SCHED]) printf("Starting STORE \n");
      
      // start the store operator
      startOperator(qpNode, qschedMbox, transId); 
      
      // get the private ports of the various processes involved
      cnt = getPortIds (qpNode, qschedMbox);
      
      // build/init/assign a split table for the child query
      splitTable = new SplitTable 
	(cnt, ((StorePlan*) qpNode)->inTupSize, FALSE);
      fillSplitAddrs (qpNode, splitTable, cnt);
      ((StorePlan*) qpNode)->child->assignSplitTbl(splitTable);
      break;
    }
    return(OK);
}


//////////////////////////////////////////////////////////////////////////////
//
//
//  start the operator (qry) on all sites where activeSite is TRUE
//
/////////////////////////////////////////////////////////////////////////////
void startOperator(QueryPlan* qry, Mailbox* qschedMbox, int transId)
{
    Mailbox *nodeMgr;
    QueryMsg *qMsg;
    int	i, cnt;

    // initialize some fields in the actual plan
    qry->activeCnt = 0;
    qry->tupleCnt = 0;
    qry->transId = transId;

    cnt=0;
    for (i=0; i < NumDiskNodes; i++) 
    {
	if (qry->activeSites[i] == TRUE)
	{
	    // operator is to run on this node 
            // get the mail box of the node manager on the appropriate node 
            nodeMgr = Nodes[i]->nodeMgrPort;
            qMsg = new QueryMsg; // get query message
            qMsg->query = qry;  
	    qMsg->send(nodeMgr, qschedMbox);
	    cnt++;
	}
    }
    qry->activeCnt = cnt;
}

/////////////////////////////////////////////////////////////////////////////
//
// get the private ports of all the processes executing an operator
// return count of number of active processes
//
/////////////////////////////////////////////////////////////////////////////
int getPortIds (QueryPlan* qry, Mailbox* qschedMbox)
{
  int		replyCnt;  // number of responses received
  CntlMsg	*cMsg;
  int		cnt;
  int		nodeId;
  int           dummy;     // dummy argument for processOperatorDone

  cnt = qry->activeCnt;
  replyCnt = 0;
  
  // loop until the message ports of all the operator processes are received
  while (replyCnt < cnt) {
    // block waiting for next message
    cMsg = (CntlMsg*) qschedMbox->recv();
    if (cMsg->type != CONTROL)
      printf("expected CONTROL message in getPortIds. got type=%d\n",
	     cMsg->type);
    else {
      switch (cMsg->opCode) {
      case NEWTASK:
	// update list of active sites
	replyCnt++;
	nodeId = cMsg->intData;  // kind of a kludge but ok
	qry->sites[nodeId] = cMsg->sourceBox; // extract opr's port
	
//	if (Flags[TRACE_SCHED])
//	  printf("reply recieved replycnt %d from node %d\n", 
//	  replyCnt, cMsg->sourceBox->owningNode->id);
	break;
      case OPDONE:
	if (cMsg->query == qry) {
//	if (Flags[TRACE_SCHED])
//	  printf("reply recieved replycnt %d from node %d\n", 
//	  replyCnt, cMsg->sourceBox->owningNode->id);
      }
	
      case TBLSTATUS:
	processOperatorDone(cMsg, dummy);
	break;
      default:
	printf("Unexpected control message type received = %d\n", 
	  cMsg->opCode);
	break;
      }
      delete cMsg;  // release message buffer
    }
  }

//    if (Flags[TRACE_SCHED])
//         printf("getPortIds done replycnt %d\n", replyCnt);

  return(replyCnt);
}
	
/////////////////////////////////////////////////////////////////////////////
//
// Use the mailbox's of the operators executing the parent operator
// to construct a split table for the child operator
// Return the number of entries in the splitTable
//
/////////////////////////////////////////////////////////////////////////////
void fillSplitAddrs (QueryPlan* qry, SplitTable* splitTable, int size)
{
    // qry points at the parent's query plan
    // splitTable points at the child's split table
    // size is the size of the splitTable

    int		i;
    int		splitCnt;

    splitCnt = 0;
    for (i=0; i  < NumDiskNodes; i++) 
    {
	if (qry->activeSites[i] == TRUE)
	{
	      // parent operator is being run on node i
#ifndef LIGHT
	      splitTable->splitArray[splitCnt].numTuples = 0;
#endif
	      splitTable->splitArray[splitCnt].destNode = qry->sites[i];
	      splitCnt++;
	}
    }
    if (splitCnt != size) 
	printf("\n\nmajor problem in fillSplitAddres, size=%d, splitCnt=%d\n\n"
	       ,size, splitCnt);
    splitTable->splitCnt = splitCnt;
}


// make a copy of a split table
SplitTable* copySplitTable(SplitTable* from)
{
    int 	i;
    SplitTable*	to;

    to = new SplitTable (from->splitCnt, from->tupleSize, from->fullBroadcast);
    to->splitCnt = from->splitCnt;
    for (i=0;i<from->splitCnt;i++) to->splitArray[i] = from->splitArray[i];
    return(to);
}


// send the split table to all sites where qry running
void sendSplitTblMsg(QueryPlan* qry, SplitTable* splitTable,
	Mailbox* qschedMbox)
{
    CntlMsg 	*cMsg;
    int		i;
	
    for (i=0; i < NumDiskNodes; i++) 
    {
	if (qry->activeSites[i] == TRUE)
	{
	    qry->activeCnt++;
	    cMsg = new CntlMsg; // get control message
	    cMsg->opCode = NEWSPLIT;  
#ifdef LIGHT
	    // appoximated split tables. all ops use same base split table
	    cMsg->splitTable = splitTable;
#else
	    cMsg->splitTable = copySplitTable(splitTable);
#endif
	    cMsg->send (qry->sites[i], qschedMbox);
	}
    }
}
