#include "dbsim.h"
#include "object.h"
#include "node.h"
#include "query.h"
#include "language.h"
#include "db.h"
#include "mailbox.h"
#include "msg.h"
#include "sched.h"
#include <math.h>
#include <string.h>


///////////////////////////////////////////////////////////////////////////////
//
//  Name:        makesched.c
//
//  Author:      MM
//
//  Function:    has procedures to form different types of schedules and
//               perform resource allocation on the query plan
//
///////////////////////////////////////////////////////////////////////////////

extern "C" {
  double ceil(double x);
  double floor(double x);
//  double sqrt(double x); kbaix
}

#define SELECT_PAGES 2
#define STORE_PAGES 2
extern int minPagesNeeded[], maxPagesNeeded[];
extern MemoryQueue *GlobalQueryWaitQueue;
extern int BufPages[];

void DeleteSched(Schedule*&);
int initialSchedule(QueryMsg*, Schedule*&, Schedule*&);
int initQueryRequirements(QueryPlan*, bool);
int makeSchedule(RootPlan*, Schedule*&, Schedule*&);
int initResourceAllocation(QueryMsg*, Schedule*, Schedule*);
int CanRun(RootPlan*);
int actualResourceAllocation(QueryMsg*, Schedule*&, Schedule*&);
void bushyExec(QueryPlan*, Schedule*&, Schedule*&, Schedule*&, Schedule*&, 
	       Schedule*&, Schedule*&, bool);

/////////////////////////////////////////////////////////////////////////////
//
//   initialSchedule
//     given the root of the query tree
//     (1) does initial resource requirement calculation
//     (2) creates a new schedule
//     (3) checks to see if the query can run, if not it is sent to the global 
//           wait queue
//
//////////////////////////////////////////////////////////////////////////////
int initialSchedule(QueryMsg* reqMsg, Schedule*& schedule, 
		    Schedule*& waitSchedule) {
  int e;
  RootPlan* root = (RootPlan*) reqMsg->query;

  // set the number of buckets and max and min pages needed by each operator
  e = initQueryRequirements(reqMsg->query, TRUE);
  if (e  != OK) return(NOTOK);

  // form the initial schedule
  if ((e = makeSchedule(root, schedule, waitSchedule)) != OK) 
    return(NOTOK);

  // perform initial query resource allocation and check if the query can run
  //   if it cannot the query is sent to the GlobalQueryWaitQueue
  // KLUDGE:: if there is load control then initial resource allocation is
  //   bypassed. this assumes that the query has smaller requirement than the
  //   global memory of the system
  if ((e = initResourceAllocation(reqMsg, schedule, waitSchedule)) != OK) 
      return(NOTOK);

  return(OK);

}

/////////////////////////////////////////////////////////////////////////
//
//  set the maximum and minimum memory requirement for the query plan
//      pointed to by root. also calculates the number of buckets to be
//      used for every join. firstTime TRUE if query requirements being
//      found for the first time.
//
/////////////////////////////////////////////////////////////////////////
int initQueryRequirements(QueryPlan* root, bool firstTime) {

  QueryPlan *qPlan = NULL;
  SelectPlan *selPlan = NULL;
  JoinPlan *joinPlan = NULL;
  int numInTuples, optMaxPagesNeeded, optMinPagesNeeded;
  int Memory, Buckets, e;
  
  if (root == NULL) return OK;  // caller was at leaf node of query tree

  switch (root->opCode) {
  case SELECT:
    selPlan = (SelectPlan*) root;
    
    // find the number of output tuples of the select
    if (selPlan->index != NULL)
      selPlan->numOutTuples = 
	(int) ceil(selPlan->file->numTuples 
		   * selPlan->indexSelectivity * selPlan->selFactor);
    else
      selPlan->numOutTuples = 
	(int) ceil(selPlan->file->numTuples * selPlan->selFactor);
    
    /* *****KLUDGES(!)*****
       
       1) the select and store operator do not reserve pages
       while executing. so they take out an option to ensure
       that they are guaranteed at least two pages at every site
       so the total number of buffer pages is decremented by 2 
       for each select or store
       
       2) min requirement also set to 2 though they can work with 
       just one page each. this causes complications in 
       resourceAllocation where we need to find the memory to be 
       allocated to the join
    */


    root->maxPagesNeeded = SELECT_PAGES;
    root->minPagesNeeded = SELECT_PAGES;
    
    if (Flags[TRACE_SCHED])
      printf("SELECT maxPages %d minPages %d numOutTuples %d\n", 
	     root->maxPagesNeeded, root->minPagesNeeded, root->numOutTuples);
    break;

  case TERMOP:
    // process the child of the terminal operator
    e = initQueryRequirements(((TermPlan*) root)->child,firstTime);
    if (e != OK) return(NOTOK);
    
    // find the number of output tuples of the operator
    root->numOutTuples = ((TermPlan*) root)->child->numOutTuples;
    root->maxPagesNeeded = 0;
    root->minPagesNeeded = 0;
    
    if (Flags[TRACE_SCHED])
      printf("TERMOP maxPages %d minPages %d, numOutTuples %d\n", 
	     root->maxPagesNeeded, root->minPagesNeeded, root->numOutTuples);
    break;
    
  case STORE:
    // process the child of the store operator
    e = initQueryRequirements(((StorePlan*)root)->child,firstTime);
    if (e != OK) return(NOTOK);
    
    // find the number of output tuples of the store
    root->numOutTuples = ((StorePlan*) root)->child->numOutTuples;
    
    // see KLUDGE above
    // set minimum and maximum memory requirements for the store
    root->maxPagesNeeded = STORE_PAGES;
    root->minPagesNeeded = STORE_PAGES;
    
    if (Flags[TRACE_SCHED])
      printf("STORE maxPages %d minPages %d, numOutTuples %d\n", 
	     root->maxPagesNeeded, root->minPagesNeeded, root->numOutTuples);
    break;
    
  case JOIN:
    joinPlan = (JoinPlan*) root;
    
    // call routine recursively for the children of the join node
    e = initQueryRequirements(joinPlan->leftChild, firstTime);
    if (e != OK) return(NOTOK);
    
    e = initQueryRequirements(joinPlan->rightChild, firstTime);
    if (e != OK) return(NOTOK);
    
    // the smaller relation is the inner(leftChild) so swap
    //     the children if needed
    if (joinPlan->leftChild->numOutTuples > joinPlan->rightChild->numOutTuples) {
	qPlan = joinPlan->leftChild;
	joinPlan->leftChild = joinPlan->rightChild;
	joinPlan->rightChild = qPlan;
    }
    
    // Records left/right input pages only if both children are select ops.
    register QueryPlan *lqp, *rqp;
    lqp = joinPlan->leftChild;
    rqp = joinPlan->rightChild;

    //set the number of input and output tuples of the join node
    joinPlan->numOutTuples = 
      (int) floor(joinPlan->rightChild->numOutTuples * joinPlan->selFactor) ;
    numInTuples = joinPlan->leftChild->numOutTuples ;
    
    // find the total number of pages needed to fit all of the hash tbl in mem
    optMaxPagesNeeded = joinPlan->optMaxPages();
    
    //set the optimally minimum number of pages needed = sqrt( |R| * F ) + 1
    optMinPagesNeeded = (int) ceil(sqrt(optMaxPagesNeeded)) + 1;
    
    if (Flags[TRACE_SCHED]) 
      printf("optmax %d optmin %d\n", optMaxPagesNeeded, optMinPagesNeeded);
    
    if (optMinPagesNeeded > BufSize) {
      printf("ERROR: Join min. memory (%d) > BufSize (%d),",
	     optMinPagesNeeded, BufSize);
      printf(" Query rejected\n");
      return(NOTOK);
    }
    
    // min and max pages needed is set only the first time this 
    //   procedure is called. the second time only the number of 
    //   buckets is calculated
    if (firstTime) {
	
        // set the min mem needed by the join
	if (joinPlan->minMemUnits == MINREQ_UNITS)
	  joinPlan->minPagesNeeded = 
	    (int) ceil(optMinPagesNeeded * joinPlan->minMemNeeded);

	else if (joinPlan->minMemUnits == MAXREQ_UNITS)
	  joinPlan->minPagesNeeded = 
	    (int) ceil(optMaxPagesNeeded * joinPlan->minMemNeeded);

	else if (joinPlan->minMemUnits == KBYTE_UNITS)
	  joinPlan->minPagesNeeded = 
	    (int) ceil(joinPlan->minMemNeeded * 1024 / PageSize);

	else // BUFPOOL_UNITS
	  joinPlan->minPagesNeeded = 
	    (int) ceil(BufSize * joinPlan->minMemNeeded);
	
	// set the max mem needed by the join
	  
	if (joinPlan->maxMemUnits == MINREQ_UNITS)
	  joinPlan->maxPagesNeeded = 
	    (int) ceil(optMinPagesNeeded *  joinPlan->maxMemNeeded);

	else if (joinPlan->maxMemUnits == MAXREQ_UNITS)
	  joinPlan->maxPagesNeeded = 
	    (int) ceil(optMaxPagesNeeded *  joinPlan->maxMemNeeded);

	else if (joinPlan->maxMemUnits == KBYTE_UNITS)
	  joinPlan->maxPagesNeeded = 
	    (int) ceil(joinPlan->maxMemNeeded * 1024 / PageSize);

	else // BUFPOOL_UNITS
	  joinPlan->maxPagesNeeded = 
	    (int) ceil(BufSize * joinPlan->maxMemNeeded);
	
	// adjust min/max if necessary and sanity check em both

	if (joinPlan->maxPagesNeeded > optMaxPagesNeeded)
	  joinPlan->maxPagesNeeded = optMaxPagesNeeded;
	
	if (joinPlan->maxPagesNeeded > BufSize)
	  joinPlan->maxPagesNeeded = BufSize;
	
	if (joinPlan->maxPagesNeeded < optMinPagesNeeded) {
	  printf("ERROR: max mem (%d) < min needed (%d)\n",
		 joinPlan->maxPagesNeeded, optMinPagesNeeded);
	  printf("Query rejected");
	  exit(NOTOK);
	}
	
	if (joinPlan->minPagesNeeded < optMinPagesNeeded) {
	  if (Flags[TRACE_SCHED]) {
	    printf("\nLess min. memory (%d) than needed (%d).", 
		   joinPlan->minPagesNeeded, optMinPagesNeeded);
	    printf("set to min needed\n\n");
	  }
	  joinPlan->minPagesNeeded = optMinPagesNeeded;
	}
	else if (joinPlan->minPagesNeeded < 3) {
	  printf("\nLess memory (%d) than needed (%d).", 
		 joinPlan->minPagesNeeded, optMinPagesNeeded);
	  printf(" Join needs at least 3 pages\n\n");
	  exit(NOTOK);
	} else 
	  if (joinPlan->minPagesNeeded > joinPlan->maxPagesNeeded) {
	    if (Flags[TRACE_SCHED]) {
	      printf("\nMore min memory given (%d) than max (%d)",
		     joinPlan->minPagesNeeded, joinPlan->maxPagesNeeded);
	      printf(" => min set to (%d)\n", 
		     joinPlan->maxPagesNeeded );
	    }
	    joinPlan->minPagesNeeded = joinPlan->maxPagesNeeded;
	  }
      
	// If the user explicitly indicated a min and/or max MPL, then uses
	// the current number of queries of this class running at min or max
	// to determine the allocation for this query.
	
	register JobClass* jc = joinPlan->jobClass;
	register JoinPlan* jp = joinPlan;
	
    } // if (firstTime)
    
    //set the number of buckets to be used for this join
	
    //  if optimal pages > max pages allocated
    //        multi bucket join so calculate the number of buckets
    //  else (* all hash table fits in memory*)
    //       so number of bucket is 1
    //
    if (optMaxPagesNeeded > joinPlan->maxPagesNeeded)

      // for a multi-bucket join, reserve 1 extra page for
      //   each extra bucket

      if (joinPlan->maxPagesNeeded >= 3)  {
	  Memory  = joinPlan->maxPagesNeeded;
	      
	  //       |R| * F - M
	  //  B = ---------------
	  //          M -  1
	  Buckets = 
	    (int) ceil((1.0 * (optMaxPagesNeeded - Memory)) / 
		       (1.0 * ( Memory - 1)));
	  
	  // formula assumes bucket 0 to bucket B
	  //    here we start from 1 onwards

	  joinPlan->numBuckets = 1 + Buckets;
	  joinPlan->bucketKMemory = 
	    (int) ceil((1.0 * optMaxPagesNeeded - Memory + Buckets) / 
		       (1.0 * Buckets));
	  
	  // weight factor = extra entries to be added to split tbl
	  //                 to send more tuples to bucket 0
	  //  = (bucket1 mem / mem for other buckets)-1
	  //           (M - B)
	  //  = ---------------------    - 1
	  //    ((|R| * F - (M - B))/B)

	  joinPlan->weightFactor = 
	    ((1.0 * (Memory - Buckets) * Buckets)
	     / (1.0 * optMaxPagesNeeded - Memory + Buckets)) - 1;
	  
	  // safeguard against any strange floating point calculation error
	  if (joinPlan->weightFactor < 0.0) joinPlan->weightFactor = 0.0;
	  
	  if (Flags[TRACE_SCHED])
	    printf("weightFactor %f opt %d memory %d bucketk %d buckets %d\n", 
		   joinPlan->weightFactor, optMaxPagesNeeded, Memory, 
		   joinPlan->bucketKMemory, Buckets);
      } else {
	  printf("\n\nERROR:multi-bucket join, at least 3 pages");
	  printf("required in memory\n");
	  joinPlan->print(5);
	  exit(NOTOK);
      }      
    else {
	// Requested maximum is great enuf to fit entire hash table in mem
	joinPlan->numBuckets = 1;
	if (Flags[TRACE_SCHED]) printf("single bucket join\n");
    }
    
    if (Flags[TRACE_SCHED]){
	printf("JOIN maxPgs %d minPgs %d numOutTups %d numBkts %d\n",
	       root->maxPagesNeeded, root->minPagesNeeded, 
	       root->numOutTuples, joinPlan->numBuckets);
    }
    break;
    
  default:
    printf("\n\nERROR: unknown operator type %d\n", root->opCode);
    exit(NOTOK);
  }

  return(OK);
}



///////////////////////////////////////////////////////////////////////////////
//
//calculate the total resources required to run the query. flag an error if 
//    more than buffer size required

int initResourceAllocation(QueryMsg* reqMsg, Schedule* schedule, 
		       Schedule* waitSchedule) {
  int nodeNo;

  Schedule *scheduleRow, *waitScheduleRow;
  Operator *opr;
  RootPlan* root;
  
  // set root of the query plan tree
  root = ((RootPlan*) reqMsg->query);
  
  // initialize the totalMinPagesNeeded and totalMaxPagesNeeded arrays to 0
  for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++) {
    root->totalMinPagesNeeded[nodeNo] = 0;
    root->totalMaxPagesNeeded[nodeNo] = 0;
  }
  
  // find the maximum and minimum pages needed per node to run the query
  //   to find this both schedule and waiting schedule are used. schedule
  //   is used to find the new operators being launched. their memory is 
  //   added to the total memory needed. waiting schedule tells us what 
  //   operators are completing and so their memory is removed from the 
  //   total requirement
    
  // initialize the minPagesNeeded and maxPagesNeeded arrays to 0
  for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
    minPagesNeeded[nodeNo] = maxPagesNeeded[nodeNo] = 0;
  
  // for each row in the schedule find the max memory needed for each node
  for(scheduleRow = schedule, waitScheduleRow = waitSchedule; 
      scheduleRow != NULL, waitScheduleRow != NULL; 
      scheduleRow = scheduleRow->nextRow, 
      waitScheduleRow = waitScheduleRow->nextRow) {
    
    // find the total memory needed per site for this scheduler row 

    for(opr = scheduleRow->oprRow; opr != NULL; opr = opr->nextOpr)
      switch(opr->oprPlan->opCode) {
      case SELECT:
      case STORE:
	for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
	  if (opr->oprPlan->activeSites[nodeNo] == TRUE) {
	    minPagesNeeded[nodeNo] += opr->oprPlan->minPagesNeeded;
	    maxPagesNeeded[nodeNo] +=  opr->oprPlan->maxPagesNeeded;
	  }
	break;
      case JOIN:
	// add joins memory if the operator seen for the first 
	// time i.e. for the build phase of the first bucket
	if ((opr->phase == BUILD) && (opr->bucketNum = 1))
	  for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
	    if (opr->oprPlan->activeSites[nodeNo] == TRUE) {
		minPagesNeeded[nodeNo] += opr->oprPlan->minPagesNeeded;
		maxPagesNeeded[nodeNo] += opr->oprPlan->maxPagesNeeded;
	    }
	break;
      case TERMOP:
	break;
      }
    
    // if for any site max(min)PagesNeeded for this row is greater
    //     than the maximum(minimum) values seen so far in the schedule
    //     reset totalMaxPagesNeeded(totalMinPagesNeeded)
      
    for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++) {
      if (minPagesNeeded[nodeNo] > root->totalMinPagesNeeded[nodeNo])
	root->totalMinPagesNeeded[nodeNo] = minPagesNeeded[nodeNo];
      if (maxPagesNeeded[nodeNo] > root->totalMaxPagesNeeded[nodeNo])
	root->totalMaxPagesNeeded[nodeNo] = maxPagesNeeded[nodeNo];
    }
    
    // if the operator is to complete remove the memory from minPagesNeeded
    //   and maxPagesNeeded.
    // if the operator is not completing now, its memory should stay as a
    //     part of the aggregate memory requirement per node
      
    for(opr = waitScheduleRow->oprRow; opr != NULL; opr = opr->nextOpr)
      switch(opr->oprPlan->opCode) {
      case SELECT:
      case STORE:
	for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
	  if (opr->oprPlan->activeSites[nodeNo] == TRUE) {
	    minPagesNeeded[nodeNo] -= opr->oprPlan->minPagesNeeded;
	    maxPagesNeeded[nodeNo] -= opr->oprPlan->maxPagesNeeded;
	  }
	break;
      case JOIN:
	// subtracts joins memory if the operator seen for the last time
	//    i.e. for the probe phase of the last bucket
	if ((opr->phase == PROBE) && 
	    (opr->bucketNum == ((JoinPlan*)opr->oprPlan)->numBuckets))
	  for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
	    if (opr->oprPlan->activeSites[nodeNo] == TRUE) {
		minPagesNeeded[nodeNo] -= opr->oprPlan->minPagesNeeded;
		maxPagesNeeded[nodeNo] -= opr->oprPlan->maxPagesNeeded;
	    }
	break;
      case TERMOP:
	break;
      }
  }
  //
  // if the minimum memory requirement of the query is greater than the 
  //     total buffer size then discard the query
  //
  for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++) {
      if (root->totalMinPagesNeeded[nodeNo] > BufSize) {
	  printf("\n\nMin. mem requirement of query (%d) > total buffer",
		 root->totalMinPagesNeeded[nodeNo]);
	  printf("size (%d) for site %d\n\n", BufSize, nodeNo);
	  return(NOTOK);
      }
  }
  return(OK);  
}


//////////////////////////////////////////////////////////////////////////
//
// check if the query can be executed given the present BufPages
//
//////////////////////////////////////////////////////////////////////////
int CanRun(RootPlan* query)
{
    int nodeNo;

    // for each site
    for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++) {
	// if available buffers less than minimum needed
	//        send the query to the GlobalQueryWaitQueue
	if (BufPages[nodeNo] < query->totalMinPagesNeeded[nodeNo]) {
	    if (Flags[TRACE_SCHED]) {
		printf("Not enough pages on site %d. needed %d, avail %d\n", 
		       nodeNo, query->totalMinPagesNeeded[nodeNo], 
		       BufPages[nodeNo]);
	    }
	    return(NOTOK);
	}
    }
    return(OK);
}

///////////////////////////////////////////////////////////////////////////////
//
// perform actual resource allocation for the query plan using the available 
//     buffers at different sites. 
//     schedule and waitSchedule point to the schedules created using 
//     the optimal memory requirement
//
//     if the available buffers are less than the minimum needed
//         query is sent to the GlobalQueryWaitQueue
//     else
//         if the available buffers are less than the maximum needed
//               query is rescheduled
//         else
//               dispatched for execution as is
//     
//////////////////////////////////////////////////////////////////////////////
int actualResourceAllocation(QueryMsg* reqMsg, Schedule*& schedule, 
			 Schedule*& waitSchedule)
{

  static int
    rowPagesAlloc[MAX_NODES],     // pages allocated to the query per site
    totalMinJoinPages[MAX_NODES]; // total no. of pages needed for 
                                  // concurrent joins per site
  int numJoins, nodeNo, reschedule, joinMemory, e, joinNum ;
  
  Schedule *scheduleRow, *waitScheduleRow;
  Operator *opr;
  
  QueryPlan *concJoins[MAXJOINS], *qPlan;
  RootPlan *root;
  
  root = (RootPlan*) (reqMsg->query);
  
  // for each site
  for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
    // if the available pages less than the maximum pages needed
    //     set pagesAllocated accordingly and mark query for rescheduling
    //   NOTE: as memory available has changed the num of buckets for the
    //          joins may also have changed
    if (BufPages[nodeNo] < root->totalMaxPagesNeeded[nodeNo]) {
	if (Flags[TRACE_SCHED])
	  printf("Less than max. pgs on site %d. needed %d, avail %d\n",
		 nodeNo, root->totalMaxPagesNeeded[nodeNo],
		 BufPages[nodeNo]);
	
	// note the buf pages allocated to the site for this site
	root->totalMaxPagesNeeded[nodeNo] = BufPages[nodeNo];
	
	reschedule = TRUE; // mark the query for rescheduling
    } 

  if (Flags[TRACE_SCHED]) {
      printf("\n\tBufPages left & allocated to query for each site\n");
      for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
	printf("\t\tsite %d BufPages %d allocated %d\n", nodeNo, 
	       BufPages[nodeNo], root->totalMaxPagesNeeded[nodeNo]);
  }

  
  // if query marked for rescheduling
  if (reschedule == TRUE) {
      if (Flags[TRACE_SCHED])
	printf("\nQuery being rescheduled\n\n");
	  
      numJoins = 0 ;
      for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
	totalMinJoinPages[nodeNo] = 0;
	  
      // for every scheduler row
      for(scheduleRow = schedule, waitScheduleRow = waitSchedule; 
	scheduleRow != NULL, waitScheduleRow != NULL; 
	scheduleRow = scheduleRow->nextRow, 
	waitScheduleRow = waitScheduleRow->nextRow) {
	  
	// copy to another array because may change in the code below
	for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
	  rowPagesAlloc[nodeNo] = root->totalMaxPagesNeeded[nodeNo];
	
	// find the number of join operators in the row and take out 
	// the pages needed for select and store from the pages available 
	// per site
	for(opr = scheduleRow->oprRow; opr != NULL; opr = opr->nextOpr)
	  switch(opr->oprPlan->opCode) {
	    case SELECT:
	    case STORE:
	      for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
		if (opr->oprPlan->activeSites[nodeNo] == TRUE)
		  rowPagesAlloc[nodeNo] -= opr->oprPlan->minPagesNeeded;
	      break;
	    case JOIN:
	      // if join seen for the first time
	      if ((opr->phase == BUILD) && (opr->bucketNum = 1)) {
		// add join to list of concurrently running joins
		concJoins[++numJoins] = opr->oprPlan;
		  
		// for every site on which the join runs, add 
		//    min memory needed to total min mem for joins
		for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
		  if (opr->oprPlan->activeSites[nodeNo] == TRUE)
		    totalMinJoinPages[nodeNo] += opr->oprPlan->minPagesNeeded;
	      }
	      break;
	  }
	      
	// KLUDGE*******
	//  divide the total memory in the row among the joins proportional
	//  to the minimum memory requirement of the join
	//  this is an arbitrary choice. other choices include size of the
	//  join, size of the result, size of inner relation etc.
	//
	      
	for(joinNum = 1; joinNum <= numJoins; joinNum++) {
	  qPlan = concJoins[joinNum];
	  for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
	    if (qPlan->activeSites[nodeNo] == TRUE) {
		joinMemory =  
		  (int) (((1.0 * qPlan->minPagesNeeded) / 
			  totalMinJoinPages[nodeNo]) * rowPagesAlloc[nodeNo]);
		
		// if the memory given < maximum needed
		//     change the allocated memory to the new value
		if (joinMemory < qPlan->maxPagesNeeded)
		  qPlan->maxPagesNeeded = joinMemory;
	    }  
	  if (joinMemory < qPlan->minPagesNeeded) {
	      printf("\n\nERROR:join mem (%d) < min req (%d) \n", 
		     joinMemory, qPlan->minPagesNeeded);
	      exit(NOTOK);
	  }
        }
	
	// if a join completes then change list of concurrent joins and
	//  total min. join mem
	for(opr = waitScheduleRow->oprRow; opr != NULL; opr = opr->nextOpr)
	  if ((opr->oprPlan->opCode == JOIN) && 
	      ((opr->phase == PROBE) && 
	       (opr->bucketNum == ((JoinPlan*)opr->oprPlan)->numBuckets))) {

	    // find the join in the concJoin array
	    for(joinNum = 1; joinNum <= numJoins; joinNum++)
	      if (concJoins[joinNum] == opr->oprPlan) break;

	    // shift the subsequent join plans once to the left
	    for (;joinNum <= (numJoins - 1); joinNum++)
	      concJoins[joinNum] = concJoins[joinNum + 1];
	    
	    numJoins--; //reduce the number of joins

	    // reset the total join minimum page requirement
	    for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
	      if (opr->oprPlan->activeSites[nodeNo] == TRUE)
		totalMinJoinPages[nodeNo] -= 
		  opr->oprPlan->minPagesNeeded;
	}
      } // for every scheduler row
	
      // perform the calculation of the memory required by the query again
      //   due to overlap the memory allocated to joins may have changed thus
      //   changing the memory requirement of the query
      if ((e = initResourceAllocation(reqMsg,schedule, waitSchedule)) != OK) 
	return(NOTOK);
      
      // recalculate only the number of buckets for the join
      e = initQueryRequirements(reqMsg->query, FALSE);
      if (e  != OK) return(NOTOK);
  } // if reschedule
  
  // for each site reduce the number of free buffer pages by the amount 
  //    allocated to the query
  for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++) {
      BufPages[nodeNo] -= root->totalMaxPagesNeeded[nodeNo];
      // add allocated memory to total memory used by the class
      // KLUDGE: correct only if query runs on only one node
      root->jobClass->usedMem += root->totalMaxPagesNeeded[nodeNo];
  }
  return(OK);
}


////////////////////////////////////////////////////////////////////////////
//
// form the schedule and the wait schedule given the root of the query plan
//
////////////////////////////////////////////////////////////////////////////
int makeSchedule(RootPlan* root, Schedule*& schedule, Schedule*& waitSchedule)
{

    Schedule   *lastWaitRow = NULL, *insertRow = NULL, *insertWaitRow = NULL;
    Schedule   *lastRow = NULL;

    // call the appropriate scheduler depending on scheduler type
    switch(root->schedType)
    {
      case RIGHTDEEPONE:
          // schedule the query by calling bushyExec
	  bushyExec(root, schedule, waitSchedule, insertRow, insertWaitRow, 
		    lastRow, lastWaitRow, TRUE);
	  break;

      case LEFTDEEP:
      case RIGHTDEEPALL:
      case BUSHYMAX:
          // schedule the query by calling bushyExec 
	  bushyExec(root, schedule, waitSchedule, insertRow, insertWaitRow, 
		    lastRow, lastWaitRow, FALSE);
	  break;
      default:
	  printf("\nERROR: unknown sched type %d\n", root->schedType);
	  exit(NOTOK);
//	  return(NOTOK);
    }
    if (Flags[TRACE_SCHED])
    {
      printf("Sending Schedule\n\n");
      schedule->print();
      printf("\nWaiting Schedule\n\n");
      waitSchedule->print();
      printf("\n");
    }
    return(OK);

}

// add a new first row to schedule
void addFirstRow(Schedule*& schedule, Schedule*& lastRow)
{
  Schedule *newRow = new Schedule();
  newRow->nextRow = schedule;
  schedule = newRow;
  if (lastRow == NULL)
     lastRow = schedule;
}

// add a new last row to the schedule
Schedule* addNewLastRow(Schedule*& schedule, Schedule*& lastRow)
{
    Schedule *newRow = new Schedule();
    newRow->nextRow = NULL;
    if (schedule == NULL)
      schedule = newRow;
    else
        lastRow->nextRow = newRow;
    lastRow = newRow;

    return(newRow);
}

//check if secondRow appears after firstRow in schedule
bool after(Schedule* firstRow, Schedule* secondRow, Schedule*& lastRow)
{
    if ((secondRow == NULL) || (firstRow == lastRow))
        return(FALSE);
    else
    {
        // check rows succeeding first row
        for(Schedule *row = firstRow->nextRow; row != NULL; row = row->nextRow)
	  if (row == secondRow)
	      return(TRUE);
	return(FALSE);
    }
}


// add opr to the schedule row. set bucketNum (default 1) 
//   and phase (ignored if opr not a join)
void addOpr2(Schedule* row, QueryPlan* opr, int bucketNum = 1, int phase = -1)
{
  Operator* newOperator;

  newOperator = new Operator();
  newOperator->oprPlan = opr;
  newOperator->nextOpr = row->oprRow;
  newOperator->bucketNum = bucketNum;
  if (opr->opCode == JOIN)
    newOperator->phase = phase;
  row->oprRow = newOperator;
}



///////////////////////////////////////////////////////////////////////////////
//
//  bushyExec
//   In - 1) currentNode - pointer to the root of the query tree
//        2) schedule - pointer to the schedule
//        3) waitSchedule - pointer to the waitSchedule
//        4) insertRow - ptr to row in schedule where op to be inserted
//        5) lastRow - last row in schedule
//        6) waitInsertRow - ptr to row in waitSchedule where op to be inserted
//        7) lastRow - last row in waitSchedule
//        8) flag - TRUE if right-deep-one schedule to be build (else FALSE)
//                right-deep-one sched -> one scan at a time
//                                        compared to all in parallel
//
//    Action - a recursive routine which traverses the query tree and
//             creates a bushy execution schedule, it generates a
//             bushy tree with the maximal parallelism possible.
//             as lef
//
//////////////////////////////////////////////////////////////////////////////
void bushyExec(QueryPlan* currentNode, Schedule*& schedule, 
	       Schedule*& waitSchedule, Schedule*& insertRow, 
	       Schedule*& lastRow, Schedule*& waitInsertRow, 
	       Schedule*& lastWaitRow, bool flag)
{

  JoinPlan *joinNode;
  Schedule *origInsertRow, *origWaitInsertRow;

  if (currentNode == NULL) return;

  switch(currentNode->opCode)
  {
    case SELECT:
      // if right child of join
      //    add to new last row
      if ((currentNode->parent->opCode == JOIN) &&
  	  (((JoinPlan*) currentNode->parent)->rightChild == currentNode))
      {
	   insertRow = addNewLastRow(schedule, lastRow);
	   waitInsertRow = addNewLastRow(waitSchedule, lastWaitRow);
      }
      else  // left child of join or child of store/termop
      {
	  // if right-deep-one schedule form new row for select
	  if (flag)
	  {
	      addFirstRow(schedule, lastRow);
	      addFirstRow(waitSchedule, lastWaitRow);
	  }
	  else // add to first row
	  {
	      if (schedule == NULL)
		  addFirstRow(schedule, lastRow);
	      if (waitSchedule == NULL)
		  addFirstRow(waitSchedule, lastWaitRow);
	  }
	  insertRow = schedule;
	  waitInsertRow = waitSchedule;
      }
      addOpr2(insertRow, currentNode);
      addOpr2(waitInsertRow, currentNode);
      break;

    case STORE:
      bushyExec(((StorePlan*) currentNode)->child, schedule, waitSchedule,
		insertRow, lastRow, waitInsertRow, lastWaitRow, flag);
      addOpr2(insertRow, currentNode);
      addOpr2(waitInsertRow, currentNode);
      break;

    case TERMOP:
      bushyExec(((TermPlan*) currentNode)->child, schedule, waitSchedule,
		insertRow, lastRow, waitInsertRow, lastWaitRow, flag);
      addOpr2(insertRow, currentNode);
      addOpr2(waitInsertRow, currentNode);
      break;

    case JOIN:
      joinNode = (JoinPlan*) currentNode;

      // make a copy as level may change below
      origInsertRow = insertRow;
      origWaitInsertRow = waitInsertRow;

      // process the left child of join
      bushyExec(joinNode->leftChild, schedule, waitSchedule, insertRow, 
		lastRow, waitInsertRow, lastWaitRow, flag);

      // add join node (BUILD) to both schedules
      addOpr2(insertRow, joinNode, 1, BUILD);
      addOpr2(waitInsertRow, joinNode, 1, BUILD);

      // if new rows before original rows in schedule, restore rows
      if (after(insertRow, origInsertRow, lastRow )) 
	  insertRow = origInsertRow;
      if (after(waitInsertRow,origWaitInsertRow, lastWaitRow)) 
	  waitInsertRow = origWaitInsertRow;

      // process right child of join
      bushyExec(joinNode->rightChild, schedule, waitSchedule, insertRow, 
		lastRow, waitInsertRow, lastWaitRow, flag);

      // add join node (PROBE) to both schedules
      addOpr2(insertRow, joinNode, 1, PROBE);
      addOpr2(waitInsertRow, joinNode, 1, PROBE);

      waitInsertRow = lastWaitRow;

      break;
  }
}
