#include "dbsim.h"
#include "object.h"
#include "msg.h"
#include "mailbox.h"
#include "query.h"
#include "sched.h"
#include "language.h"

/////////////////////////////////////////////////////////////////////////////
//
//  scheduler.c
//
//     contains methods for all the classes declared in sched.h 
//               (scheduler, operator, queryqueue)
//
/////////////////////////////////////////////////////////////////////////////

extern "C" {
  void exit(int);
}

extern int BufPages[];
extern allocateResources(QueryMsg*, Schedule*&, Schedule*&);
extern void QueryScheduler(Node*, QueryMsg*, Schedule*, Schedule*);
extern CanRun(RootPlan*);

int Schedule::numSched = 0;
int QueryReq::numQReq = 0;
int Operator::numOper = 0;

Schedule* Schedule::freeSchedule = NULL;	// no free Schedules initially
Operator* Operator::freeOperator = NULL;	// no free operators initially
QueryReq* QueryReq::freeQueryReq = NULL;	// no free QueryReqs initially

////////////////////////////////////////////////////////////////////////

void Dqtable::note_entry() {
   qtable::note_entry();
   if ((clock < delay) || (statWaitExpired)) return;
   
   int curLen = qtable::cur();
   for (int f = 0; f < curLen; f++) qtable::note_exit();
   qtable::reset();
   for (f = 0; f < curLen; f++) qtable::note_entry();
   statWaitExpired = TRUE;
 }

////////////////////////////////////////////////////////////////////////

void Dqtable::note_exit() {
   qtable::note_exit();
   if ((clock < delay) || (statWaitExpired)) return;

   int curLen = qtable::cur();
   for (int f = 0; f < curLen; f++) qtable::note_exit();
   qtable::reset();
   for (f = 0; f < curLen; f++) qtable::note_entry();
   statWaitExpired = TRUE;
 }

////////////////////////////////////////////////////////////////////////
//
// Schedule class desctructor - destroy all operators in the row
//
////////////////////////////////////////////////////////////////////////
Schedule::~Schedule() {
    Operator *opr, *startOpr;

    startOpr = this->oprRow;
    while ( startOpr != NULL)
    {
        opr = startOpr;
	startOpr = startOpr->nextOpr;
	delete opr;
    }
}

////////////////////////////////////////////////////////////////////////
//
// iteratively deletes all the rows in schedule
//
////////////////////////////////////////////////////////////////////////
void DeleteSched(Schedule*& schedule)
{

    Schedule *startRow, *row;

    // for each operator row delete all the query plans 
    startRow = schedule;
    while (startRow != NULL)
    {
        row = startRow;
	startRow = startRow->nextRow;
	delete row;
    }

}

void Operator::print()
{
    switch(oprPlan->opCode)
    {
        case JOIN:
            if (phase == BUILD)
	      printf(" BUILD");
	    else
	      if (phase == PROBE)
		printf(" PROBE");
	    break;
	case SELECT:
	    printf(" SELECT");
	    break;
	case STORE:
	    printf(" STORE");
	case TERMOP:
	    printf(" TERMOP");
	    break;
    }
    printf("(%d)", bucketNum);
    if (nextOpr != NULL) nextOpr->print();
}



void Schedule::print()
{
    printf("\t");
    oprRow->print();
    printf("\n");
    if (nextRow != NULL) nextRow->print();
}

//
// add a new query to a query queue
//
void QueryQueue::enqueue(Node *node, QueryMsg *reqMsg, 
			 Schedule* schedule, Schedule* waitSchedule)
{
    QueryReq *newReq;
    
    newReq = new QueryReq();
    newReq->node = node;
    newReq->reqMsg = reqMsg;
    newReq->schedule = schedule;
    newReq->waitSchedule = waitSchedule;
    for(int nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
    {
	if (Flags[TRACE_SCHED])
	    printf("site %d minPages %d\n", 
 	         nodeNo, ((RootPlan*) reqMsg->query)->totalMinPagesNeeded[nodeNo]);
    }
    newReq->nextReq = NULL;

    if (qhead == NULL) {
	// queue is empty so just stick the new request at the front
	qhead = newReq;
	qtail = newReq;
    } else {
        // queue non-empty so add the new request to the tail
	qtail->nextReq = newReq;
	qtail = newReq;
    }

    queryQueueStat->note_entry();

}


// remove query from the queryqueue
QueryReq* QueryQueue::dequeue(QueryReq*& query)
{
    QueryReq* prev;

    if (query == NULL) query = qhead;

    if (query == qhead)
    {
          qhead = query->nextReq;
	  prev = NULL;
    }
    else
    {
         // query in middle of queue. find prev node to adjust pointers
         for(prev = qhead; prev->nextReq != query; prev = prev->nextReq);
         prev->nextReq = query->nextReq;
    }
    if (query == qtail) qtail = prev;

    queryQueueStat->note_exit();
    return(query);
};


// find next query from the queryqueue
QueryReq* QueryQueue::findNext()
{
    QueryReq* qReq = qhead;

    while (qReq != NULL) {
        if ((qReq->reqMsg->query->jobClass->currentMPL < 
	     qReq->reqMsg->query->jobClass->maxMPL) &&
	    (((RootPlan*) (qReq->reqMsg->query))->randPlan->currentMPL <
	     ((RootPlan*) (qReq->reqMsg->query))->randPlan->maxMPL)) {
            return(dequeue(qReq));
	} else {
	   qReq = qReq->nextReq;
        }
    }
    return(NULL);
}


// print contents of queryqueue
void QueryQueue::print()
{
    QueryReq *query;
    int n;

    for(query = qhead, n = 0; query != NULL; query = query->nextReq, n++)
    {
        printf("\n\t\tQuery #%d\n", n);
        query->reqMsg->query->print(5);
    }
}

// print out the query queue statistics
void QueryQueue::printStats()
{
  printf("\n\tTotal\t\t%d\n\tMeanLength\t%5.3f\n\tMaxLength\t%d\n\tAvgWaitingTime\t%5.3f secs\n", 
	 queryQueueStat->cnt(), 	 queryQueueStat->qlen(), 
	 queryQueueStat->max(), 	 (queryQueueStat->qtime() / 1000.0) );
}




// add a new query to a query queue, update job class table also
void MemoryQueue::enqueue(Node *node, QueryMsg *reqMsg, 
		 	  Schedule* schedule, Schedule* waitSchedule)
{
  reqMsg->query->jobClass->memoryWaitTable->note_entry();

  // call the enqueue method on the parent class QueryQueue
  QueryQueue::enqueue(node, reqMsg, schedule, waitSchedule);
}

// extract first request from the head of the queue, update job class
//       table also
QueryReq* MemoryQueue::dequeue(QueryReq*& query)
{
    QueryReq* nextReq = QueryQueue::dequeue(query);

    //record the exit from the queue on the job class memory queue
    nextReq->reqMsg->query->jobClass->memoryWaitTable->note_exit();

    return(nextReq);
}


// check if any site clash present in the two memory arrays
bool MemoryQueue::siteClash(int queryPagesNeeded[MAX_NODES],
			    int currentPagesNeeded[MAX_NODES])
{
      int siteNum;

      for(siteNum = 0; siteNum < NumDiskNodes; siteNum++)
	    if ((queryPagesNeeded[siteNum] > 0) &&
		(currentPagesNeeded[siteNum] > 0))
	    {
	        // waiting quere and current query want pages on siteNum
	        if (Flags[TRACE_SCHED])
		  printf("Memory clash found: site %d Query waits\n", siteNum);

	        return TRUE;
	    }
      return FALSE;
}


// check if any of the waiting queries can run given the present number
//     of pages in BufPages and fork off a query scheduler if it can
//     the query does not run if it has a site conflict with any of the
//     waiting queries
void MemoryQueue::checkQueue()
{
    QueryReq *nextQuery, *query, *waiter;
    bool clash = FALSE;
    int e;

    // if queue is nonempty
    if (this->nonempty())
    {
         if (Flags[TRACE_SCHED])
	     printf("Waiting Queries in Query Queue\n");

        // for each waiting query check if query can run
        for(query = qhead; query != NULL; query = query->nextReq) {
  	    // if the head of the queryqueue can run now and no site clash
	    //    with other waiters run the query
	    if  (CanRun((RootPlan*) (query->reqMsg->query))) {
	        clash = FALSE;
		// check for site clashes with other waiters
	        for(waiter = qhead; waiter != query; waiter = waiter->nextReq)
		   if (siteClash(((RootPlan*) waiter->reqMsg->query)->totalMinPagesNeeded, 
				 ((RootPlan*) query->reqMsg->query)->totalMinPagesNeeded))
		      clash = TRUE;

		// if no site conflict found
		if (clash == FALSE)
		{
	           nextQuery = this->dequeue(query);
		   // form the schedules. also does the resource allocation
		   e = allocateResources(nextQuery->reqMsg, 
			    nextQuery->schedule, nextQuery->waitSchedule);
		
		   if (e == OK)
		       // fork off a new query scheduler to handle the query
		       QueryScheduler(nextQuery->node, nextQuery->reqMsg, 
			      nextQuery->schedule, nextQuery->waitSchedule);
		   else
		     // report scheduling error to the user
		     printf("*ERROR in scheduling query. execution aborted\n");
	           
		   delete nextQuery;
	           if (Flags[TRACE_SCHED])
	             printf("Next Waiting Query in GlobalQ can run now\n");
		   return; // exit loop after first such query found
	        }
	    }
	}
    }
}


// check if any of the queries waiting in the queue have a site clash with
//   the site requirement of the new query. return TRUE if clash present
bool MemoryQueue::checkWaiters(int currentPagesNeeded[MAX_NODES])
{
    QueryReq *query;

    // for each waiting query check for site clashes
    for(query = qhead; query != NULL; query = query->nextReq)
        if (siteClash(((RootPlan*) query->reqMsg->query)->totalMinPagesNeeded, 
		      currentPagesNeeded))
	    return TRUE;
    return FALSE;
}


// find next query from an mpl queue. checks if readyMPL < maxMPL
QueryReq* MPLQueue::findNext()
{
    QueryReq* qReq = qhead;
    JobClass* jclass;

    if (qReq != NULL) {
	jclass = qReq->reqMsg->query->jobClass;
	if (jclass->readyMPL < jclass->maxMPL) {
//	    if (Flags[TRACE_LOCK]) 
//	        printf("next query in mpl queue started\n");
            return(dequeue(qReq));  
	}
    }
	
    return(NULL);
}

