#include <limits.h>
#include <dbsim.h>
#include <object.h>
#include <node.h>
#include <msg.h>
#include <mailbox.h>
#include <terminal.h>
#include <split.h>
#include <op.h>
#include <query.h>
#include <language.h>
#include <db.h>
#include <disk.h>
#include <sched.h>
#include <bufmgr.h>
#include <cost.h>
#include <string.h>

///////////////////////////////////////////////////////////////////////////////
//
//  Name:        sim.c
//
//  Title:       UW Database Simulator controller
//
//  Function:    Instantiates sim objects, starts primary sim
//               processes, writes out simulation results, and terminates.
//                 
//  Change Activity:
//
//     08/25/91 (KPB) - Initial version.
//
///////////////////////////////////////////////////////////////////////////////

// Hardware/sofware configuration global variables

int        ActiveTerms = 0,
           Inst_StartIO = INST_STARTIO, // instructions to start an I/O
           NumNodes = DEFAULT_NUMNODES,
           NumDiskNodes = DEFAULT_NUMNODES-1,
           DisksPerNode = DEFAULT_DISKSPERNODE,
           CPUsPerNode =  DEFAULT_CPUSPERNODE,
           KBytesPerNode = DEFAULT_KBYTESPERNODE, // 1K = 1024 bytes (duh!)
           BufSize, // calculated at run time from other config vars
           PagesPerCyl = DEFAULT_PAGESPERCYL,
           PageSize = DEFAULT_PAGESIZE,
           DiskCacheContexts = DEFAULT_DISKCACHECONTEXTS, // disabled if = zero
           DiskCachePages = DEFAULT_DISKCACHEPAGES, // disabled if = zero
           MaxPacketBytes = DEFAULT_MAXPACKETBYTES,
           MaxRequests = 0, // infinite server
           GlobKBytes = DEFAULT_GLOBKBYTES,
           LocKBytes = DEFAULT_LOCKBYTES,
           CurTempFileNum = MAX_FILES;
float      SeekFactor = DEFAULT_SEEKFACTOR,
           MinLatency = DEFAULT_MINLATENCY,
           DiskRotation = DEFAULT_DISKROTATION,
           SettleTime = DEFAULT_SETTLETIME,
           XferRate = DEFAULT_XFERRATE,
           XferTime, // calculated at run time from other config vars
           Mips =  DEFAULT_MIPS,
           MinDelay = DEFAULT_MINDELAY,
           DelayPerByte = DEFAULT_DELAYPERBYTE,
           MinSendProtocol = DEFAULT_MINSENDPROTOCOL,
           SendProtocolPerByte = DEFAULT_SENDPROTOCOLPERBYTE,
           MinRecvProtocol = DEFAULT_MINRECVPROTOCOL,
           RecvProtocolPerByte = DEFAULT_RECVPROTOCOLPERBYTE,
           MinShortInstr = DEFAULT_MINSHORTINSTR,
           ShortInstrPerByte = DEFAULT_SHORTINSTRPERBYTE;

// Simulated objects that need to be globally accessible

Mailbox*   GlobalSchedPort;
Node       *Nodes[MAX_NODES] = {NULL};
RandomFile *ConstFiles[MAX_FILES] = {NULL};
File       *Relations[MAX_FILES] = {NULL};
Index      *Indices[MAX_FILES] = {NULL};
Terminal   *Terms[MAX_TERMINALS] = {NULL}; 
JobClass   *JobClasses[MAX_CLASSES] = {NULL};
facility   *Network;
MemoryQueue *GlobalMemoryWaitQueue ;
QueryQueue  *GlobalMPLWaitQueue ;
int        BufPages[MAX_NODES] = {0};
storage    *GlobMsgPool[MAX_NODES];

// Global simulator control data

bool       Flags[MAXFLAGS] = {FALSE},
           CSIMReports = FALSE,
           PlanStats = FALSE,
           ResetBufferFlag = 0,     // to clear buffer pool after each query
           PrimaryClassNeeded = FALSE,
           PrimaryClassFound = FALSE,
	   PlotMode = FALSE;
event      SimDone("SimDone");
int	   SimRunning,
           MsgLevel = 0,
           NumClasses = 0,
           TermIDCounter = 0,
	   MaxMPL = 0;              // maintained by scheduler
double     TimeWeightedMPL = 0.0,  // maintained by scheduler
           FirstMPLEntry = 0.0;
float      TotNWBytesSent = 0.0,
           TotNWBytesShort = 0.0,
           TotSends = 0.0,
           TotBlockedSends = 0.0,
	   GlobalStatWaitSecs = 0.0;

int 
      // total no. of minimum pages for operators on one site per scheduler row
      minPagesNeeded[MAX_NODES],
      
      // total no. of maximum pages for operators on one site per scheduler row
      maxPagesNeeded[MAX_NODES];

// Global vars used to communicate with SetFlags:

float      PingInterval = 0.0, SimTime = 0.0;  // default is run forever

char       *ConfigFileName = NULL, 
           *SchemaFileName = NULL, 
           *WkldFileName = NULL,
           *SourceFileName = NULL,
           *SimTimeChars,
           BufMgrType = '3',     // default is 3 level LRU policy
           DiskType = 'c';       // default is CareyJitsu disk

// Global vars used to communicate with parser/lex:

extern FILE* yyin;        // parser/lex input file
extern int yynerrs;       // parser/lex error count
extern int yylineno;      // lex current line number
SymTable   RealSyms;      // real num symbol table
SymTable   FileSyms;      // file symbol table
SymTable   PlanSyms;      // plan symbol table
SymTable   JobClassSyms;  // job class symbol table
bool       OpenClasses;   // one or more open classes exist
int*       AllDiskIDs;    // array of one integer per DisksPerNode

// Function prototypes

void      SetFlags(int argc, char** argv);
void      SetCSIMLimits();
bool      OpenFiles(FILE* &configFile, FILE* &schemaFile, FILE* &wkldFile);
void      OutputConfig();
void      OutputSchema();
void      OutputWorkload();
void      OutputResults();
void      globalScheduler(Node *n);
int       yyparse();
void      WritePing();
void      printBufStats();
void      PrintHelp();

///////////////////////////////////////////////////////////////////////////////

//extern "C" int sim(int argc, char** argv) { kbaix
extern "C" void sim(int argc, char** argv) {
  int i, waitStat;
  FILE *configFile = NULL, *schemaFile = NULL, *wkldFile = NULL;
  float holdTime;

  create("sim");           // defines itself as "distinguished" CSIM process

  // Sets global flags from the command line.

  SetFlags(argc, argv);
    
  // Verifies language input files were specified and open up OK
    
//  if (!OpenFiles(configFile, schemaFile, wkldFile)) return(0); kbaix
  if (!OpenFiles(configFile, schemaFile, wkldFile)) return;

  // Parses configuration definition file, possibly resetting config variables

  yyin = configFile;
  SourceFileName = ConfigFileName;
  yylineno = 1;
  yyparse();
  if (Flags[TRACE_PARSER]) printf("config:  yynerrs = %d\n", yynerrs);
//  if (yynerrs) return(0); kbaix
  if (yynerrs) return;

  // Calculates any derived configuration constants

  BufSize = (int) ((KBytesPerNode * 1024.0) / PageSize); 
  XferTime = ((float) PageSize)/(XferRate * 1048.6);
  if (GlobKBytes > 0) {
      for (i = 0; i < NumNodes; i++)
	  GlobMsgPool[i] = new storage ("GlobMsgPool", GlobKBytes * 1024);
  } else
      for (i = 0; i < NumNodes; i++)
	GlobMsgPool[i] = NULL;

  // If a global message pool was specified, make sure that each process has
  // at least one private message buffer, or deadlock could occur.  This is
  // because the global pool is a shared resource and can be overrun by a 
  // fast process.  Local pools cannot be overrun, and will always free up
  // eventually.

  if (GlobKBytes != 0) {
      LocKBytes = Max(LocKBytes, (int)(ceil(DEFAULT_MAXPACKETBYTES / 1024.0)));
  }

  OutputConfig();

  // Given the configuration at this point, resets the various CSIM limits.

  SetCSIMLimits();

  // Performs any final initialization needed before cranking up nodes

  if (MaxRequests == 0) {
    Network = new facility("network"); // instantiates network
    Network->set_servicefunc(inf_srv);
  } else {
    Network = new facility_ms("network", MaxRequests); // instantiates network
  }

  SimDone.clear();                   // enables this process to exit gracefully
  SimRunning = TRUE;                 // enables polling processes

  // Instatiates nodes, each will start its own processes

  for (i = 0; i < NumNodes; i++) Nodes[i] = new Node(i);

  // initialize the buffer array to be used by the scheduler
  for(i = 0; i < NumNodes; i++) {
      BufPages[i] = BufSize;
  }

  // Build a list of disk IDs, one per DisksPerNode (used in parser to 
  // decluster files across all disks as default).

  AllDiskIDs = new int[DisksPerNode];
  for (i = 0; i < DisksPerNode; i++) AllDiskIDs[i] = i;

  // Parses the database definition file

  yyin = schemaFile;
  SourceFileName = SchemaFileName;
  yylineno = 1;
  yyparse();
  if (Flags[TRACE_PARSER]) printf("schema:  yynerrs = %d\n", yynerrs);
//  if (yynerrs) return(0); kbaix
  if (yynerrs) return;
  OutputSchema();

  // Instantiates the buffer managers on each node.

  for (i = 0; i < NumNodes; i++) Nodes[i]->startBufMgr();

  // Parses the workload definition file

  yyin = wkldFile;
  SourceFileName = WkldFileName;
  yylineno = 1;
  yyparse();
  if (Flags[TRACE_PARSER]) printf("wkload:  yynerrs = %d\n", yynerrs);
//  if (yynerrs) return(0); kbaix
  if (yynerrs) return;
  OutputWorkload();

  // Resets CSIM limits again based on additional info in workload file

  SetCSIMLimits();

  // Starts a global query scheduler process on node "NumNodes"

  if (Flags[TRACE_CONTROL]) 
    printf("starting the global scheduler node on node %d\n", NumNodes - 1);
  if (Flags[TRACE_CONTROL]) 
    printf("queries run on nodes 0 to %d\n",NumNodes-2);

  GlobalSchedPort = new Mailbox("GlobalSchedPort", Nodes[NumNodes-1]);
  globalScheduler(Nodes[NumNodes-1]);

  // instantiates a global query queue to be used by the scheduler
  GlobalMemoryWaitQueue = new MemoryQueue(GlobalStatWaitSecs);
  GlobalMPLWaitQueue = new QueryQueue(GlobalStatWaitSecs);

  // Instantiates/starts up terminal objects/processes for each job class

  for (i = 0; i < NumClasses; i++) JobClasses[i]->start();

  // Waits for the simulation to finish, writes simulation output and exits.
  // Writes out periodic progress msgs if -p flag was specified.

  if (SimTime != 0.0) {                  // if user specified a simulation time
    fprintf(stderr, "\n%s:  Specified simulation time:  %s (%.0f msecs)\n", 
	   SIMNAME, SimTimeChars, SimTime);

    if (PingInterval)		         // user wants periodic progress msgs
      do {
	if (simtime()+PingInterval > SimTime){ // should we quit b4 next ping?
	  holdTime = SimTime - simtime();
	  if (holdTime < 0.0) holdTime = 0.0; // cant be too careful!
        }				      
	else                             // can fit next ping in b4 sim done
	  holdTime = PingInterval;
	hold(holdTime);                  // wait till next interval
	WritePing();
      } while (simtime() < SimTime);     // keep pinging until time up
	
    else                                 // no ping msgs, just wait quietly
      hold(SimTime);

    fprintf(stderr, "\n%s:  Simulation time expired\n\n", SIMNAME);
    SimRunning = FALSE;                  // disables polling processes 

  // No simulation time parm.  Stops with cntl C or all terminals having finite
  // plan sequences, or a "primary" job class exists.

  } else {
    if ((PrimaryClassNeeded) && (!PrimaryClassFound)) {
	printf("ERROR: Random Class in workload but no primary class\n");
	SimRunning = FALSE;
    }

    fprintf(stderr, "\n%s:  No specified simulation time limit\n", SIMNAME);

    if (PingInterval)
      do {                               // until no more active terminals
        waitStat = SimDone.timed_wait(PingInterval);
	WritePing();
      } while (waitStat != EVENT_OCCURRED); // loops while still active terms 

    else                                 // no ping msgs, just wait quietly
      SimDone.wait();

    SimRunning = FALSE;                  // disables polling processes
  } 

  OutputResults();
//  return(0); kbaix
  return;
}

///////////////////////////////////////////////////////////////////////////////

bool OpenFiles(FILE* &configFile, FILE* &schemaFile, FILE* &wkldFile) {

  if (ConfigFileName == NULL ||
      SchemaFileName == NULL ||
      WkldFileName == NULL) {
    PrintHelp();  
    return(FALSE);
  }

  configFile = fopen(ConfigFileName, "r");
  if (configFile == NULL) {
    printf("%s:  error attempting to open configuration definition file %s\n", 
	   SIMNAME, ConfigFileName);
    return(FALSE);
  }

  schemaFile = fopen(SchemaFileName, "r");
  if (schemaFile == NULL) {
    printf("%s:  error attempting to open schema definition file %s\n", 
	   SIMNAME, SchemaFileName);
    return(FALSE);
  }

  wkldFile = fopen(WkldFileName, "r");
  if (wkldFile == NULL) {
    printf("%s:  error attempting to open workload definition file %s\n", 
	   SIMNAME, WkldFileName);
    return(FALSE);
  }

  return(TRUE);
}

///////////////////////////////////////////////////////////////////////////////

void PrintHelp() {
    printf("usage:\n\n%s \t-c config_file -s schema_file -w wkld_file -t timespec -d flags\n\t\t-B bufmgrtype -D disktype -e percentile -CpPqQ\n\nwhere:\n\n\t\t- timespec is one of ss, mm:ss, or hh:mm:ss in simulated time\n\t\t- bufmgrtype is '3' (3 level LRU)\n\t\t- disktype is 'c' (carey)\n\t\t- flags is for debug traces and is one of:\n\t\t\tB - Buffer Manager\n\t\t\tD - Disk\n\t\t\tI - I/O\n\t\t\tT - Terminals\n\t\t\tS - Scheduler\n\t\t\tO - Operators\n\t\t\tM - Messages\n\t\t\tL - Lock Manager\n\t\t\tQ - Queries\n\t\t\tP - Parser\n\t\t\tC - General control flow\n\t\t\tK - memory leak trace\n\t\t\tA - All traces\n\t\t-C flag turns on the CSIM trace output\n\t\t-P flag turns on 'plot mode' (makes results-extracting scripts easier)\n\t\t-p flag turns on 'ping mode' (prints progress messages)\n\t\t   The ping flag can take an optional parameter to specify the 'inter-ping'\n\t\t   time (in milliseconds of simulated time).\n",
	   SIMNAME);
}

///////////////////////////////////////////////////////////////////////////////

void SetCSIMLimits(){
  int i;

  // 1 per disk per node, one per node for cpu, 1 for network, 9 to spare
  i = max_facilities (NumNodes * (1 + DisksPerNode) + 10);  
  if (Flags[TRACE_CONTROL]) printf("number of CSIM facilities set to %d\n",i);

  // 1 per disk per node, one per cpu per node, n for network, 10 to spare
  i = max_servers(NumNodes * (CPUsPerNode + DisksPerNode) + 
		  MaxRequests + 10);  
  if (Flags[TRACE_CONTROL]) printf("number of CSIM servers set to %d\n",i);

  // 5 per job class, 1 per disk per node, one per node for cpu,
  // 1 for network, 9 to spare
  i = max_qtables (5 * MAX_CLASSES + NumNodes * (5 + 2 * DisksPerNode) + 10);  
  if (Flags[TRACE_CONTROL]) printf("number of CSIM qtables set to %d\n",i);

  // 5 per disk per node + 40 per max job classes + 8*20 for plan stats
  i = max_tables (NumNodes * 5 * DisksPerNode + 40 * MAX_CLASSES + 8 * 20);  
  if (Flags[TRACE_CONTROL]) printf("number of CSIM tables set to %d\n",i);

  // one histogram for each facility???
  i = max_histograms (NumNodes * (1 + DisksPerNode) + 10);  
  if (Flags[TRACE_CONTROL]) printf("number of CSIM histograms set to %d\n",i);

  // one mailbox for each operator on each node, plus global scheduler,
  // global deadlock detector, 1000 query schedulers (max MPL), and terminals
  i = max_mailboxes (NumNodes * (MAX_OPSPERNODE + 5) + MAX_TERMINALS + 1000);
  if (LocKBytes > 0) {
      if (GlobKBytes > 0) {
	  max_storages(i + NumNodes);
      } else {
	  max_storages(i);  // conditionally alloc local msg pool
      }
  } else { 
      if (GlobKBytes > 0) {
	  max_storages(NumNodes);
      }
  }
  if (Flags[TRACE_CONTROL]) printf("number of CSIM mailboxes set to %d\n",i);

  // 2 per disk per node (assume 30 I/Os in queue), 1 per buf page per node 
  // plus one per mailbox (processes wait on a recv when mb is empty)
  i = max_events (NumNodes*(2 * DisksPerNode * 30 + BufSize)+max_mailboxes(0));
  if (Flags[TRACE_CONTROL]) printf("number of CSIM events set to %d\n",i);

  // one mailbox for each operator on each node, plus global scheduler,
  // global deadlock detector, 1000 query schedulers (max MPL), and terminals
  i = max_processes (NumNodes * (MAX_OPSPERNODE + 5) + MAX_TERMINALS + 1000);
  if (Flags[TRACE_CONTROL]) printf("number of CSIM processes set to %d\n",i);

  // assume each operator on each node is sending messages to
  // operators on each of the other nodes, and a backlog of 1000 msgs max
  i = max_messages (500*NumNodes * (MAX_OPSPERNODE * NumNodes) + MAX_TERMINALS);
  if (Flags[TRACE_CONTROL]) printf("number of CSIM messages set to %d\n",i);

}

///////////////////////////////////////////////////////////////////////////////

void OutputConfig(){
  char *diskCacheStat = "", *bufmgrType = "??", *diskType = "??";
  
  if (DiskCachePages == 0 || DiskCacheContexts == 0) 
    diskCacheStat = "(disabled)";
  
  if (BufMgrType == '3') bufmgrType = "3 level LRU";
    
  if (DiskType == 'c') diskType = "Carey";

  printf("\nHardware Configuration\n----------------------\n  %15d nodes (1 host, %d with disks)\n  %15d disks/node\n  %15d processors/node\n  %15.5f MIPS/processor\n  %15d Kbytes RAM/node\n  %15d byte pages\n  %15d instr to start an I/O\n  %15s buffer manager\n\n  %15s disk\n  %15d disk cache contexts %s\n  %15d disk cache pages per context %s\n  %15d pages/cylinder\n  %15.5f msecs min rotational latency\n  %15.5f msecs/rotation\n  %15.5f msecs settle time\n  %15.5f seek factor\n  %15.5f Mbytes/sec transfer rate\n  %15.5f msecs/page transfer time\n\n  %15d maximum n/w packet bytes\n  %15d maximum concurrent n/w requests\n  %15d local n/w buffer Kbytes (per mailbox)\n  %15d global n/w buffer Kbytes\n  %15.5f minimum send protocol instructions\n  %15.5f send protocol instructions per byte\n  %15.5f minimum receive protocol instructions\n  %15.5f receive protocol instructions per byte\n  %15.5f mimimum wire delay\n  %15.5f wire delay per byte\n  %15.5f minimum short circuit instructions\n  %15.5f short circuit instructions per byte\n\n",
	 NumNodes, NumNodes-1,
	 DisksPerNode, 
	 CPUsPerNode, 
	 Mips, 
	 KBytesPerNode,
	 PageSize,
	 Inst_StartIO,
	 bufmgrType,
	 diskType,
	 DiskCacheContexts,
	 diskCacheStat,
	 DiskCachePages,
	 diskCacheStat,
	 PagesPerCyl,
	 MinLatency,
	 DiskRotation,
	 SettleTime,
	 SeekFactor,
	 XferRate,
	 XferTime,
	 MaxPacketBytes,
	 MaxRequests,
	 LocKBytes,
	 GlobKBytes,
	 MinSendProtocol,
	 SendProtocolPerByte,
	 MinRecvProtocol,
	 RecvProtocolPerByte,
	 MinDelay,
	 DelayPerByte,
	 MinShortInstr,
	 ShortInstrPerByte);
}

///////////////////////////////////////////////////////////////////////////////

void OutputSchema(){}

///////////////////////////////////////////////////////////////////////////////

void OutputWorkload(){}

///////////////////////////////////////////////////////////////////////////////

void OutputResults(){
  int i;
  char plotID[5];

  printf("\n\nElapsed simulated time:  %f msecs\n", simtime());

  // Writes out global scheduler statistics

  printf("\n\nScheduling Statistics:\n----------------------\n\n");
  printf("\tAverageMPL\t%.2f\n\tMaxMPL\t\t%d\n", 
	 TimeWeightedMPL / (clock - FirstMPLEntry), MaxMPL);
  printf("\n\nGlobal Memory Wait Queue Statistics:\n------------------------------------\n");
  GlobalMemoryWaitQueue->printStats();

  printf("\n\nGlobal MPL Wait Queue Statistics:\n------------------------------------\n");
  GlobalMPLWaitQueue->printStats();

  // Writes out job class related statistics for each class

  printf("\n\nJob Class Statistics:\n---------------------\n");
  for (i = 0; i < NumClasses; i++) JobClasses[i]->writeStats();

  // If one or more plans have been marked for statistics gathering, prints
  // out the stats for each one.

  if (PlanStats) {
    printf("\n\nPlan Statistics:\n---------------\n");
    PlanSyms.writeStats();
  }
    
  // Writes out network related statistics
  
  if (PlotMode) {
      strcpy(plotID, "NW");
  } else {
      strcpy(plotID, "");
  }
  
  printf("\n\nNetwork Statistics:\n-------------------\n\n");
  printf("   tot Mbytes   tot Mbytes    b/w reqmt   mean      mean      mean     total  %%send\n");      
  printf("      on wire   short circ  (Mbits/sec)   util  q length  svc time     compl  block\n");
  printf("  -----------  -----------  -----------  -----  --------  --------  --------  -----\n%s  %11.3f  %11.3f  %11.5f  %5.3f  %8.3f  %8.3f  %8d  %5.1f\n",
	 plotID,
	   (TotNWBytesSent == 0.0) ? 0.0 : 
	 TotNWBytesSent / (1024.0 * 1024.0),
	 (TotNWBytesShort == 0.0) ? 0.0 : 
	 TotNWBytesShort / (1024.0 * 1024.0),
	 (TotNWBytesSent == 0.0) ? 0.0 : 
	 ((TotNWBytesSent * 8.0) / (1024.0 * 1024.0)) / (simtime()/1000.0),
	 Network->util(),
	 Network->qlen(),
	 Network->serv(),
	 Network->completions(),
	 (TotSends == 0.0) ? 0.0 : 
	 TotBlockedSends / TotSends * 100.0
	 );

  // Writes out CPU related statistics for each node

  printf("\n\nCPU Statistics:\n---------------\n\n");
  printf("  Node   util   thruput  response  q length  svc time     compl\n");
  printf("  ----  -----  --------  --------  --------  --------  --------\n");
  for (i = 0; i < NumNodes; i++) Nodes[i]->printCPUStats();

  // Writes out buffer related statistics for each node

  printBufStats();
    
  // Writes out disk related statistics for each node
      
  printf("\n\nDisk Statistics:\n----------------\n\n");
  printf("              arm    resp avg svc min svc max svc std dev cache   cntlr     arm   avg  max\n");
  printf("  Node disk  util @ cntlr  @ disk  @ disk  @ disk  @ disk hit %%   compl   compl  qlen qlen\n");
  printf("  ---- ---- ----- ------- ------- ------- ------- ------- ----- ------- ------- ----- ----\n");
  for (i = 0; i < NumNodes; i++) Nodes[i]->printDiskStats();

  if (CSIMReports) report();             // writes out CSIM standard reports
  mdlstat();                 // prints resources used by simulation

}

///////////////////////////////////////////////////////////////////////////////

void      WritePing(){
  int i, totCompletions = 0;
  char primClassMsg[80];

  primClassMsg[0] = '\0';

  for (i = 0; i < NumClasses; i++) {
    totCompletions += (int) JobClasses[i]->completions;
    if (JobClasses[i]->stopCount >= 0) // primary job class?
      sprintf(primClassMsg, "(%8d for primary job class)",
	      JobClasses[i]->completions);
  }
  
  fprintf(stderr, "%s: time %10.2f, %8d completions %s\n",
	  SIMNAME, simtime(), totCompletions, primClassMsg);
}

///////////////////////////////////////////////////////////////////////////////

void printBufStats() 
{
  int i;
  char plotID[256];

  if (PlotMode) {
      sprintf(plotID, "       "); 
  } else {
      sprintf(plotID, "  ");
  }

  printf("\n\nBuffer Statistics:\n------------------\n\n");
  printf("%s                  total     total     total     total  avg qlen  avg qlen  avg qlen  avg qlen  avg qlen\n", plotID);  
  printf("%s  Node  hit %%      hits      refs    writes     reads   invalid     valid     hated     liked     loved\n", plotID);
  printf("%s  ----  -----  --------  --------  --------  --------  --------  --------  --------  --------  --------\n", plotID);
  
  for (i = 0; i < NumNodes; i++) {
      Nodes[i]->bufMgr->PrintBufferPoolStats();
  }
}

