#include "dbsim.h"
#include "object.h"
#include "language.h"
#include "terminal.h"

///////////////////////////////////////////////////////////////////////////////
//
//  Name:        source.c
//
//  Function:    Open class arrival source.  Waits for appropriate inter-
//               arrival time, instantiates a terminal which submits a single
//               plan and deletes itself, and waits again.  Exits when the
//               simulation timer pops.
//                 
//  Change Activity:
//
//     10/92 (KPB) - Initial version.
//
///////////////////////////////////////////////////////////////////////////////

extern int TermIDCounter;

extern "C" {
  void exit(int code);
}

void Source(JobClass *jc) {
  float         waitTime;
  Terminal*     t;
  
  create("source");                       // define a CSIM process
  if (Flags[TRACE_TERM] || Flags[TRACE_CONTROL]) 
    printf("Arrival source process, class %s started\n", jc->name);

  // If the "waitsecs" parameter was specified for this class, then 
  // sits and waits the specified amount of time before submitting any
  // queries.

  if (jc->waitSecs != NULL) {
    waitTime = jc->waitSecs->choose() * 1000.0; // convert secs->msecs
    hold(waitTime);
  }

  // Loops until the simulation timer pops or the plan sequence is exhausted.

  while(SimRunning) {
     // Before sending in another query, see if the system is already 
    // saturated (current MPL of this class is already at its maximum MPL).
    // Scheduler kindly maintains these statistics for us.

    if (jc->currentMPL == jc->maxMPL) {
      printf("\nSystem saturated by open job class %s.\n", jc->name);
      fprintf(stderr, "\nSimulation terminated, system saturated.\n",
	      jc->name);
      SimRunning = FALSE;
      SimDone.set();
      return;
    }

    t = new Terminal(TermIDCounter++, jc); // start a (short-lived) terminal
    
    // Checks if the plan sequence is exhausted, and stops if so.

    if (jc->plans->atEnd()) break;

    // Waits for the specified interarrival time

    waitTime = jc->interArrivalSecs->choose() * 1000.0; // convert sec->msec
    hold(waitTime);
  }
  SimDone.wait(); // just wait for last query to complete
}
