#include "dbsim.h"
#include "object.h"
#include "mailbox.h"
#include "terminal.h"
#include "language.h"
#include <string.h>

///////////////////////////////////////////////////////////////////////////////
//
//  Name:        terminal.c
//
//  Function:    Terminal class implementations
//                 
//  Change Activity:
//
//     08/xx/91 (KPB) - Initial version.
//
///////////////////////////////////////////////////////////////////////////////

Terminal::Terminal(int ID, JobClass* jc) {
  ActiveTerms++; // update total number of active terminals
  id = ID;
  jobClass = jc;

  // For closed classes, there may be multiple terminals active within the
  // class, and therefore we must clone a private copy of plans and think times
  // for each of them.  Plans and think times may contain or be sequences, 
  // and each terminal should have its own independent sequence counters.  
  // The clone methods will recursively clone as needed.  
  // 
  // This is not an issue for open classes, since even tho in some instances 
  // there may be multiple terminals active for the class, each terminal 
  // only submits one plan and then exits.  So there is no notion of a 
  // sequence within any particular terminal, only within the class.  
  // 
  // Avoiding private copies of the random plan structure for open 
  // class terminals also eliminates the need to clean up those private copies
  // when the terminal disappears after submitting its single plan.  Even tho 
  // closed class terminals have private copies of the random plan, when they
  // disappear the simulation is in the process of terminating, so we dont
  // have to clean them up either (whew).

  if (jc->interArrivalSecs == NULL) {       // this term for a closed class?
    thinksecs = jc->thinksecs->clone();     // clone private copy of think time
    plans = jc->plans->clone();             // clone private copy of plans
    if (jc->waitSecs != NULL)               // need to wait before submitting?
      waitSecs = jc->waitSecs->choose();    // pick a wait time for this term
  }
  else {
    thinksecs = NULL;                       // not applicable
    waitSecs = 0.0;                         // source waits, not terminals
    plans = jc->plans;                      // just point to the class plans
  }

  TermProc(this);                           // start up terminal process
}

void Terminal::print(int indent) {
  printf("Terminal id %d:", id);
  NewSubLine;
  jobClass->print(indent+PRINTINDENT);
}
