#include "dbsim.h"
#include "object.h"
#include "msg.h"
#include "mailbox.h"
#include "node.h"
#include "language.h"
#include "db.h"
#include "terminal.h"
#include "split.h"
#include "query.h"
#include "bufmgr.h"
#include <math.h>

///////////////////////////////////////////////////////////////////////////////
//
//  Name:        termproc.c
//
//  Function:    Terminal process.  Thinks and submits queries according to
//               its job class description.  Gathers statistics.
//                 
//  Change Activity:
//
//     10/91 (KPB/DJD) - Initial version.
//
///////////////////////////////////////////////////////////////////////////////

extern Mailbox* GlobalSchedPort;
extern event    SimDone;
extern Node*    Nodes[];
extern float WarmupPeriod;

extern void     attemptAdmission();
void            cleanUp(QueryPlan*);

extern "C" {
  void exit(int code);
//  double sqrt(double); kbaix
}

void TermProc(Terminal *t) {
  float         thinkTime;
  TIME	        startTime, endTime;
  char*         planName;
  bool          reset = FALSE;         // plan sequence has reset to beginning
  bool          closedClass;           // terminal is part of a closed class
  distObjType   objType;               // dummy used to satisfy plan::choose
  RandomPlan*   rp;
  QueryPlan*	qp;
  QueryMsg*	reqMsg;
  QueryMsg*	replyMsg;
  JobClass*     jc = t->jobClass;      // to shorten ptr expressions
  
  create("terminal");                  // define a CSIM process
  if (Flags[TRACE_TERM] || Flags[TRACE_CONTROL]) 
    printf("Terminal process %d, class %s started\n",t->id, t->jobClass->name);

  t->port = new TermMailbox("TermPort", t); // get a mail address (must be
	       // allocated inside this process so it can be deleted here too)

  closedClass = (jc->interArrivalSecs == NULL); //think time will exist if true

  // If this terminal has to wait before submitting queries, then does so
  // (closed class terminals only -- source waits for open classes).

  if (t->waitSecs != 0.0) hold(t->waitSecs * 1000.0);

  // For closed class terminals, keeps submitting plans until simulation timer
  // pops (SimRunning == FALSE) or there are no more plans (reset == TRUE).

  do {

      // If this terminal is part of a closed class population, then waits for
      // the appropriate think time (may be zero -- batch jobs).

      if (closedClass) {
	  thinkTime = t->thinksecs->choose() * 1000.0; // pick time in msec
	  hold(thinkTime);                  // waits for think time
      }
      
      planName = (char*) t->plans->choose(objType, reset); // pick a plan name
      if (reset && closedClass) break;    // end of plan list for this terminal
    
      rp = (RandomPlan*) PlanSyms.lookup(planName); // find random plan object
      if (rp == NULL) {
	  printf("ERROR:  terminal %d found an undefined plan: %s\n",
		 t->id, planName);
	  exit(4);
      }
      
      qp = rp->choose(NULL, jc);         // stamp out a specific plan instance
      
      // Croaks if the root operator of a plan is not a STORE or TERMSEND.  
      // This must be the case, since the result tuples from a query must go
      // somewhere.  KLUDGE !!!!!!!!  Future update/insert ops could also be
      // valid roots!

      if ((qp->opCode != STORE) && (qp->opCode != TERMOP))
        printf("\nERROR: Query rejected - root of query plan not a store/termsend\n\n");
      else {
	  reqMsg = new QueryMsg();           // get a query message
	  reqMsg->query = qp;                // stuff plan in message
    
	  // Sends the query to the global query scheduler

	  if(Flags[TRACE_TERM]) {
	      printf("%f:  term %d class %s sending query to global sched.\n ",
		   clock, t->id, jc->name);
	      if (Flags[TRACE_QUERY]) qp->print(0);
	  }
    
	  startTime = clock;
	  reqMsg->send(GlobalSchedPort, (Mailbox*) t->port);
	  
	  // Waits for a reply
	  
	  replyMsg = (QueryMsg*) t->port->recv();
	  endTime = clock;
	  float resp = endTime - startTime;
	  
	  // Accumulates query stats into job class statistics in CSIM tables 
	  // (these table objects delay updates until statwaitsecs pass).
	  
	  qp->accumStats();                 // roll up plan stats to root node
	  
	  jc->bufRefs->record(qp->cnt_bufRef);
	  jc->bufHits->record(qp->cnt_bufHit);
	  jc->bufDirtys->record(qp->cnt_bufDirty);
	  jc->diskReads->record(qp->cnt_diskRead);
	  jc->diskCacheHits->record(qp->cnt_diskCacheHit);
	  jc->CPUInstr->record(qp->cnt_CPUInstr);
	  jc->respTimes->record(resp);
	  jc->execTimes->record(qp->execTime);
	  jc->msgsSent->record(qp->cnt_msgsSent);
	  jc->msgsReceived->record(qp->cnt_msgsReceived);
	  jc->blockedSends->record(qp->cnt_blockedSends);
	  jc->blockedRecvs->record(qp->cnt_blockedRecvs);
	  jc->bytesSent->record(qp->cnt_bytesSent);
	  jc->bytesReceived->record(qp->cnt_bytesReceived);
	  jc->sendTime->record(qp->sendTime / qp->cnt_msgsSent);
	  jc->recvTime->record(qp->recvTime / qp->cnt_msgsReceived);
	  if (simtime() > jc->statWaitSecs * 1000.0) {
	      jc->completions++;         // bump completions for throughput
	      if (jc->completions == jc->maxCompletions) reset = TRUE;
	  }
	  
	  // If root node of query is flagged for plan based statistics, record
	  // response times on a plan basis, as well.
	  
	  if (qp->statsPlan != NULL) {
	      qp->statsPlan->respTimes->record(resp);
	      qp->statsPlan->execTimes->record(qp->execTime);
	      qp->statsPlan->completions++;
	  }

	  // start queries from classes until they are at their maxMPLs
	  attemptAdmission();
	  
	  if(Flags[TRACE_TERM]) {
	      printf("%f:  term %d class %s, query executed in %5.2f ms.", 
		     clock, t->id, jc->name, resp);
	      printf(" started at %f\n", startTime);
	  }
	
	  delete replyMsg;
      }
      cleanUp(qp);                         // get rid of query plan storage
  } while(SimRunning && !reset && closedClass);
  
  // If this is the last active terminal in the system, or this is the last
  // active terminal of a "primary" job class, then shuts down the simulation
  
  if (closedClass) {
    if ((--ActiveTerms == 0) || (--(jc->stopCount) == 0) ||
	(jc->completions == jc->maxCompletions)) {
      SimRunning = FALSE;
      SimDone.set(); 
    }
  } else {
    if ((jc->stopCount >= 0) && jc->plans->atEnd()) {
      SimRunning = FALSE;
      SimDone.set(); 
    }
  }
  
  delete t->port;                       // delete terminal mailbox
  delete t;                             // delete terminal object 
}
     
///////////////////////////////////////////////////////////////////////////////

void cleanUp(QueryPlan* node) { // recursive proc to clean up nodes of a query
  StorePlan*	storeOp;
  JoinPlan*	joinOp;
  SelectPlan*	selectOp;
  TermPlan*	termOp;
  
  switch (node->opCode)   {
  case STORE:
    storeOp = (StorePlan *) node;
    cleanUp(storeOp->child);
    delete storeOp;
    break;
  case TERMOP:
    termOp = (TermPlan *) node;
    cleanUp(termOp->child);
    delete termOp;
    break;
    
  case SELECT:
    selectOp = (SelectPlan *) node;
    delete selectOp;
    break;
    
  case JOIN:
    joinOp = (JoinPlan *) node;
    cleanUp(joinOp->leftChild);
    cleanUp(joinOp->rightChild);
    delete joinOp;
    break;

  default:
    printf("Termproc.cleanUp, unknown opcode=%d\n",node->opCode);
  }
}
